/****************************************************************************
**
** Copyright (C) 2012 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Lesser General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public
** License along with this library; if not, write to the
** Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
**
****************************************************************************/
/*
 * ChangeLog -  
 *				July 05, 2012
 * 				1. Added logic to limit the max frame update rate to 10.
 *              2. Added checks for detecting and handling socket errors.
 *              3. Appended missing '\n' in RFB protocol version string
 *              4. Added readSecurityFailReason() function to read the fail reason string and
 *                 also handles socket reconnection in this case.
 *              5. Changed the sequence of sending the framebufferUpdateRequest().
 *                 Its now sent only after complete VNC frame is received.
 *                 Previously it was getting sent before reading last frame rectangle.
 *              6. Added logic to determine the up/down scaling support based on the server side scaling capability.
 *              7. Added logic to request the best fitresolution values to  TM server that support scaling.
 *              8. Added ogic to calculate the scalling factors depending on the negotiated resolution values
 *                 and then notifying it to application.
 *				March 19, 2013
 *              1. Updated the fps limit to 20.
 *              2. Added vncPauseFbUpdateRequest() and vncResumeFbUpdateRequest() api for pausing/resuming the fb updates from the ML server.
 *              3. Added support to enable/disable the consideration of pixel compensation of client display panel.
 *              4. Added support to handle knob key and media key events.
 *              5. Added suooprt for integrating the double buffer implementation.
 *
 *              August 02, 2013
 *              1. Changed the location of sending of 1st FB update request from after clientDisplayConfig() to after clientEventConfig().
 *              2. Added 500msec delay before sending 1st FB update request.
 */

#include "QTmConfiguration.h"
#include "QTmRfbClient.h"
#include "QTmApplication.h"
#include <TmHeaders/TCPSocket.h>
#include <QProcess>
#include <QTimer>
#include <sys/timeb.h>
#include <errno.h>
#include <TmHeaders/libinterface.h>
#include <TmHeaders/appinterface.h>

#define MAX_FB_RESOUTION_WIDTH_BYTES (1920*4)       // considering RGB888 as color format as it has max value of bits per pixel
/*!
  \class QTmRfbClient

  \brief This class implements the RFB client protocol, including the Terminal Mode extensions, defined in the version 1.0.

  Default values
  \list
  \o QTmConfiguration::ConfigContextInfoEncoding = true
  \o QTmConfiguration::ConfigTerminalModeEncoding = true
  \o QTmConfiguration::ConfigFbOrientation = true
  \o QTmConfiguration::ConfigFbRotation = false
  \o QTmConfiguration::ConfigFbReplacement = false
  \o QTmConfiguration::ConfigDisplaySizeWidth, = 0
  \o QTmConfiguration::ConfigDisplaySizeHeight = 0
  \o QTmConfiguration::ConfigDisplayDistance = 0
  \o QTmConfiguration::ConfigVirtualKeyboard = true
  \o QTmConfiguration::ConfigKeyEventListing = true
  \o QTmConfiguration::ConfigKeyEventMapping = true
  \o QTmConfiguration::ConfigIncrementalUpdate = true
  \o QTmConfiguration::ConfigRunLengthEncoding = true
  \o QTmConfiguration::ConfigFbUpScaling = true
  \o QTmConfiguration::ConfigFbDownScaling = true
  \o QTmConfiguration::ConfigDesktopSizeEncoding = true
  \o QTmConfiguration::ConfigDisplayPixelWidth = 0
  \o QTmConfiguration::ConfigDisplayPixelHeight = 0
  \endlist
 */

const int MSG_FRAMEBUFFER_UPDATE          =   0;     // RFB FramebufferUpdate message identifier
const int MSG_SET_COLOUR_MAP_ENTRIES      =   1;     // RFB SetColourMapEntries message identifier
const int MSG_BELL                        =   2;     // RFB Bell message identifier
const int MSG_SERVER_CUT_TEXT             =   3;     // RFB ServerCutText message identifier

const int MSG_TERMINAL_MODE               = 128;     // RFB TerminalMode extension message identifier

const int MSG_SET_PIXEL_FORMAT            =   0;     // RFB SetPixelFormat message identifier
const int MSG_SET_ENCODINGS               =   2;     // RFB SetEncodings message identifier
const int MSG_FRAMEBUFFER_UPDATE_REQUEST  =   3;     // RFB FramebufferUpdateRequest message identifier
const int MSG_KEY_EVENT                   =   4;     // RFB KeyEvent message identifier
const int MSG_POINTER_EVENT               =   5;     // RFB PointerEvent message identifier
const int MSG_CLIENT_CUT_TEXT             =   6;     // RFB ClientCutText message identifier

//VNC Extensions message subtypes
const int TM_MSG_SERVER_DISPLAY_CONFIG    =   1;    // RFB ServerDisplayConfiguration message
const int TM_MSG_CLIENT_DISPLAY_CONFIG    =   2;    // RFB ClientDisplayConfiguration message
const int TM_MSG_SERVER_EVENT_CONFIG      =   3;    // RFB ServerEventConfiguration message
const int TM_MSG_CLIENT_EVENT_CONFIG      =   4;    // RFB ClientEventConfiguration message
const int TM_MSG_EVENT_MAPPING            =   5;    // RFB EventMapping message
const int TM_MSG_EVENT_MAPPING_REQ        =   6;    // RFB EventMappingRequest message
const int TM_MSG_KEY_EVENT_LISTING        =   7;    // RFB KeyEventListing message
const int TM_MSG_KEY_EVENT_LISTING_REQ    =   8;    // RFB KeyEventListingRequest message
const int TM_MSG_KEYBOARD_TRIGGER         =   9;    // RFB KeyboardTrigger message
const int TM_MSG_KEYBOARD_TRIGGER_REQ     =  10;    // RFB KeyboardTriggerRequest message
const int TM_MSG_DEVICE_STATUS            =  11;    // RFB DeviceStatus message
const int TM_MSG_DEVICE_STATUS_REQ        =  12;    // RFB DeviceStatusRequest message
const int TM_MSG_CONTENT_ATTESTATION_RESP =  13;    // RFB DeviceAttestationResponse message
const int TM_MSG_CONTENT_ATTESTATION_REQ  =  14;    // RFB DeviceAttestationRequest message
// Value 15 reserved
const int TM_MSG_FRAMEBUFFER_BLOCKING     =  16;    // RFB FramebufferBlockingNotification message
// Value 17 reserved
const int TM_MSG_AUDIO_BLOCKING           =  18;    // RFB AudioBlockingNotification message
// Value 19 reserved
const int TM_MSG_TOUCH_EVENT              =  20;    // RFB TouchEvent message
const int TM_MSG_FRAMEBUFFER_ALT_TEXT     =  21;    // RFB FramebufferAlternativeText message
const int TM_MSG_FRAMEBUFFER_ALT_TEXT_REQ =  22;    // RFB FramebufferAlternativeTextRequest message

const int EXT_MSG_TERMINAL_MODE           = 128;
const int EXT_ENC_RAW                     =   0;    // Encoding identifier for Raw encoding (i.e. no encoding applied)
const int EXT_ENC_DESKTOP_SIZE            =-223;
const int EXT_ENC_TERMINAL_MODE           =-523;
const int EXT_ENC_TM_CONTEXT_INFO         =-524;
const int EXT_ENC_RUN_LENGTH_ENCODING     =-525;    // Encoding identifier for Run Length Encoding (scan-line based)

#define MIN_FB_UPDATE_REQUEST_INTERVAL      50     //msec

#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
#include <sys/timeb.h>
static struct timeb PerformanceStartTime;
struct timeb        PerformanceEndTime;
double              PerformanceDiffTime, PerformanceFpsCount;
int                 PerformanceSampleWindow = 100;
int                 PERFORMANCE_counter = 0;
int                 PERFORMANCE_number_of_rectangles = 0;
int	                PERFORMANCE_number_of_pixel = 0;
int                 PERFORMANCE_number_of_runs;
#endif


/*!
  \enum QTmRfbClient::QTmSessionStatus
  This enumberation lists the possible status values the VNC client can be in.
  \value RFB_SESSION_NONE RFB session is in default state.
  \value RFB_SESSION_CONNECTED_ACTIVE RFB session is initialized and active, i.e. framebuffer updates are ongoing.
  \value RFB_SESSION_CONNECTED_PASSIVE RFB session is initialized but passive, i.e. no framebuffer updates are ongoing.
  \value RFB_SESSION_SERVER_VERSION VNC client is receiving VNC server RFB protocol version
  \value RFB_SESSION_SECURITY_TYPES VNC client is receiving the supported security types
  \value RFB_SESSION_SECURITY_RESPONSE VNC client is receiving the security response
  \value RFB_SESSION_SERVER_INIT VNC client is receiving the RFB Server Init message
  \value ERROR_RFB_SESSION_DISCONNECTED Error, VNC session got disconnected.
  \value ERROR_WRONG_SERVER_VERSION Error, wrong RFB server version received from VNC server
  \value ERROR_WRONG_SECURITY_TYPES Error, wrong security type received from VNC server
  \value ERROR_WRONG_FRAMEBUFFER_ENCODING Error, unknown framebuffer encoding received from VNC server
  \value ERROR_WRONG_MSG_TYPE Error, wrong RFB message type received from VNC server
  \value ERROR_WRONG_FSM_STATE Error, wrong FSM state.
  \value ERROR_CANNOT_ALLOCATE_SURFACE Error, client cannot allocate framebuffer surface.
 */

// ********************************************************************
// *****            Class Constructor and Destructor              *****
// ********************************************************************
/*!
  Class Constructor.
  Set the default values. \a parent gives the objects parent.
 */
#ifdef QT_UI_CLEANUP
QTmRfbClient::QTmRfbClient(QObject *parent)
    : QThread(parent),m_threadListenr(0),m_clientId(0),m_socketError(false),m_displayStatusFlag(false)
#else
QTmRfbClient::QTmRfbClient(QObject *parent)
    : QObject(parent)
#endif
    , m_connected(false)
    , m_eventMap(0)
    , m_surface(0)
    , m_url(0)
    , m_socket(0)
    , m_rfbSessionStatus(RFB_SESSION_NONE)
    , m_keyEventCounter(0)
    , m_framebuffer(0)
    , m_frameCount(0)
    , m_timeLastFBRequest(0)
    , m_fbUpdateRequestFlag(false)
    , m_timeRoundTrip(0)
{
#ifdef QT_DEBUG_TM_RFBCLIENT
    m_rfbConfigClient = new QTmConfiguration("(Client)");
    m_rfbConfigServer = new QTmConfiguration("(Server)");
#else
    m_rfbConfigClient = new QTmConfiguration();
    m_rfbConfigServer = new QTmConfiguration();
#endif
    m_eventMap        = new QHash<unsigned int, unsigned int>();
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigFbUpdates,     false);
    //m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::RGB888);
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::RGB565);
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigDesktopSizeEncoding, true);
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigTerminalModeEncoding, true);
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigContextInfoEncoding, true);
    qDebug () << "QTmRfbClient::QTmRfbClient:" << "Instance constructed at" << this;

    /*!@{*/
    m_isSendingFbUpdateRequestFirstTime = true;
    /*!@}*/

#ifdef VNC_OPTIMIZATION
    m_encodingCmdBuffer = (unsigned char*) malloc(MAX_FB_RESOUTION_WIDTH_BYTES);
#endif
}
/*!
  Class destructure.
  Disconnect the session prior deletion.
 */
QTmRfbClient::~QTmRfbClient() {
    qDebug () << "QTmRfbClient::~QTmRfbClient:" << "Instance destroyed at" << this;
    if (m_connected)
    {
        qDebug () << "QTmRfbClient::~QTmRfbClient: sessionStop()";
        sessionStop();
    }
    if(m_socket)
    {
        qDebug () << "QTmRfbClient::~QTmRfbClient: delete m_socket";
        delete m_socket;
        m_socket = 0;
    }

    free(m_encodingCmdBuffer);
}


// ********************************************************************
// *****                  VNC Sessiong Handling                   *****
// ********************************************************************
/*!
  \property QTmRfbClient::sessionActive
  \brief Holds the information, whether the VNC session is active or not.
 */
/*!
  \property QTmRfbClient::sessionUrl
  \brief Holds the URL to the VNC server. Is set to QUrl(), if client is not connected.
 */
/*!
  \fn QTmRfbClient::sessionConnected(QString serverName)
  Signal emitted from a RFB client to inform that the VNC client is connected to the server and
  the VNC session is initialized. The name of the VNC server is given in \a serverName
  This signal is emitted, when the client has received the RFB Server Init message.
 */
/*!
  \fn QTmRfbClient::sessionDisconnected()
  Signal emitted from a RFB client to inform that the VNC session is lost.
  This signal is emitted from the protocolFsm() on reception of an invalid message or whenever
  the TCP connection is broken while transmitting or receiving data.
 */
/*!
  Start a VNC session and connect the VNC client to a VNC server at the URL \a url.
  The TCP socket is connected and the readReady() signal is connected to the protocolFsm().
  Returns true, if the VNC session has been started. False is returned otherwise.
  Note: The VNC session is not connected (i.e. initialized), until the sessionConnected() signal is received.
 */
#ifdef QT_UI_CLEANUP
bool QTmRfbClient::sessionStart(QUrl url) {
    m_connected = false;
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::sessionStart:" << "Start VNC Session with" << url;
#endif
    //m_mutex.lock();
    m_socket = new TCPSocket();
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "TCPSocket created" << m_socket->getState();
#endif
    int connectStatus = m_socket-> connectToHost(url.host().toLatin1().data(), url.port());
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "TCPSocket connect status :" << connectStatus;
#endif
    if(!connectStatus)
    {
        m_connected = true;
        m_url = url;
        m_rfbSessionStatus = RFB_SESSION_SERVER_VERSION;
    }
    else
    {
        m_socket->closeSocket();
    }
    //m_mutex.unlock();
    return m_connected;
}

void QTmRfbClient::sessionStop() {
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::sessionStop:" << "Stop VNC Session with" << m_url << m_connected;
#endif
    //m_mutex.lock();
    if (!m_connected) {
      //  m_mutex.lock();
        return;
    }
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "About to close socket" << m_socket;
#endif
    int closeStatus = -1;
    if(m_socket){
        closeStatus = m_socket->closeSocket();
    }

    m_connected = false;
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "Close socket status :" << closeStatus;
#endif
    m_url       = QUrl();
    m_rfbSessionStatus = RFB_SESSION_NONE;
    if (m_surface)
        m_surface->deallocateSurface();
    qDebug()<< "About to exit from QTmRfbClient::sessionStop() " << m_connected;
    return;
}


#else
bool QTmRfbClient::sessionStart(QUrl url) {
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::sessionStart:" << "Start VNC Session with" << url;
#endif
    m_mutex.lock();
    if (m_connected) {
        m_mutex.unlock();
        return false;
    }
    m_socket = new QTcpSocket(this);
    m_socket->connectToHost(url.host(), url.port());
    if (!m_socket->waitForConnected()) {
#ifdef QT_DEBUG_TM_RFBCLIENT
        qDebug() << "QTmRfbClient::sessionStart:" << m_socket->errorString();
#endif
        m_mutex.unlock();
        sessionStop();
        return false;
    }
    m_connected = true;
    m_url = url;
    m_rfbSessionStatus = RFB_SESSION_SERVER_VERSION;
    connect(m_socket, SIGNAL(readyRead()),
            this,     SLOT  (protocolFsm()),
            Qt::AutoConnection);
    m_mutex.unlock();
    return true;
}

void QTmRfbClient::sessionStop() {
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::sessionStop:" << "Stop VNC Session with" << m_url;
#endif
    m_mutex.lock();
    if (!m_connected) {
        m_mutex.unlock();
        return;
    }
    m_connected = false;
    m_url       = QUrl();
    m_rfbSessionStatus = RFB_SESSION_NONE;
    //deallocate surface
    if (m_surface)
        m_surface->deallocateSurface();
    //disconnect socket
    disconnect(m_socket, SIGNAL(readyRead()), this, SLOT(protocolFsm()));
    m_socket->disconnectFromHost();
    m_socket->close();
    delete m_socket;
    m_socket = 0;
    m_mutex.unlock();
    return;
}
#endif

/*!
  Stop the current VNC session.
  The surface, holding the framebuffer update is de-allocated. The TCP socket is disconnected.
 */

/*!
  Return true if the session is connected. False is retured otherwise.
 */
bool QTmRfbClient::sessionActive() {
    return m_connected;
}
/*!
  Returns the URL to the VNC server, the client is connected to.
 */
QUrl QTmRfbClient::sessionUrl() {
    return m_url;
}
/*!
  Returns the current RFB Session Status.
 */
QTmRfbClient::QTmSessionStatus QTmRfbClient::sessionStatus() {
    return m_rfbSessionStatus;
}
/*!
  Returns the pointer to the currently valid key event map.
 */
QHash<unsigned int, unsigned int> *QTmRfbClient::eventMap() {
    return m_eventMap;
}
/*!
  The the surface, the VNC client must use for the framebuffer updates.
  \a surface provides the pointer to the instance to use. The instance is already allocated.
 */
void QTmRfbClient::setSurface(QTmSurface *surface) {
    m_surface = surface;
}


// ********************************************************************
// *****                    VNC Configuration                     *****
// ********************************************************************
/*!
  Change the VNC client configuration.
  The \a key is set to \a value.
 */
void QTmRfbClient::setClientConfig(QTmConfiguration::QTmConfig key, unsigned int value) {
    m_rfbConfigClient->setConfiguration(key, value);
}
/*!
  Returns the VNC client configuration of \a key.
  In case the key has not been defined, the \a defaultValue is returned instead.
 */
unsigned int QTmRfbClient::clientConfig(QTmConfiguration::QTmConfig key, unsigned int defaultValue) {
    return m_rfbConfigClient->configuration(key, defaultValue);
}
/*!
  Returns the VNC server configuration of \a key.
  In case the key has not been defined, the \a defaultValue is returned instead.
 */
unsigned int QTmRfbClient::serverConfig(QTmConfiguration::QTmConfig key, unsigned int defaultValue) {
    return m_rfbConfigServer->configuration(key, defaultValue);
}


/**
 * RFB Protocol Handshake
 */
void QTmRfbClient::protocolHandshake() {
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::protocolHandshake()";
#endif
    unsigned char cmdBuffer[12];
    QByteArray protocolVer;
    int receiveBytes = receiveData(12, (char *)cmdBuffer);
    if(receiveBytes < 12)
    {
        qDebug()<< "Error coccured while protocolHandshake()";
        handleSocketError();
        return;
    }
    int major = ((int) cmdBuffer[ 6]-48) + ((int) cmdBuffer[5]-48)*10 + ((int) cmdBuffer[4]-48)*100;
    int minor = ((int) cmdBuffer[10]-48) + ((int) cmdBuffer[9]-48)*10 + ((int) cmdBuffer[8]-48)*100;
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::protocolHandshake:" << "Receive Server RFB Protocol Version" << major << "." << minor;
#endif
    if (major != 3 && (minor != 3 || minor != 7 || minor != 8)) {
        //m_rfbSessionStatus = ERROR_WRONG_SERVER_VERSION;
        return;
    }
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRfbVersionMajor, 3);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigRfbVersionMajor, 3);
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRfbVersionMinor, minor);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigRfbVersionMinor, minor);
    m_rfbSessionStatus = RFB_SESSION_SECURITY_TYPES;
    switch (minor) {
    case 3:
        protocolVer.append("RFB 003.003", 11).append('\n');
        transmitData(12, protocolVer.data());
        break;
    case 7:        
        protocolVer.append("RFB 003.007", 11).append('\n');
        transmitData(12, protocolVer.data());
        break;
    case 8:        
        protocolVer.append("RFB 003.008", 11).append('\n');
        transmitData(12, protocolVer.data());
        break;
    default:
        //m_rfbSessionStatus = ERROR_WRONG_SERVER_VERSION;
        return;
    }
}


/**
 * RFB Security Handshake
 */
void QTmRfbClient::securityHandshake() {
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::securityHandshake()";
#endif
    unsigned char cmdBuffer[4];
    int receivedBytes = 0;
    if (m_rfbConfigServer->configuration(QTmConfiguration::ConfigRfbVersionMinor) == 3) { //RFB Server decides the security type
        qDebug() << "QTmRfbClient::1 waiting for data ..";
        receivedBytes = receiveData(4, (char*)cmdBuffer);
        if(receivedBytes < 1)
        {
            qDebug()<< "Error coccured while securityHandshake()";
            handleSocketError();
            return;
        }

        //check for security type 'None'
        if (cmdBuffer[0] == 0x00 && cmdBuffer[1] == 0x00 && cmdBuffer[2] == 0x00 && cmdBuffer[3] == 0x01) {
            m_rfbSessionStatus = RFB_SESSION_SERVER_INIT;
            //Send RFB Client Init message
            clientInit();
        }
        else
        {
            //m_rfbSessionStatus = ERROR_WRONG_SECURITY_TYPES;
            qDebug()<< "Error : invalid / unsupported security type";
        }
        return;
    }
    else { //Go through list to find if type 0x01 (None) is supported.
        qDebug() << "QTmRfbClient::2 waiting for data ..";
        receivedBytes = receiveData(1, (char*)cmdBuffer);
        if(receivedBytes < 1)
        {
            qDebug()<< "Error coccured while securityHandshake()";
            handleSocketError();
            return;
        }

        if(cmdBuffer[0]!= 0x00)
        {
            int noOfSecurityTypes = (int)cmdBuffer[0];
            for (int i=noOfSecurityTypes; i>0; i--)
            {
                receivedBytes = receiveData(1, (char*)cmdBuffer);
                if(receivedBytes < 1)
                {
                    qDebug()<< "Error occured while securityHandshake()";
                    handleSocketError();
                    return;
                }

                //check for security type 'None'
                if (cmdBuffer[0]==0x01) {
                    break;
                }
            }

            //check for security type 'None'
            if (cmdBuffer[0]==0x01)
            {
                qDebug() << "Got security type 'None'";
                transmitData(1, (char*)cmdBuffer);

                if(m_rfbConfigServer->configuration(QTmConfiguration::ConfigRfbVersionMinor) < 8)
                {
                    //3.7
                    m_rfbSessionStatus = RFB_SESSION_SERVER_INIT;
                    //Send RFB Client Init message
                    clientInit();
                }
                else
                {
                    //3.8
                    m_rfbSessionStatus = RFB_SESSION_SECURITY_RESPONSE;
                }
            }

        }
        else
        {
            //this means protocol version not suported by server
            if(m_rfbConfigServer->configuration(QTmConfiguration::ConfigRfbVersionMinor) >= 7)
            {
                //this means we need to read the fail reason
               readSecurityFailReason();
            }

            //now server closes the connection
            qDebug() << "Error : protocol version not supported by server!!";
        }
    }
}

void QTmRfbClient::securityResponse()                       //<<for ML cer
{
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::securityResponse()";
#endif
    unsigned char cmdBuffer[4];

    int receivedBytes = receiveData(4, (char*)cmdBuffer);
    if(receivedBytes < 4)
    {
        qDebug() << "Error occured while securityResponse()";
        handleSocketError();
        return;
    }

    if (cmdBuffer[0] == 0x00 && cmdBuffer[1] == 0x00 && cmdBuffer[2] == 0x00 && cmdBuffer[3] == 0x01)
    {
        //m_rfbSessionStatus = ERROR_WRONG_SECURITY_TYPES;

        //this means security result failed

        if(m_rfbConfigServer->configuration(QTmConfiguration::ConfigRfbVersionMinor) >= 8)
        {
            //this means we need to read the fail reason
            readSecurityFailReason();
        }

        m_rfbSessionStatus = RFB_SESSION_SERVER_VERSION;
    }
    else
    {
        m_rfbSessionStatus = RFB_SESSION_SERVER_INIT;

        //Send RFB Client Init message
        clientInit();
    }
}

void QTmRfbClient::readSecurityFailReason()
{
    unsigned char cmdBuffer[4];

    qDebug() << "Reading the security fail reason";

    //read the fail reason length
    int receivedBytes = receiveData(4, (char*)cmdBuffer);
    if(receivedBytes < 4)
    {
        qDebug() << "Error occured while securityResponse()";
        handleSocketError();
        return;
    }

    int reasonLength = ((cmdBuffer[0] << 24) | (cmdBuffer[1] << 16) | (cmdBuffer[2] << 8) | (cmdBuffer[3])) ;
    qDebug() << "reasonLength : " << reasonLength;
    //ignore the fail reason
    for(int i = 0; i < reasonLength; i++)
    {
        receiveData(1, (char*)cmdBuffer);
    }

    //now server closes the connection
    qDebug() << "Error : protocol version not supported by server!!";
    int connectStatus = -1;
    do
    {
        delete m_socket;

        qDebug() << "Try Reconnecting to server : " << m_url.host() << m_url.port();
        QThread().msleep(100);
        m_socket = new TCPSocket();
        connectStatus = m_socket->connectToHost(m_url.host().toLatin1().data(), m_url.port());

    }while(connectStatus);

    qDebug() << "Reconnected to server !!";
}


/**
 * RFB Client Server Initialization
 */
void QTmRfbClient::clientInit() {
    unsigned char cmdBuffer[1];
    cmdBuffer[0] = 0x00; //Sharing is set to false
    transmitData(1, (char*)cmdBuffer);
}

void QTmRfbClient::serverInit() {
#ifdef QT_DEBUG_TM_RFBCLIENT
        qDebug()<<"m_rfbConfigServer->serverInit()";
#endif
    unsigned char cmdBuffer[24];
    //RFB Server Init Message
    int receivedBytes = receiveData(24, (char *)cmdBuffer);
    if(receivedBytes < 24)
    {
        qDebug()<< "Error occured while serverInit()";
        handleSocketError();
        return;
    }

    //Framebuffer Size
    unsigned int width  = (cmdBuffer[0] << 8) | cmdBuffer[1];
    unsigned int height = (cmdBuffer[2] << 8) | cmdBuffer[3];

    qDebug()<<"m_rfbConfigServer->serverInit()  width -" << width;
    qDebug()<<"m_rfbConfigServer->serverInit()  height -" << height;

    //Color Format
    int bpp         = cmdBuffer[4];
    int depth       = cmdBuffer[5]; //Note: Ingnore ByteOrder at cmdBuffer[6]
    int true_color  = cmdBuffer[7];
    int red_max     = (cmdBuffer[8]  << 8) | cmdBuffer[9];
    int green_max   = (cmdBuffer[10] << 8) | cmdBuffer[11];
    int blue_max    = (cmdBuffer[12] << 8) | cmdBuffer[13];
    int red_shift   = cmdBuffer[14];
    int green_shift = cmdBuffer[15];
    int blue_shift  = cmdBuffer[16];
    if (true_color == 0)
    {
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::RGBNone);
#ifdef QT_CODE_DEBUG
        qDebug()<<"m_rfbConfigServer->setConfiguration()" <<QTmGlobal::RGBNone;
#endif
    }
    else if (bpp  == 32   && depth       == 24   &&
        red_max   == 0xFF && green_max   == 0xFF && blue_max   == 0xFF &&
        red_shift == 0x10 && green_shift == 0x08 && blue_shift == 0x00)
    {
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::RGB888);
#ifdef QT_CODE_DEBUG
        qDebug()<<"m_rfbConfigServer->setConfiguration()" <<QTmGlobal::RGB888;
#endif
    }
    else if (bpp  == 16   && depth       == 16   &&
        red_max   == 0x1F && green_max   == 0x3F && blue_max   == 0x1F &&
        red_shift == 0x0B && green_shift == 0x05 && blue_shift == 0x00)
    {
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::RGB565);
#ifdef QT_CODE_DEBUG
        qDebug()<<"m_rfbConfigServer->setConfiguration()" <<  QTmGlobal::RGB565;
#endif
    }
    else if (bpp  == 16   && depth       == 15   &&
        red_max   == 0x1F && green_max   == 0x1F && blue_max   == 0x1F &&
        red_shift == 0x0A && green_shift == 0x05 && blue_shift == 0x00)
    {
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::RGB555);
#ifdef QT_CODE_DEBUG
        qDebug()<<"m_rfbConfigServer->setConfiguration()" <<  QTmGlobal::RGB555;
#endif
    }
    else if (bpp  == 16   && depth       == 12   &&
        red_max   == 0x0F && green_max   == 0x0F && blue_max   == 0x0F &&
        red_shift == 0x08 && green_shift == 0x04 && blue_shift == 0x00)
    {
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::RGB444);
#ifdef QT_CODE_DEBUG
        qDebug()<<"m_rfbConfigServer->setConfiguration()" << QTmGlobal::RGB444;
#endif
    }
    else if (bpp  == 16   && depth       == 10   &&
        red_max   == 0x07 && green_max   == 0x0F && blue_max   == 0x07 &&
        red_shift == 0x07 && green_shift == 0x03 && blue_shift == 0x00)
    {
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::RGB343);
#ifdef QT_CODE_DEBUG
        qDebug()<<"m_rfbConfigServer->setConfiguration()" << QTmGlobal::RGB343;
#endif
    }
    else if (bpp  == 16   && depth       == 16   &&
        red_max   == 0xFF && green_max   == 0x00 && blue_max   == 0x00 &&
        red_shift == 0x00 && green_shift == 0x00 && blue_shift == 0x00)
    {
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::Gray16);
#ifdef QT_CODE_DEBUG
        qDebug()<<"m_rfbConfigServer->setConfiguration()" << QTmGlobal::Gray16;
#endif
    }
    else if (bpp  == 8    && depth       == 8    &&
        red_max   == 0x0F && green_max   == 0x00 && blue_max   == 0x00 &&
        red_shift == 0x00 && green_shift == 0x00 && blue_shift == 0x00)
    {
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::Gray8);
#ifdef QT_CODE_DEBUG
        qDebug()<<"m_rfbConfigServer->setConfiguration()" << QTmGlobal::Gray8;
#endif
    }
    else
    {
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::RGBNone);
#ifdef QT_CODE_DEBUG
        qDebug()<<"m_rfbConfigServer->setConfiguration()" << QTmGlobal::RGBNone;
#endif
    }
    //Client follows Server Color format, if no color format has been defined otherwise
    if (m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::RGBNone) == (int) QTmGlobal::RGBNone)
        m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigFbColorFormat, m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbColorFormat));
    //VNC Server name
    unsigned int length = (cmdBuffer[20] << 24) | (cmdBuffer[21] << 16) | (cmdBuffer[22] << 8) | cmdBuffer[23];
    unsigned char *serverName = (unsigned char *) malloc(length+1);
    receivedBytes = receiveData(length, (char *)serverName);
    if(receivedBytes < length)
    {
        qDebug()<< "Error occured server init";
        handleSocketError();
        return;
    }
    serverName[length] = '\0';
    //Create new image
    //desktopSizePseudoEncoding(QSize(width, height));
    setServerFbSize(QSize(width, height));
    //VNC Client is now fully connected
    m_rfbSessionStatus = QTmRfbClient::RFB_SESSION_CONNECTED_PASSIVE;
    //RFB SetEncodings messages
    setEncodings();
    setPixelFormat();
    qDebug() << "QTmRfbClient:" << "VNC Connected";
    emit sessionConnected(QString((char *)serverName));
#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
    ftime(&PerformanceStartTime);
#endif
}


/**
 * Send RFB SetEncoding message to the RFB server.
 */
//void QTmRfbClient::setEncodings(int encoding) {
void QTmRfbClient::setEncodings() {
#ifdef QT_DEBUG_TM_RFBCLIENT
    //qDebug() << "QTmRfbClient::setEncodings:" << "Send Set Encoding message" << encoding;
    //qDebug() << "QTmRfbClient::setEncodings:" << "sizeof(encoding)" << sizeof(encoding);
#endif
    unsigned char cmdBuffer[25] = {0};
    int data_enc = 0;
    int other_enc[6] = {0};
    int supportedEncCount = 0;

    if (m_rfbConfigClient->configuration(QTmConfiguration::ConfigRunLengthEncoding, false))
    {
        qDebug() << "QTmRfbClient::setEncodings:" << EXT_ENC_RUN_LENGTH_ENCODING;
        data_enc = EXT_ENC_RUN_LENGTH_ENCODING;
    }
    else
    {
        qDebug() << "QTmRfbClient::setEncodings:" << EXT_ENC_RUN_LENGTH_ENCODING;
        data_enc = EXT_ENC_RAW;
    }

    other_enc[supportedEncCount] = data_enc;
    supportedEncCount++;

    if (m_rfbConfigClient->configuration(QTmConfiguration::ConfigDesktopSizeEncoding, false))
    {
        qDebug() << "QTmRfbClient::setEncodings:" << EXT_ENC_DESKTOP_SIZE;
        other_enc[supportedEncCount] = EXT_ENC_DESKTOP_SIZE;
        supportedEncCount++;
    }

    if (m_rfbConfigClient->configuration(QTmConfiguration::ConfigContextInfoEncoding, false))
    {
        qDebug() << "QTmRfbClient::setEncodings:" << EXT_ENC_TM_CONTEXT_INFO;
        other_enc[supportedEncCount] = EXT_ENC_TM_CONTEXT_INFO;
        supportedEncCount++;
    }

    if (m_rfbConfigClient->configuration(QTmConfiguration::ConfigTerminalModeEncoding, false))
    {
        qDebug() << "QTmRfbClient::setEncodings:" << EXT_ENC_TERMINAL_MODE;
        other_enc[supportedEncCount] = EXT_ENC_TERMINAL_MODE;
        supportedEncCount++;
    }

    cmdBuffer[0] = MSG_SET_ENCODINGS;
    cmdBuffer[1] = 0x00;
    cmdBuffer[2] = 0x00;
    cmdBuffer[3] = (supportedEncCount);
    int i,j;
    for(i=4, j=0; j < supportedEncCount; i+=4,j++)
    {
        cmdBuffer[i+3] = other_enc[j] & 0xFF;
        cmdBuffer[i+2] = (other_enc[j] & 0xFF00) >> 8;
        cmdBuffer[i+1] = (other_enc[j] & 0xFF0000) >> 16;
        cmdBuffer[i] = (other_enc[j] & 0xFF000000) >> 24;
    }

    transmitData((4 + (sizeof(int) * supportedEncCount)), (char*)cmdBuffer);
}


/**
 * Send RFB SetPixelFormat message to the RFB server.
 */
void QTmRfbClient::setPixelFormat() {
    unsigned char cmdBuffer[20];
    bzero(cmdBuffer, 20);
    cmdBuffer[0]  = MSG_SET_PIXEL_FORMAT;
    cmdBuffer[1]  = 0;
    cmdBuffer[2]  = 0;
    cmdBuffer[3]  = 0;
   // cmdBuffer[6]  = 0x01; //Byte order flag (big endian)
    cmdBuffer[6]  = 0x00; //Byte order flag (little endian)
    cmdBuffer[7]  = 0x01; //True color flag
    QTmGlobal::TmColorFormat color = (QTmGlobal::TmColorFormat) m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::RGB888);
#ifdef QT_CODE_DEBUG
    qDebug()<<"QTmRfbClient::setPixelFormat() << COLOR FORMAT "<<color;
#endif
    switch (color) {
    case QTmGlobal::RGB565:
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::setPixelFormat:" << "Set Color Format" << "RGB565";
#endif
        cmdBuffer[4]  = 16;   //BPP
        cmdBuffer[5]  = 16;   //Color Depth
        cmdBuffer[9]  = 0x1F; //Red Max (LSB) 2 raiseto 5
        cmdBuffer[11] = 0x3F; //Green Max (LSB) 2 raiseto 6
        cmdBuffer[13] = 0x1F; //Blue Max (LSB) 2 raiseto 5
        cmdBuffer[14] = 0x0B; //Red Shift
        cmdBuffer[15] = 0x05; //Green Shift
        cmdBuffer[16] = 0x00; //Blue Shift
        break;
    case QTmGlobal::RGB555:
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::setPixelFormat:" << "Set Color Format" << "RGB555";
#endif
        cmdBuffer[4]  = 16;
        cmdBuffer[5]  = 15;
        cmdBuffer[9]  = 0x1F;
        cmdBuffer[11] = 0x1F;
        cmdBuffer[13] = 0x1F;
        cmdBuffer[14] = 0x0A;
        cmdBuffer[15] = 0x05;
        cmdBuffer[16] = 0x00;
        break;
    case QTmGlobal::RGB444:
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::setPixelFormat:" << "Set Color Format" << "RGB444";
#endif
        cmdBuffer[4]  = 16;
        cmdBuffer[5]  = 12;
        cmdBuffer[9]  = 0x0F;
        cmdBuffer[11] = 0x0F;
        cmdBuffer[13] = 0x0F;
        cmdBuffer[14] = 0x08;
        cmdBuffer[15] = 0x04;
        cmdBuffer[16] = 0x00;
        break;
    case QTmGlobal::RGB343:
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::setPixelFormat:" << "Set Color Format" << "RGB343";
#endif
        cmdBuffer[4]  = 16;
        cmdBuffer[5]  = 10;
        cmdBuffer[9]  = 0x07;
        cmdBuffer[11] = 0x0F;
        cmdBuffer[13] = 0x07;
        cmdBuffer[14] = 0x07;
        cmdBuffer[15] = 0x03;
        cmdBuffer[16] = 0x00;
        break;
    case QTmGlobal::RGB888:
    default:
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::setPixelFormat:" << "Set Color Format" << "RGB888";
#endif
        cmdBuffer[4]  = 32;
        cmdBuffer[5]  = 24;
        cmdBuffer[9]  = 0xFF;
        cmdBuffer[11] = 0xFF;
        cmdBuffer[13] = 0xFF;
        cmdBuffer[14] = 0x10;
        cmdBuffer[15] = 0x08;
        cmdBuffer[16] = 0x00;
        break;
    }
    transmitData(20, (char*)cmdBuffer);
}


/*!
  Send an RFB Key Event message, as given in \a event.
  If the event is key-press event, the key event counter is set to zero.
  The \a event is not deleted.
 */
void QTmRfbClient::keyEvent(QTmKeyEvent *event) {
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::keyEvent:" << "Send key event" << event->key() << event->pressed();
#endif
    unsigned char cmdBuffer[8];
    cmdBuffer[0] = MSG_KEY_EVENT;
    cmdBuffer[1] = event->pressed();
    cmdBuffer[2] = 0x00;
    cmdBuffer[3] = 0x00;
    cmdBuffer[4] = event->key() >> 24;
    cmdBuffer[5] = event->key() >> 16;
    cmdBuffer[6] = event->key() >>  8;
    cmdBuffer[7] = event->key();
    //Increase key event counter on press events only
    if (event->pressed())
        m_keyEventCounter++;
    transmitData(8, (char*)cmdBuffer);
}
/*!
  Send an RFB Pointer Event message, as given in \a event.
  The \a event is not deleted.
 */
#ifdef QT_UI_CLEANUP
void QTmRfbClient::pointerEvent(int x, int y, unsigned int pointer_pressed) {
#ifdef QT_CODE_DEBUG
    qDebug()<<"QTmRfbClient::pointerEvent() "<< x << "," << y <<", pressed :" <<  pointer_pressed;
#endif
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::pointerEvent:" << "Send pointer event" << x << "," << y;
#endif
    unsigned char cmdBuffer[6];
    qDebug()<<"QTmRfbClient::pointerEvent() @@@@";
    cmdBuffer[0] = MSG_POINTER_EVENT;
    cmdBuffer[1] = pointer_pressed;
    cmdBuffer[2] = x >> 8;
    cmdBuffer[3] = x;
    cmdBuffer[4] = y >> 8;
    cmdBuffer[5] = y;
    transmitData(6, (char *) cmdBuffer);
}
#else
void QTmRfbClient::pointerEvent(QTmPointerEvent *event) {
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::pointerEvent:" << "Send pointer event" << event->point();
#endif
    unsigned char cmdBuffer[6];
    cmdBuffer[0] = MSG_POINTER_EVENT;
    cmdBuffer[1] = event->pressed();
    cmdBuffer[2] = event->point().x() >> 8;
    cmdBuffer[3] = event->point().x();
    cmdBuffer[4] = event->point().y() >> 8;
    cmdBuffer[5] = event->point().y();
    transmitData(6, (char *) cmdBuffer);
}
#endif
/*!
  Send an RFB Touch Event message, as given in \a touchEvent. This message is a Terminal Mode VNC extension message.
  This message sends all simultaneous individual events within a single message.
  These individual events are removed from the QList, while sending. \a touchEvent itself is not deleted though.
 */
void QTmRfbClient::touchEvent(QList<QTmTouchEvent *> *touchEvent) {
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::touchEvent:" << "Send touch event" << touchEvent->size();
#endif
    int n = touchEvent->size() & 0xFF;
    int j = 5;
    QByteArray buffer(5+n*6, 0);
    buffer[0] = EXT_MSG_TERMINAL_MODE;
    buffer[1] = TM_MSG_TOUCH_EVENT;
    buffer[2] = (1+n*6) >> 8;
    buffer[3] = (1+n*6);
    buffer[4] = n;
    foreach(QTmTouchEvent *event, *touchEvent) {
        buffer[j++] = event->point().x() >> 8;
        buffer[j++] = event->point().x();
        buffer[j++] = event->point().y() >> 8;
        buffer[j++] = event->point().y();
        buffer[j++] = event->eventId();
        buffer[j++] = event->pressure();
    }
    transmitData(5+n*6, (char*)buffer.constData());
}

/*!
  Send an RFB ClientCutText message. The message to be sent is given in \a text.
  The text is translated into Latin1. Other character sets are not supported.
 */
void QTmRfbClient::clientCutText(QString text) {
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::clientCutText:" << "Send Client Touch Text" << text;
#endif
    unsigned long int length = text.size();
    unsigned char *cmdBuffer = (unsigned char *) malloc(8+length);
    cmdBuffer[0] = MSG_CLIENT_CUT_TEXT;
    cmdBuffer[1] = 0x00;
    cmdBuffer[2] = 0x00;
    cmdBuffer[3] = 0x00;
    cmdBuffer[4] = length >> 24;
    cmdBuffer[5] = length >> 16;
    cmdBuffer[6] = length >>  8;
    cmdBuffer[7] = length;
    memcpy(cmdBuffer+8, text.toLatin1(), length);
    transmitData(8+length, (char *) cmdBuffer);
}
#ifdef QT_UI_CLEANUP
void QTmRfbClient::serverCutText() {
    unsigned char cmdBuffer[8];
#if 0
    int receivedBytes = receiveData(7, (char*)cmdBuffer+1);
    if(receivedBytes < 7)
    {
        qDebug()<<"Error occured while serverCutText";
        handleSocketError();
        return;
    }
    int length = (cmdBuffer[4] << 24) | (cmdBuffer[5] << 16) | (cmdBuffer[6] << 8) | (cmdBuffer[7]);
    char *serverText = (char *) malloc(length + 1);
    receivedBytes = receiveData(length, serverText);
    if(receivedBytes < length)
    {
        qDebug()<<"Error occured while serverCutText";
        handleSocketError();
        return;
    }
    serverText[length] = '\0';
    QString copyText(serverText);
    /** emit signal as needed*/
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::serverCutText:" << "Receive Client Touch Text" << serverText;
#endif
    free(serverText);
#else                           //<<for ML cer
    if(m_socket){
       m_socket->readAll();
    }
#endif
}

#else
void QTmRfbClient::serverCutText() {
    unsigned char cmdBuffer[8];
    receiveData(7, (char*)cmdBuffer+1);
    int length = (cmdBuffer[4] << 24) | (cmdBuffer[5] << 16) | (cmdBuffer[6] << 8) | (cmdBuffer[7]);
    char *serverText = (char *) malloc(length + 1);
    receiveData(length, serverText);
    serverText[length] = '\0';
    QString copyText(serverText);
    QClipboard *clipboard = QApplication::clipboard();
    clipboard->setText(copyText, QClipboard::Clipboard);
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::serverCutText:" << "Receive Client Touch Text" << serverText;
#endif
    free(serverText);
}
#endif



/**
 * Receive RFB SetColourMapEntries message from the RFB server.
 */
void QTmRfbClient::setColourMapEntries() {
    unsigned char cmdBuffer[6];
#if 0
    int receivedBytes = receiveData(5, (char*)cmdBuffer+1);
    if(receivedBytes < 5)
    {
        qDebug() << "Error occured while setColourMapEntries";
        handleSocketError();
        return;
    }
    int n = (cmdBuffer[4] << 8) | (cmdBuffer[5]);
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient:" << "Received " << n << "colours. NOTE: Colour Maps are not supported. Ignore colours";
#endif
    for (int i=n; i>0; i++)
    {
        receivedBytes = receiveData(6, (char*)cmdBuffer);
        if(receivedBytes < 6)
        {
            qDebug() << "Error occured while setColourMapEntries";
            handleSocketError();
            break;
        }
    }
#else
    m_socket->readAll();
#endif
}


/**
 * Receive RFB Bell message from the RFB server and call system beep function.
 */

#ifdef QT_UI_CLEANUP
void QTmRfbClient::bell() {
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient:" << "BELL";
#endif
    if(m_socket){
        m_socket->readAll();                       //<<for ML cer
    }
}
#else
void QTmRfbClient::bell() {
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient:" << "BELL";
#endif
    QApplication::beep();
}
#endif



/*!
  Send an RFB Framebuffer Update Request message to the RFB server.
  The client will request a non-incremental update for the entire framebuffer region.
  The message is only send, when framebuffer updates have previously been disabled.
 */
void QTmRfbClient::triggerFramebufferUpdates() {
    if (m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbUpdates, false))
    {
#ifndef BUFFER_MANAGER_INTEGRATED
        framebufferUpdateRequest(false);
#endif
    }
}

void QTmRfbClient::framebufferUpdateRequest(bool incremental) {
    double currTimeMsec;
    double fbUpdateRequestTimeDiff;

    if (!m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbUpdates, false))
        return;
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::framebufferUpdateRequest:" << "Send Framebuffer Update Request message"
             << QRect(0, 0,
                      m_rfbConfigClient->configuration(QTmConfiguration::ConfigRequestedFbWidth, 0),
                      m_rfbConfigClient->configuration(QTmConfiguration::ConfigRequestedFbHeight, 0))
             << incremental;
#endif

    if(m_timeLastFBRequest == 0)
    {
        //this means, this is 1st FB update request
        sendFbUpdateRequest(incremental);

        m_timeLastFBRequest = getCurrMillisec();

        //qDebug() << "m_timeLastFBRequest : " << QString::number(m_timeLastFBRequest,'f',0);
    }
    else
    {
        //this means, this is NOT 1st FB update request

        //qDebug() << "m_timeLastFBRequest : " << QString::number(m_timeLastFBRequest,'f',0);

        currTimeMsec = getCurrMillisec();

        //qDebug() << "currTimeMsec : " << QString::number(currTimeMsec,'f',0);

        if(currTimeMsec > m_timeLastFBRequest)
        {
            fbUpdateRequestTimeDiff = (currTimeMsec - m_timeLastFBRequest);

            m_timeRoundTrip += fbUpdateRequestTimeDiff;
            m_frameCount ++;
            if(m_frameCount == 10)
            {
                qDebug() << "avg round trip msec: " <<  m_timeRoundTrip/m_frameCount;
                m_frameCount = 0;
                m_timeRoundTrip = 0;
            }

            //qDebug() << "fbUpdateRequestTimeDiff : " << QString::number(fbUpdateRequestTimeDiff,'f',0);

            if(fbUpdateRequestTimeDiff >= MIN_FB_UPDATE_REQUEST_INTERVAL)
            {
                sendFbUpdateRequest(incremental);

                m_timeLastFBRequest = getCurrMillisec();
            }
            else
            {
                //this means, the FB update msg from TM server are too fast.
                //so need to schedule FB update request certain msec later
                //which will eventually control the FB update rate from TM server.
                //this regulation helps in below way,
                //1. gui module gets sufficient time to render(i.e. bitmap creation, scaling, padding etc.)
                //2. keep the CPU load moderate
                //NOTE: This static regulation is temporary.

                double delayMsec = (MIN_FB_UPDATE_REQUEST_INTERVAL - fbUpdateRequestTimeDiff);

    #ifdef QT_DEBUG_TM_RFBCLIENT
                qDebug() << "QTmRfbClient::framebufferUpdateRequest: sleep msec" << QString::number(delayMsec,'f',0);
    #endif

                msleep(delayMsec);

                //since this is not 1st FB update request
                //here 'incremental' flag should be 'true' if incremental updates are enabled
                sendFbUpdateRequest(clientConfig(QTmConfiguration::ConfigIncrementalUpdate, false));

                m_timeLastFBRequest = getCurrMillisec();
            }
        }
        else
        {
            qDebug() << "framebufferUpdateRequest() : system time changed in -ve ";
        }
    }
}

void QTmRfbClient::resetFbUpdateRequestTimer()
{
    qDebug() << "QTmRfbClient::resetFbUpdateRequestTimer():";
    m_fbUpdateRequestFlag = false;
#ifndef BUFFER_MANAGER_INTEGRATED
    sendFbUpdateRequest(false);
#endif
    m_timeLastFBRequest = getCurrMillisec();
}

void QTmRfbClient::pauseFbUpdateRequest()
{
    qDebug() << "QTmRfbClient::pauseFbUpdateRequest():";
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigFbUpdates, false);
    
    if(m_surface)
    {
        m_surface->resetSurface();
    }
    else
    {
        qDebug() << "QTmRfbClient::m_surface is null !!";
    }
}

void QTmRfbClient::resumeFbUpdateRequest()
{
    m_isSendingFbUpdateRequestFirstTime = true;

    qDebug() << "QTmRfbClient::resumeFbUpdateRequest():";
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigFbUpdates, true);

    m_fbUpdateRequestFlag = false;

    if(m_surface)
    {
	    m_surface->resetSurface();
        m_surface->registerObserver(this);
        m_surface->start();
    }
    else
    {
        qDebug() << "QTmRfbClient::m_surface is null !!";
    }
}

void QTmRfbClient::framebufferUpdate() {

    bool updateSurfaceFlag = true;
    unsigned char cmdBuffer[12];
    unsigned int i;
    unsigned int nRectangle;
#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
    bool PERFORMANCE_sample = false;
#endif
#ifdef QT_DEBUG_TM_RFBCLIENT
        //qDebug() << "QTmRfbClient::framebufferUpdate:" << "Receive Framebuffer Update message";
#endif
    int totalReceivedBytes = 0;
    int rectByteSize = 0;
    int receivedBytes = receiveData(3, (char *)cmdBuffer+1);

    if(receivedBytes < 3)
    {
        qDebug() << "Error occured while framebufferUpdate";
        handleSocketError();
        return;
    }

    //check whether FB update request was sent for this FB update message,
    //if 'yes' then
    //          surfaceUpdate should be done.
    //if 'no' then
    //          surfaceUpdate should NOT be done, since this violates the ML specs.
    //          also 'FB update request' wont be sent for this invalid FB update msg.
    if(m_fbUpdateRequestFlag == true)
    {
        m_fbUpdateRequestFlag = false;
        updateSurfaceFlag = true;
    }
    else
    {
        updateSurfaceFlag = false;
    }


    totalReceivedBytes += receivedBytes;
    nRectangle = (cmdBuffer[2] << 8) | cmdBuffer[3];
    for (i=nRectangle; i>0; i--) {
        receivedBytes = receiveData(12, (char*)cmdBuffer);
        if(receivedBytes < 12)
        {
            qDebug() << "Error occured while framebufferUpdate";
            handleSocketError();
            break;
        }
        totalReceivedBytes += receivedBytes;
        unsigned int x      = (cmdBuffer[0] << 8)  | cmdBuffer[1];
        unsigned int y      = (cmdBuffer[2] << 8)  | cmdBuffer[3];
        unsigned int width  = (cmdBuffer[4] << 8)  | cmdBuffer[5];
        unsigned int height = (cmdBuffer[6] << 8)  | cmdBuffer[7];
        int          enc    = (cmdBuffer[8] << 24) | (cmdBuffer[9] << 16) | (cmdBuffer[10] << 8) | cmdBuffer[11];

        qDebug() << "QTmRfbClient::framebufferUpdate:" << "encoding type - " << enc;
#ifdef QT_DEBUG_TM_RFBCLIENT
        qDebug() << "QTmRfbClient::framebufferUpdate:" << "Rectangle" << i << QRect(x, y, width, height);
#endif
        switch (enc) {
        case EXT_ENC_RAW:
            if (width == 0 || height == 0)
                break;
#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
            PERFORMANCE_number_of_pixel += height * width;
            PERFORMANCE_number_of_rectangles++;
            PERFORMANCE_sample = true;
#endif
            rectByteSize  = rawEncoding(x, y, width, height);
            //qDebug()<< "----------------------------------    rectByteSize ::::::: "<<rectByteSize;
            totalReceivedBytes+= rectByteSize;

            if ((i == 1) && (updateSurfaceFlag == true))
            {
#ifndef BUFFER_MANAGER_INTEGRATED
                framebufferUpdateRequest(m_rfbConfigClient->configuration(QTmConfiguration::ConfigIncrementalUpdate, true));
#endif
            }

            //Update the graphical user interface
            if (!m_surface)
                return;

#ifndef VNC_OPTIMIZATION
            if(updateSurfaceFlag == true)
#else
            if ((i == 1) && (updateSurfaceFlag == true))
#endif
               m_surface->updateSurface(QRect(x, y, width, height));

            break;
        case EXT_ENC_RUN_LENGTH_ENCODING:
            if (width == 0 || height == 0)
                break;
#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
            PERFORMANCE_number_of_rectangles++;
            PERFORMANCE_number_of_pixel += height * width;
            PERFORMANCE_number_of_runs  += rleEncoding(x, y, width, height);
            PERFORMANCE_sample = true;
#else
            totalReceivedBytes += rleEncoding(x, y, width, height);
#endif
            if ((i == 1) && (updateSurfaceFlag == true))
            {
#ifndef BUFFER_MANAGER_INTEGRATED
               framebufferUpdateRequest(m_rfbConfigClient->configuration(QTmConfiguration::ConfigIncrementalUpdate, true));
#endif
            }

            //Update the graphical user interface
            if (!m_surface)
                return;

#ifndef VNC_OPTIMIZATION
            if(updateSurfaceFlag == true)
#else
            if ((i == 1) && (updateSurfaceFlag == true))
#endif
                m_surface->updateSurface(QRect(x, y, width, height));
            break;
        case EXT_ENC_DESKTOP_SIZE:
#ifndef BUFFER_MANAGER_INTEGRATED
            if ((i == 1) && (updateSurfaceFlag == true))
            {
                framebufferUpdateRequest(false);
            }
#endif

            desktopSizePseudoEncoding(QSize ((cmdBuffer[4] << 8) | cmdBuffer[5], (cmdBuffer[6] << 8) | cmdBuffer[7]));

#ifdef BUFFER_MANAGER_INTEGRATED
            if ((i == 1) && (updateSurfaceFlag == true))
            {
                m_surface->updateSurface(QRect(0, 0, (cmdBuffer[4] << 8) | cmdBuffer[5], (cmdBuffer[6] << 8) | cmdBuffer[7]));
            }
#endif
            break;
        case EXT_ENC_TM_CONTEXT_INFO:
            totalReceivedBytes += contextInformationPseudoEncoding(cmdBuffer);
            if ((i == 1) && (updateSurfaceFlag == true))
            {
#ifndef BUFFER_MANAGER_INTEGRATED
                framebufferUpdateRequest(m_rfbConfigClient->configuration(QTmConfiguration::ConfigIncrementalUpdate, true));
#endif
            }
            break;
        default:
            //m_rfbSessionStatus = ERROR_WRONG_FRAMEBUFFER_ENCODING;
            qDebug() << "ERROR" << "Unknown Framebuffer Encoding";
            if(m_socket){
                m_socket->readAll();
            }
            break;
        }
    }
#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
    if (!PERFORMANCE_sample)
        return;
    PERFORMANCE_counter++;
    if ((PERFORMANCE_counter % PerformanceSampleWindow) == 0)
    {
        ftime(&PerformanceEndTime);
        PerformanceDiffTime = (1000.0*PerformanceEndTime.time   + PerformanceEndTime.millitm -
                               1000.0*PerformanceStartTime.time - PerformanceStartTime.millitm)/
                PerformanceSampleWindow;
        PerformanceFpsCount = 1000.0/PerformanceDiffTime;
        qDebug("***** \nPerformance: After %i(th) FrameBufferUpdates", PERFORMANCE_counter);
        qDebug("***** - Achieved framerate/s   %.2f", PerformanceFpsCount);
        qDebug("***** - Number of Pixels       %i",   PERFORMANCE_number_of_pixel);
        qDebug("***** - Average rectangle size %.2f", (double) PERFORMANCE_number_of_pixel / (double) PERFORMANCE_number_of_rectangles);
        qDebug("***** - Average Run Length     %.2f", (double) PERFORMANCE_number_of_pixel / (double) PERFORMANCE_number_of_runs);
        PERFORMANCE_number_of_pixel = 0;
        PERFORMANCE_number_of_rectangles = 0;
        PERFORMANCE_number_of_runs = 0;
        ftime(&PerformanceStartTime);
    }

#else
    qDebug() << "QTmRfbClient::framebufferUpdate() :   " << "bytes rcvd :" << totalReceivedBytes;
#endif

    //qDebug()<<"QTmRfbClient::framebufferUpdateRequest << TOTAL REQUEST DUARATION TIMESTAMP::" << tsResponseDuration << ", totalbytes :" << totalReceivedBytes;
}


// ********************************************************************
// *****              Display Configuration message               *****
// ********************************************************************
/*!
  \fn QTmRfbClient::displayConfigurationReceived()
  Signal emitted from a RFB client to inform about an server display configuration message received.
  The receiving component should use this information to decide on the pixel format to request.
 */
/*!
  Receive an RFB Server Display Configuration message. This message is a Terminal Mode VNC extension message.
  The server display configuration message provides information about the display properties at the VNC server.
  This includes the following elements:
  \list
    \o Termina Mode Server version (major and minor)
    \o Supported framebuffer configurations (orientation, rotaton, scaling, replacement
    \o Relative pixel width
    \o Supported color formats
  \endlist
  The received information is stored in the server configuration isnstace. After emitting the
  displayConfigurationReceived() signal, the client sends its own Client Display Configuration message automatically.
  Returns the number of bytes read.
 */
int QTmRfbClient::serverDisplayConfig() {
    unsigned char cmdBuffer[12];
    int read = receiveData(12, (char*)cmdBuffer);
    if (read < 12) {
        qDebug() << "QRfbClientTm:" << "Error receiving Server Display Configuration";
        handleSocketError();
        return read;
    }

    if(!m_displayStatusFlag){/*START ML Certification issue: server display config issue*/
    //Terminal Mode version
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigTmVersionMajor, ((int) cmdBuffer[0])-48);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigTmVersionMinor, ((int) cmdBuffer[1])-48);
    //Framebuffer Configuration
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbOrientation,  (cmdBuffer[3]  >> 0) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbRotation,     (cmdBuffer[3]  >> 1) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbUpScaling,    (cmdBuffer[3]  >> 2) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbDownScaling,  (cmdBuffer[3]  >> 3) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbReplacement,  (cmdBuffer[3]  >> 4) & 0x01);
    //Relative Pixel Size
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigPixelWidth,     (cmdBuffer[4]  << 8) | cmdBuffer[5]);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigPixelHeight,    (cmdBuffer[6]  << 8) | cmdBuffer[7]);
    //Pixel Format support
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigColorARGB888,   (cmdBuffer[11] >> 0) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigColorRGB888,    (cmdBuffer[10] >> 0) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigColorRGB565,    (cmdBuffer[9]  >> 0) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigColorRGB555,    (cmdBuffer[9]  >> 1) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigColorRGB444,    (cmdBuffer[9]  >> 1) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigColorRGB343,    (cmdBuffer[9]  >> 1) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigColorGray16,    (cmdBuffer[8]  >> 0) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigColorGray8,     (cmdBuffer[8]  >> 1) & 0x01);
    //Emit signal on message reception
    emit displayConfigurationReceived();
    //Send response message
     clientDisplayConfig();
     //setPixelFormat();
    m_displayStatusFlag = true;
    return read;
    }
}
/*!
  Sends an RFB Client Display Configuration message. This message is a Terminal Mode VNC extension message.
  The client display configuration message provides information about the display properties at the VNC Client.
  This includes the following elements:
  \list
    \o Termina Mode Server version (major and minor)
    \o Supported framebuffer configurations (orientation, rotaton, scaling, replacement
    \o Screen resolution
    \o Phyiscal screen dimensions and distance from the driver
  \endlist
  The transmitted information is taken from the client configuration isnstace.
 */
void QTmRfbClient::clientDisplayConfig() {
    unsigned char cmdBuffer[18];
    QSize clientFrameBufferSize;
    unsigned int width, height;
    QTmConfiguration::ServerScalingStatus serverScalingStatus;

    cmdBuffer[0]  = EXT_MSG_TERMINAL_MODE;
    cmdBuffer[1]  = TM_MSG_CLIENT_DISPLAY_CONFIG;
    cmdBuffer[2]  = 0x00;
    cmdBuffer[3]  = 0x0E;
    cmdBuffer[4]  = 0x01; //Terminal Mode Major version is 1
    cmdBuffer[5]  = 0x00; //Terminal Mode Minor version is 0

    //decide the framebuffer size
    getExpectedClientSize(QSize(m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbWidth,  0), m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbHeight,  0)),
                          &clientFrameBufferSize, false);

    //Framebuffer Configuration
    cmdBuffer[6]  = 0x00;
    cmdBuffer[7]  = ((m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbOrientation,  true) & 0x01) << 0) |
                    ((m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbRotation,    false) & 0x01) << 1) |
                    ((m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbUpScaling,    true) & 0x01) << 2) |
                    ((m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbDownScaling,  true) & 0x01) << 3) |
                    ((m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbReplacement, false) & 0x01) << 4);

    cmdBuffer[8]  = clientFrameBufferSize.width() >> 8;
    cmdBuffer[9]  = clientFrameBufferSize.width();
    cmdBuffer[10] = clientFrameBufferSize.height() >> 8;
    cmdBuffer[11] = clientFrameBufferSize.height();

    //m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigDisplayPixelWidth, clientFrameBufferSize.width());
    //m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigDisplayPixelHeight, clientFrameBufferSize.height());
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbWidth, clientFrameBufferSize.width());
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbHeight, clientFrameBufferSize.height());
    //clientFrameBufferSize.setHeight(height);
    //clientFrameBufferSize.setWidth(width);
    QTmGlobal::TmColorFormat colorFormat = (QTmGlobal::TmColorFormat) m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbColorFormat);


    //Client display sizes
    cmdBuffer[12] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplaySizeWidth,   0) >> 8;
    cmdBuffer[13] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplaySizeWidth,   0);
    cmdBuffer[14] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplaySizeHeight,  0) >> 8;
    cmdBuffer[15] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplaySizeHeight,  0);
    cmdBuffer[16] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayDistance,    0) >> 8;
    cmdBuffer[17] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayDistance,    0);
    transmitData(18, (char*)cmdBuffer);

#if 0
    //allocate the framebuffer
    if(m_surface)
    {
        m_framebuffer = m_surface->allocateSurface(clientFrameBufferSize, colorFormat, m_calculatedXScaleFactor, m_calculatedYScaleFactor);
        if(!m_framebuffer)
        {
            qDebug() << "QTmRfbClient::clientDisplayConfig:" << "Allocate surface failed!";
            //m_rfbSessionStatus = QTmRfbClient::ERROR_CANNOT_ALLOCATE_SURFACE;
            //emit sessionDisconnected();
            return;
        }
#ifdef BUFFER_MANAGER_INTEGRATED
		if(m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbUpdates, false))
        m_surface->start();
#endif
    }
    else
    {
        qDebug() << "QTmRfbClient::clientDisplayConfig:" << "No surface registered with rfb client !!";
    }

#endif
}

void QTmRfbClient::getExpectedClientSize(QSize serverSize, QSize * clientFrameBufferSize, bool afterFbResize)
{
    int width, height;
    QTmConfiguration::ServerScalingStatus serverScalingStatus;

    //int serverDisplayPixelWidth = m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbWidth,  0);
    //int serverDisplayPixelHeight = m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbHeight,  0);

    int serverDisplayPixelWidth = serverSize.width();
    int serverDisplayPixelHeight = serverSize.height();

    int clientDisplayPixelWidth = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelWidth,  0);
    int clientDisplayPixelHeight = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelHeight,  0);

    calculateScalingFactors(clientDisplayPixelWidth, clientDisplayPixelHeight, serverDisplayPixelWidth, serverDisplayPixelHeight);

    qDebug() << "QTmRfbClient::" << "serverDisplayPixelWidth : " << serverDisplayPixelWidth;
    qDebug() << "QTmRfbClient::" << "serverDisplayPixelHeight : " << serverDisplayPixelHeight;
    qDebug() << "QTmRfbClient::" << "clientDisplayPixelWidth : " << clientDisplayPixelWidth;
    qDebug() << "QTmRfbClient::" << "clientDisplayPixelHeight : " << clientDisplayPixelHeight;

    if(!afterFbResize)
    {
        if((m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbUpScaling, 0)) ||
            (m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbDownScaling, 0)))
        {
            //server side scaling supported

            double serverDisplayAspectRatio = (double)((double)serverDisplayPixelWidth / (double)serverDisplayPixelHeight);

            qDebug() << "QTmRfbClient::" << "serverDisplayAspectRatio : " << serverDisplayAspectRatio;

            double clientDisplayAspectRatio = ((double)clientDisplayPixelWidth / (double)clientDisplayPixelHeight);

            qDebug() << "QTmRfbClient::" << "clientDisplayAspectRatio - " << clientDisplayAspectRatio;

            //get the scaling support status
            if(m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbUpScaling, 0) &&
                    !(m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbDownScaling, 0)))
            {
                qDebug() << "QTmRfbClient::clientDisplayConfig:" << "server side up scaling supported";
                serverScalingStatus = QTmConfiguration::onlyUpScaling;
            }
            if(!(m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbUpScaling, 0)) &&
                    (m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbDownScaling, 0)))
            {
                qDebug() << "QTmRfbClient::clientDisplayConfig:" << "server side down scaling supported";
                serverScalingStatus = QTmConfiguration::onlyDownScaling;

                // So client will not do downScaling (this saves bandwidth !!)
                m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigFbDownScaling, false);

            }
            if(m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbUpScaling, 0) &&
                    m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbDownScaling, 0))
            {
                qDebug() << "QTmRfbClient::clientDisplayConfig:" << "server side both scaling supported";
                serverScalingStatus = QTmConfiguration::bothUpDownScaling;

                // So client will not do downScaling (this saves bandwidth !!)
                m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigFbDownScaling, false);
            }

            //check whether client needs upscaling or downscaling
            if((clientDisplayAspectRatio > serverDisplayAspectRatio))
            {
                //client screen is wider
                qDebug() << "QTmRfbClient::" << "client screen is wider";
                if((serverScalingStatus == QTmConfiguration::onlyUpScaling) || (serverScalingStatus == QTmConfiguration::bothUpDownScaling))
                {
                    width = (serverDisplayAspectRatio*clientDisplayPixelHeight);
                    height = clientDisplayPixelHeight;

                    //recalculate the scaling factors
                    calculateScalingFactors(m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelWidth,  0),
                                            m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelHeight,  0),
                                            width, height);
                }
                else
                {
                    //server side scaling NOT supported
                    width = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelWidth,  0);
                    height = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelHeight,  0);
                }
            }
            else if(clientDisplayAspectRatio < serverDisplayAspectRatio)
            {
                //client screen is narrower
                qDebug() << "QTmRfbClient::" << "client screen is narrower";
                if((serverScalingStatus == QTmConfiguration::onlyDownScaling) || (serverScalingStatus == QTmConfiguration::bothUpDownScaling))
                {
                    width = clientDisplayPixelWidth;
                    height = (unsigned int)((double)(1/serverDisplayAspectRatio)*(double)clientDisplayPixelWidth);

                    //recalculate the scaling factors
                    calculateScalingFactors(m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelWidth,  0),
                                            m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelHeight,  0),
                                            width, height);
                }
                else
                {
                    //server side scaling NOT supported
                    width = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelWidth,  0);
                    height = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelHeight,  0);
                }
            }
            else
            {
                //client screen has same aspect ratio as server
                qDebug() << "QTmRfbClient::" << "client screen has same aspect ratio as server";
                width = clientDisplayPixelWidth;
                height = clientDisplayPixelHeight;

                //recalculate the scaling factors
                calculateScalingFactors(m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelWidth,  0),
                                        m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelHeight,  0),
                                        width, height);
            }
        }
        else if(!(m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbUpScaling, 0)) &&
            !(m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbDownScaling, 0)))
        {
            //server side scaling NOT supported
            //width = m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbWidth,  0);
            //height = m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbHeight,  0);
            width = serverSize.width();
            height = serverSize.height();
        }

    }
    else
    {
        //afterResize we have to follow server size
        width = serverSize.width();
        height = serverSize.height();
    }

    clientFrameBufferSize->setWidth(width);
    clientFrameBufferSize->setHeight(height);
}

#if 0

QSize QTmRfbClient::setServerFbSize(QSize size) {
#ifdef QT_DEBUG_TM_RFBCLIENT
        qDebug() << "QTmRfbClient::setServerFbSize:" << "Native FB size" << size;
#endif
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbWidth,  size.width());
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbHeight, size.height());
    bool scaling = m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbDownScaling, true);
    if (!scaling) {
        int clientWidth  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelWidth, 0);
        int clientHeight = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelHeight, 0);
#ifdef QT_DEBUG_TM_RFBCLIENT
        qDebug() << "QTmRfbClient::setServerFbSize:" << "FB Scaling not supported";
        qDebug() << "QTmRfbClient::setServerFbSize:" << "Client FB size" << QSize(clientWidth, clientHeight);
#endif
        if (size.width() > clientWidth)   size.setWidth(clientWidth);
        if (size.height() > clientHeight) size.setHeight(clientHeight);
    }
#ifdef QT_DEBUG_TM_RFBCLIENT
    else {
        qDebug() << "QTmRfbClient::setServerFbSize:" << "FB Scaling supported";
    }
    qDebug() << "QTmRfbClient::setServerFbSize:" << "Request FB size" << size;
#endif
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbWidth,  size.width());
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbHeight, size.height());
    return size;
}

#else

QSize QTmRfbClient::setServerFbSize(QSize size) {
#ifdef QT_DEBUG_TM_RFBCLIENT
        qDebug() << "QTmRfbClient::setServerFbSize:" << "Native FB size" << size;
#endif
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbWidth,  size.width());
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigFbHeight, size.height());

    //m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbWidth,  size.width());   //??
    //m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbHeight, size.height());  //??
    return size;
}

#endif


#if 0
QSize QTmRfbClient::setClientFbSize(QSize size, bool scaling) {
#ifdef QT_DEBUG_TM_RFBCLIENT
        qDebug() << "QTmRfbClient::setClientFbSize:" << "Native FB size" << size << scaling;
#endif
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigFbUpScaling,   scaling);
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigFbDownScaling, scaling);
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigDisplayPixelWidth,  size.width());
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigDisplayPixelHeight, size.height());

    if (!scaling) {
        int serverWidth  = m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbWidth, 0);
        int serverHeight = m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbHeight, 0);
#ifdef QT_DEBUG_TM_RFBCLIENT
        qDebug() << "QTmRfbClient::setClientFbSize:" << "FB Scaling not supported";
        qDebug() << "QTmRfbClient::setClientFbSize:" << "Server FB size" << QSize(serverWidth, serverHeight);
#endif
        if (size.width() > serverWidth)   size.setWidth(serverWidth);
        if (size.height() > serverHeight) size.setHeight(serverHeight);
    }
#ifdef QT_DEBUG_TM_RFBCLIENT
    else {
        qDebug() << "QTmRfbClient::setClientFbSize:" << "FB Scaling supported";
    }
    qDebug() << "QTmRfbClient::setClientFbSize:" << "Request FB size" << size;
#endif

    if(((size.width() > QTmConfiguration::ConfigMaxDisplayWidth) | (size.width() > QTmConfiguration::ConfigMaxDisplayHeight))
    {
        m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbWidth,  QTmConfiguration::ConfigMaxDisplayWidth);
        m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbHeight, QTmConfiguration::ConfigMaxDisplayHeight);
        size.setWidth(QTmConfiguration::ConfigMaxDisplayWidth);
        size.setHeight(QTmConfiguration::ConfigMaxDisplayHeight);
    }
    else
    {
        m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbWidth,  size.width());
        m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbHeight, size.height());
    }

    return size;
}

#else

QSize QTmRfbClient::setClientFbSize(QSize size, bool scaling) {
#ifdef QT_DEBUG_TM_RFBCLIENT
        qDebug() << "QTmRfbClient::setClientFbSize:" << "Native FB size" << size << scaling;
#endif
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigFbUpScaling,   scaling);
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigFbDownScaling, scaling);

    if((size.width() > QTmConfiguration::ConfigMaxDisplayWidth) || (size.height() > QTmConfiguration::ConfigMaxDisplayHeight))
    {
        m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbWidth,  QTmConfiguration::ConfigMaxDisplayWidth);
        m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbHeight, QTmConfiguration::ConfigMaxDisplayHeight);

        m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigDisplayPixelWidth,  QTmConfiguration::ConfigMaxDisplayWidth);
        m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigDisplayPixelHeight, QTmConfiguration::ConfigMaxDisplayHeight);

        size.setWidth(QTmConfiguration::ConfigMaxDisplayWidth);
        size.setHeight(QTmConfiguration::ConfigMaxDisplayHeight);
    }
    else
    {
        m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbWidth,  size.width());
        m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbHeight, size.height());

        m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigDisplayPixelWidth,  size.width());
        m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigDisplayPixelHeight, size.height());
    }

    return size;
}
#endif

// ********************************************************************
// *****               Event Configuration message                *****
// ********************************************************************
/*!
  \fn QTmRfbClient::eventConfigurationReceived()
  Signal emitted from a RFB client to inform about an server event configuration message received.
  The receiving component should use this information to adapt its own eventing mechanisms, to support as
  many server events as possible.
 */
/*!
  Receive an RFB Server Event Configuration message. This message is a Terminal Mode VNC extension message.
  The server event configuration message provides information about the event mechanisms at the VNC server.
  This includes the following elements:
  \list
    \o Keyboard Language and Country
    \o User Interface Language and Country
    \o Supported Device, Multimedia and Function keys
    \o Supported key and pointer event mechanisms
  \endlist
  The received information is stored in the server configuration isnstace. After emitting the
  eventConfigurationReceived() signal, the client sends its own Client Event Configuration message automatically.
  In case framebuffer updates are still disabled, the client will send a framebuffer update request message.
  Returns the number of bytes read.
 */
int QTmRfbClient::serverEventConfig() {
    unsigned char cmdBuffer[28];
    int read = receiveData(28, (char*)cmdBuffer);
    if (read < 28) {
        qDebug() << "QRfbClientTm:" << "Error receiving Server Event Configuration";
        handleSocketError();
        return read;
    }
    //Read the Server Language settings
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigKeyboardLanguage,   (cmdBuffer[0] << 8) | cmdBuffer[1]);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigKeyboardCountry,    (cmdBuffer[2] << 8) | cmdBuffer[3]);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigUiLanguage,         (cmdBuffer[4] << 8) | cmdBuffer[5]);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigUiCountry,          (cmdBuffer[6] << 8) | cmdBuffer[7]);
    //Read the Server Event settings
    int eventKnob    = (cmdBuffer[8]  << 24) | (cmdBuffer[9]  << 16) | (cmdBuffer[10] << 8) | cmdBuffer[11];
    int eventDevice  = (cmdBuffer[12] << 24) | (cmdBuffer[13] << 16) | (cmdBuffer[14] << 8) | cmdBuffer[15];
    int eventMedia   = (cmdBuffer[16] << 24) | (cmdBuffer[17] << 16) | (cmdBuffer[18] << 8) | cmdBuffer[19];

    qDebug()<< "Server supported Knob keys :[4]=" << cmdBuffer[8] << "[3]=" << cmdBuffer[9] << "[2]=" << cmdBuffer[10] << "[1]=" << cmdBuffer[11];
    qDebug()<< "Server supported Device keys :[4]=" << cmdBuffer[12] << "[3]=" << cmdBuffer[13] << "[2]=" << cmdBuffer[14] << "[1]=" << cmdBuffer[15];
    qDebug()<< "Server supported Media keys :[4]=" << cmdBuffer[16] << "[3]=" << cmdBuffer[17] << "[2]=" << cmdBuffer[18] << "[1]=" << cmdBuffer[19];

    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigEventKnobKey,      eventKnob);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigEventDeviceKey,    eventDevice);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigEventMediaKey,     eventMedia);
    //Key event related
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigEventItuKeypad,   (cmdBuffer[23] >> 0) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigVirtualKeyboard,  (cmdBuffer[23] >> 1) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigKeyEventListing,  (cmdBuffer[23] >> 2) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigKeyEventMapping,  (cmdBuffer[23] >> 3) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigEventFunctionKeys, cmdBuffer[22]);
    //Pointer event related
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigEventPointer,     (cmdBuffer[27] >> 0) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigEventTouch,       (cmdBuffer[27] >> 1) & 0x01);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigButtonMask,        cmdBuffer[26]);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigTouchNumber,       cmdBuffer[25]);
    m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigPressureMask,      cmdBuffer[24]);
    //Emit signal on message reception
    qDebug()<< "QRfbClientTm:" << "KnobEvent Conf Received is "<< eventKnob;
    qDebug()<< "QRfbClientTm:" << "DeviceEvent Conf Received is "<< eventDevice;
    qDebug()<< "QRfbClientTm:" << "MediaEvent Conf Received is "<< eventMedia;
    emit eventConfigurationReceived(eventKnob, eventDevice, eventMedia);
    //Send response message
    clientEventConfig();
    //Send framebuffer updates, if not already enabled
    if (m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbUpdates, false)) {
        triggerFramebufferUpdates();
    }
    return 28;
}
/*!
  Send an RFB Client Event Configuration message. This message is a Terminal Mode VNC extension message.
  The client event configuration message provides information about the event mechanisms at the VNC server.
  This includes the following elements:
  \list
    \o Keyboard Language and Country
    \o User Interface Language and Country
    \o Supported Device, Multimedia and Function keys
    \o Supported key and pointer event mechanisms
  \endlist
  The transmitted information is taken from the client configuration isnstace.
 */
void QTmRfbClient::clientEventConfig() {
    QSize clientFrameBufferSize;
    unsigned char cmdBuffer[32];
    cmdBuffer[0]  = EXT_MSG_TERMINAL_MODE;
    cmdBuffer[1]  = TM_MSG_CLIENT_EVENT_CONFIG;
    cmdBuffer[2]  = 0x00;
    cmdBuffer[3]  = 0x1C;
    //Language Setting
    cmdBuffer[4]  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigKeyboardLanguage, false) >> 8;
    cmdBuffer[5]  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigKeyboardLanguage, false);
    cmdBuffer[6]  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigKeyboardCountry,  false) >> 8;
    cmdBuffer[7]  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigKeyboardCountry,  false);
    cmdBuffer[8]  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigUiLanguage,       false) >> 8;
    cmdBuffer[9]  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigUiLanguage,       false);
    cmdBuffer[10] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigUiCountry,        false) >> 8;
    cmdBuffer[11] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigUiCountry,        false);


    qDebug()<< "QRfbClientTm:" << "KnobEvent Conf Sending is : "<< m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventKnobKey, false);
    qDebug()<< "QRfbClientTm:" << "DeviceEvent Conf Sending is : "<< m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventDeviceKey, false);
    qDebug()<< "QRfbClientTm:" << "MediaEvent Conf Sending is : "<< m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventMediaKey, false);


    // Multifunction knob configuration
    cmdBuffer[12] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventKnobKey, false) >> 24;
    cmdBuffer[13] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventKnobKey, false) >> 16;
    cmdBuffer[14] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventKnobKey, false) >>  8;
    cmdBuffer[15] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventKnobKey, false);
    // Mobile device key configuration
    cmdBuffer[16] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventDeviceKey, false) >> 24;
    cmdBuffer[17] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventDeviceKey, false) >> 16;
    cmdBuffer[18] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventDeviceKey, false) >>  8;
    cmdBuffer[19] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventDeviceKey, false);
    // Multimedia key configuration
    cmdBuffer[20] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventMediaKey, false) >> 24;
    cmdBuffer[21] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventMediaKey, false) >> 16;
    cmdBuffer[22] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventMediaKey, false) >>  8;
    cmdBuffer[23] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventMediaKey, false);
    // Key event configuration
    cmdBuffer[24] = 0x00;
    cmdBuffer[25] = 0x00;
    cmdBuffer[26] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventFunctionKeys, false);
    cmdBuffer[27] = ((m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventItuKeypad,  false) & 0x01) << 0) |
                    ((m_rfbConfigClient->configuration(QTmConfiguration::ConfigVirtualKeyboard, false) & 0x01) << 1) |
                    ((m_rfbConfigClient->configuration(QTmConfiguration::ConfigKeyEventListing, false) & 0x01) << 2) |
                    ((m_rfbConfigClient->configuration(QTmConfiguration::ConfigKeyEventMapping, false) & 0x01) << 3);
    // Pointer event configuration
    cmdBuffer[28] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigPressureMask, false);
    cmdBuffer[29] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigTouchNumber,  false);
    cmdBuffer[30] = m_rfbConfigClient->configuration(QTmConfiguration::ConfigButtonMask,   false);
    cmdBuffer[31] = ((m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventPointer, false) & 0x01) << 0) |
                    ((m_rfbConfigClient->configuration(QTmConfiguration::ConfigEventTouch,   false) & 0x01) << 1);
    transmitData(32, (char*)cmdBuffer);
    m_keyEventCounter = 0;


    QTmGlobal::TmColorFormat colorFormat = (QTmGlobal::TmColorFormat) m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbColorFormat);

    //decide the framebuffer size
    getExpectedClientSize(QSize(m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbWidth,  0), m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbHeight,  0)),
                          &clientFrameBufferSize, false);

    //allocate the framebuffer
    if(m_surface)
    {
        m_framebuffer = m_surface->allocateSurface(clientFrameBufferSize, colorFormat, m_calculatedXScaleFactor, m_calculatedYScaleFactor);
        if(!m_framebuffer)
        {
            qDebug() << "QTmRfbClient::clientDisplayConfig:" << "Allocate surface failed!";
            //m_rfbSessionStatus = QTmRfbClient::ERROR_CANNOT_ALLOCATE_SURFACE;
            //emit sessionDisconnected();
            return;
        }
#ifdef BUFFER_MANAGER_INTEGRATED
                if(m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbUpdates, false))
        m_surface->start();
#endif
    }
    else
    {
        qDebug() << "QTmRfbClient::clientDisplayConfig:" << "No surface registered with rfb client !!";
    }

}


// ********************************************************************
// *****                  Event Mapping message                   *****
// ********************************************************************
/*!
  \fn QTmRfbClient::eventMappingReceived(unsigned int clientKey, unsigned int serverKey)
  Signal emitted from a RFB client to inform about an event mapping message received.
  \a serverKey defines the key symbol value at the server, to which the client symboal value \a clientKey
  is mapped.
 */
/*!
  Receive an RFB Event Mapping message. This message is a Terminal Mode VNC extension message.
  Read the new key mapping and emit a eventMappingReceived() signal.
  Returns the number of bytes read.
 */
int QTmRfbClient::eventMapping() {
    unsigned char cmdBuffer[8];
    int read = receiveData(8, (char *) cmdBuffer);
    qDebug() << "QTmRfbClient: eventMapping() read : " << read;
    if (read < 8) {
        qDebug() << "QRfbClientTm:" << "Error receiving Event Mapping message";
        handleSocketError();
        return read;
    }
    unsigned int clientKey = (cmdBuffer[0]<<24) | (cmdBuffer[1]<<16) | (cmdBuffer[2]<<8) | cmdBuffer[3];
    unsigned int serverKey = (cmdBuffer[4]<<24) | (cmdBuffer[5]<<16) | (cmdBuffer[6]<<8) | cmdBuffer[7];
    m_eventMap->insert(clientKey, serverKey);
    emit eventMappingReceived(clientKey, serverKey);
    return 8;
}
/*!
  Send an RFB Event Mapping Request message. This message is a Terminal Mode VNC extension message.
  \a serverKey defines the key symbol value at the server, to which the client symboal value \a clientKey
  should be mapped.
 */
void QTmRfbClient::eventMappingRequest(unsigned int clientKey, unsigned int serverKey) {
    unsigned char cmdBuffer[12];
    cmdBuffer[0]  = EXT_MSG_TERMINAL_MODE;
    cmdBuffer[1]  = TM_MSG_EVENT_MAPPING_REQ;
    cmdBuffer[2]  = 0x00;
    cmdBuffer[3]  = 0x08;
    cmdBuffer[4]  = clientKey >> 24;
    cmdBuffer[5]  = clientKey >> 16;
    cmdBuffer[6]  = clientKey >>  8;
    cmdBuffer[7]  = clientKey;
    cmdBuffer[8]  = serverKey >> 24;
    cmdBuffer[9]  = serverKey >> 16;
    cmdBuffer[10] = serverKey >>  8;
    cmdBuffer[11] = serverKey;
    transmitData(12, (char*)cmdBuffer);
}


// ********************************************************************
// *****                Key Event Listing message                 *****
// ********************************************************************
/*!
  \fn QTmRfbClient::keyEventListReceived(QList<unsigned int> *list, unsigned int counter)
  Signal emitted from a RFB client to inform about a key event listing message received.
  This message is meant to inform the client, that a new key event list is available.
  The client can use the new list to change the available virtual key (if applicable).
  \a list provides a pointer to a QList<unsigned int>, containing the supported key symbol values.
  \a counter provides the current key event counter. This can be used to check, whether the list
  is relevant for the current input, or not.
 */
/*!
  Receive an RFB Key Event Listing message. This message is a Terminal Mode VNC extension message.
  Read the supported key events and store them in a QList.
  After receiving all events, a keyEventListReceived() signal is emitted.
  Returns the number of bytes read.
 */
int QTmRfbClient::keyEventListing() {
    unsigned char cmdBuffer[4];
    int read = receiveData(4, (char*)cmdBuffer);
    if (read < 4) {
        qDebug() << "QRfbClientTm:" << "Error receiving Key Event Listing message";
        handleSocketError();
        return read;
    }
    //Get the key event listing configuration. Determine the number of key events in the list.
    unsigned char      key_event_config  =  cmdBuffer[0];
    unsigned short int key_event_number  =  cmdBuffer[1];
    unsigned short int key_event_counter = (cmdBuffer[2] << 8) | cmdBuffer[3];
    Q_UNUSED(key_event_config);
    Q_UNUSED(key_event_counter);
    //Read key event list
    char *buffer = (char *) malloc(key_event_number<<2);
    read = receiveData(key_event_number<<2, buffer);
    if (read < (key_event_number<<2)) {
        qDebug() << "QRfbClientTm:" << "Error receiving Key Event List";
        free(buffer);
        handleSocketError();
        return 4 + read;
    }
    //Create QList of events
    QList<unsigned int> *list = new QList<unsigned int>();
    for (int i=0; i<key_event_number; i++) {
        unsigned int event = (buffer[i<<2] << 24) | (buffer[(i<<2)+1] << 16) | (buffer[(i<<2)+2] << 8) | buffer[(i<<2)+3];
        list->append(event);
    }
    free(buffer);
    emit keyEventListReceived(list, m_keyEventCounter);
    return 4 + read;
}
/*!
  Send an RFB Key Event Listing Request message. This message is a Terminal Mode VNC extension message.
  This message indicates to the VNC server, whether to enable (\a enable = true) or disable (\a enable = false) Key Event Listing.
 */
void QTmRfbClient::keyEventListingRequest(bool enable) {
    if (!m_rfbConfigServer->configuration(QTmConfiguration::ConfigKeyEventListing, true))
        return;
    unsigned char cmdBuffer[8];
    bzero(cmdBuffer, 8);
    cmdBuffer[0] = EXT_MSG_TERMINAL_MODE;
    cmdBuffer[1] = TM_MSG_KEY_EVENT_LISTING_REQ;
    cmdBuffer[3] = 0x04;
    cmdBuffer[7] = enable;
    transmitData(8, (char*)cmdBuffer);
}


// ********************************************************************
// *****             Virtual Keyboard Trigger message             *****
// ********************************************************************
/*!
  \fn QTmRfbClient::virtualKeyboardTriggerReceived(QPoint point, bool remove)
  Signal emitted from a RFB client to inform about a virtual keyboard trigger message received.
  This message is meant to inform the client, that either a virtual keyboard should be shown on
  the remote display (\a remove = false), or a virtual keyboard should be removed (\a remove = true).
  \a point provides the location of the cursor.
 */
/*!
  Receive an RFB Virtual Keyboard Trigger message. This message is a Terminal Mode VNC extension message.
  If the message contains a valid cursor position, a virtualKeyboardTriggerReceived() signal is emitted.
  Returns the number of bytes read.
 */
int QTmRfbClient::virtualKeyboardTrigger() {
    unsigned char cmdBuffer[16];
    int read = receiveData(16, (char *) cmdBuffer);
    if (read < 16) {
        qDebug() << "QRfbClientTm:" << "Error receiving Virtual Keyboard Trigger message";
        handleSocketError();
        return read;
    }
    //Determine a valid cursor position
    if (cmdBuffer[3] & 0x01) {
        int xCursor = (cmdBuffer[4]  << 8) | cmdBuffer[5];
        int yCursor = (cmdBuffer[6]  << 8) | cmdBuffer[7];
        emit virtualKeyboardTriggerReceived(QPoint(xCursor, yCursor), (cmdBuffer[3] & 0x08) ? true : false);
    }
    // Note: Text input size & location and following key event listing indicator are ignored.
    return 16;
}
/*!
  Send an RFB Virtual Keyboard Trigger Request message. This message is a Terminal Mode VNC extension message.
  The message tells the VNC server that the client will handle virtual keyboard trigger messages (\a enable = true)
  or not (\a enable = false).
 */
void QTmRfbClient::virtualKeyboardTriggerRequest(bool enable) {
    unsigned char cmdBuffer[8];
    bzero(cmdBuffer, 8);
    cmdBuffer[0]  = EXT_MSG_TERMINAL_MODE;
    cmdBuffer[1]  = TM_MSG_KEYBOARD_TRIGGER_REQ;
    cmdBuffer[3]  = 0x04;
    cmdBuffer[7]  = (enable) ? 0x01 : 0x00;
    transmitData(8, (char*)cmdBuffer);
}


// ********************************************************************
// *****                   Device Status message                  *****
// ********************************************************************
/*!
  \fn QTmRfbClient::deviceStatusReceived(unsigned int deviceStatus)
  Signal emitted from a RFB client to inform about a device status message received.
  This message is meant to inform the client, that the server has changed its device status.
  The configuration vector \a deviceStatus contains the new status.
 */
/*!
  Receive an RFB Device Status message. This message is a Terminal Mode VNC extension message.
  Read the device status and set the Server configuration.
  A deviceStatusReceived() signal is emitted.
  Returns the number of bytes read.
 */
int QTmRfbClient::deviceStatus() {
    unsigned char cmdBuffer[4];
    int read = receiveData(4, (char *) cmdBuffer);
    if (read < 4) {
        qDebug() << "QRfbClientTm:" << "Error receiving Device Status message";
        handleSocketError();
        return read;
    }
    if (( cmdBuffer[3]       & 0x03) != 1)
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigDeviceKeyLock,     (cmdBuffer[3]       & 0x03));
    if (((cmdBuffer[3] >> 2) & 0x03) != 1)
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigDeviceDeviceLock, ((cmdBuffer[3] >> 2) & 0x03));
    if (((cmdBuffer[3] >> 4) & 0x03) != 1)
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigDeviceScreenSaver,((cmdBuffer[3] >> 4) & 0x03));
    if (((cmdBuffer[3] >> 6) & 0x03) != 1)
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigDeviceNightMode,  ((cmdBuffer[3] >> 6) & 0x03));
    if (( cmdBuffer[2]       & 0x03) != 1)
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigDeviceVoiceInput,  (cmdBuffer[2]       & 0x03));
    if (((cmdBuffer[2] >> 2) & 0x03) != 1)
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigDeviceMicInput,   ((cmdBuffer[2] >> 2) & 0x03));
    if (( cmdBuffer[1]       & 0x03) != 1)
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigDeviceDistraction, (cmdBuffer[1]       & 0x03));
    if (( cmdBuffer[0] & 0x03) != 1 && (cmdBuffer[0] & 0x03) != 2 && (cmdBuffer[0] & 0x03) != 3)
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigDeviceRotation,    (cmdBuffer[0]       & 0x07));
    if (((cmdBuffer[0] >> 3) & 0x03) != 1)
        m_rfbConfigServer->setConfiguration(QTmConfiguration::ConfigDeviceOrientation,((cmdBuffer[0] >> 3) & 0x03));
    //emit device status
    unsigned int deviceStatus = (cmdBuffer[0]<<24) | (cmdBuffer[1]<<16) | (cmdBuffer[2]<<8) | cmdBuffer[3];

    emit deviceStatusReceived(deviceStatus);
    return 4;
}
/*!
  Send an RFB Device Status Request message. This message is a Terminal Mode VNC extension message.
  The device status requested is taken from the server configuration class instance.
 */
void QTmRfbClient::deviceStatusRequest() {
    unsigned char cmdBuffer[8];
    bzero(cmdBuffer, 8);
    cmdBuffer[0] = EXT_MSG_TERMINAL_MODE;
    cmdBuffer[1] = TM_MSG_DEVICE_STATUS_REQ;
    cmdBuffer[3] = 0x04;
    cmdBuffer[7] = (m_rfbConfigServer->configuration(QTmConfiguration::ConfigDeviceKeyLock)     << 0) |
                   (m_rfbConfigServer->configuration(QTmConfiguration::ConfigDeviceDeviceLock)  << 2) |
                   (m_rfbConfigServer->configuration(QTmConfiguration::ConfigDeviceScreenSaver) << 4) |
                   (m_rfbConfigServer->configuration(QTmConfiguration::ConfigDeviceNightMode)   << 6);
    cmdBuffer[6] = (m_rfbConfigServer->configuration(QTmConfiguration::ConfigDeviceVoiceInput)  << 0) |
                   (m_rfbConfigServer->configuration(QTmConfiguration::ConfigDeviceMicInput)    << 2);
    cmdBuffer[5] = (m_rfbConfigServer->configuration(QTmConfiguration::ConfigDeviceDistraction) << 0);
    cmdBuffer[4] = (m_rfbConfigServer->configuration(QTmConfiguration::ConfigDeviceRotation)    << 0) |
                   (m_rfbConfigServer->configuration(QTmConfiguration::ConfigDeviceOrientation) << 3);
    transmitData(8, (char*)cmdBuffer);
}


// ********************************************************************
// *****                Content Attestation message               *****
// ********************************************************************
/*!
  Receive an RFB Content Attestation message. This message is a Terminal Mode VNC extension message.
  Returns the number of bytes read.

  TODO: Implementation (currently empty body)
 */
int QTmRfbClient::contentAttestation() {
    return 0;
}
/*!
  Send an RFB Content Attestation Request message. This message is a Terminal Mode VNC extension message.
  The VNC server must attest the next framebuffer update message, according to the provided configuration.
  \a nounce, provides a random nounce the VNC server must use in its response.
  \a attestation, describes the elements of the framebuffer update message to be attested.
  \a signature TBD
  \a key TBD
  \a session_key TBD

  TODO: Implementation (currently empty body)
 */
void QTmRfbClient::contentAttestationRequest(char */*nounce*/, char /*attestation*/, char /*signature*/, char /*key*/, char */*session_key*/) {
    return;
}


// ********************************************************************
// *****               Blocking Notification message              *****
// ********************************************************************
/*!
  Send an RFB Framebuffer Blocing Notification message. This message is a Terminal Mode VNC extension message.
  The message tells the VNC server that the client has blocked the rectangular framebuffer area, as given in
  \a fbArea. This area corresponds to the application  with the application identifier \a appId.
  The reason for blocking is given in \a reason.
 */
void QTmRfbClient::framebufferBlockingNotification(QRect fbArea, unsigned int appId, unsigned int reason) {
    unsigned char cmdBuffer[18];
    cmdBuffer[0]  = EXT_MSG_TERMINAL_MODE;
    cmdBuffer[1]  = TM_MSG_FRAMEBUFFER_BLOCKING;
    cmdBuffer[2]  = 0x00;
    cmdBuffer[3]  = 0x0E;
    cmdBuffer[4]  = fbArea.x() >> 8;
    cmdBuffer[5]  = fbArea.x();
    cmdBuffer[6]  = fbArea.y() >> 8;
    cmdBuffer[7]  = fbArea.y();
    cmdBuffer[8]  = fbArea.width() >> 8;
    cmdBuffer[9]  = fbArea.width();
    cmdBuffer[10] = fbArea.height() >> 8;
    cmdBuffer[11] = fbArea.height();
    cmdBuffer[12] = appId >> 24;
    cmdBuffer[13] = appId >> 16;
    cmdBuffer[14] = appId >>  8;
    cmdBuffer[15] = appId;
    cmdBuffer[16] = reason >> 8;
    cmdBuffer[17] = reason;
    transmitData(18, (char*)cmdBuffer);
}
/*!
  Send an RFB Audio Blocing Notification message. This message is a Terminal Mode VNC extension message.
  The message tells the VNC server that the client has blocked an incoming audio stream.
  The stream belongs to the application with the application identifier \a appId.
  The reason for blocking is given in \a reason.
 */
void QTmRfbClient::audioBlockingNotification(unsigned int appId, unsigned int reason) {
    unsigned char cmdBuffer[10];
    cmdBuffer[0]  = EXT_MSG_TERMINAL_MODE;
    cmdBuffer[1]  = TM_MSG_AUDIO_BLOCKING;
    cmdBuffer[2]  = 0x00;
    cmdBuffer[3]  = 0x06;
    cmdBuffer[4] = appId >> 24;
    cmdBuffer[5] = appId >> 16;
    cmdBuffer[6] = appId >>  8;
    cmdBuffer[7] = appId;
    cmdBuffer[8] = reason >> 8;
    cmdBuffer[9] = reason;
    transmitData(10, (char*)cmdBuffer);
}


// ********************************************************************
// *****           Framebuffer Alternative Text message           *****
//*********************************************************************
/*!
  \fn QTmRfbClient::framebufferAlternativeTextReceived(unsigned int appId, QString text)
  Signal emitted from a RFB client to inform about a new framebuffer alternative text message received.
  The application \a appId uniquely identfies the application providing the alternative text \a text.
 */
/*!
  Receive an RFB Framebuffer Alternative Text message.
  The signal framebufferAlternativeTextReceived() is emitted after the message has been fully received.
  Returns the number of bytes read.
 */
int QTmRfbClient::framebufferAlternativeText() {
    unsigned char cmdBuffer[6];
    int read = receiveData(6, (char *) cmdBuffer);
    if (read < 6) {
        qDebug() << "QRfbClientTm:" << "Error receiving Framebuffer Alternative Text message";
        handleSocketError();
        return read;
    }
    unsigned int appId  = (cmdBuffer[0]<<24) | (cmdBuffer[1]<<16) | (cmdBuffer[2]<<8) | cmdBuffer[3];
    unsigned int length = (cmdBuffer[4]<<16) |  cmdBuffer[5];
    char *text = (char *) malloc(length+1);
    text[length] = '\0';
    read = receiveData(length, text);
    if (read < (int) length) {
        qDebug() << "QRfbClientTm:" << "Error receiving Framebuffer Alternative Text message";
        handleSocketError();
        free(text);
        return read + 6;
    }
    emit framebufferAlternativeTextReceived(appId, QString(text));
    free(text);
    return 6 + length;
}
/*!
  Send an RFB Framebuffer Alternative Text Request message. The message is a Terminal Mode VNC extension message.
  The message requests framebuffer alternative text messages from the VNC server.
  The maximum length of the text in those messages in given in \a max_length.
 */
void QTmRfbClient::framebufferAlternativeTextRequest(unsigned short int max_length) {
    unsigned char cmdBuffer[6];
    cmdBuffer[0]  = EXT_MSG_TERMINAL_MODE;
    cmdBuffer[1]  = TM_MSG_FRAMEBUFFER_ALT_TEXT_REQ;
    cmdBuffer[2]  = 0x00;
    cmdBuffer[3]  = 0x02;
    cmdBuffer[4]  = max_length >> 8;
    cmdBuffer[5]  = max_length;
    transmitData(6, (char*)cmdBuffer);
}

/**
 * Protocol FSM
 */
void QTmRfbClient::protocolFsm() {
#ifdef QT_UI_CLEANUP
    int receivedBytes = 0;
#endif

    qDebug() << "QTmRfbClient - " << this;

    unsigned char cmdBuffer[4];
    int length, read;
    switch (m_rfbSessionStatus) {
    case RFB_SESSION_CONNECTED_ACTIVE:
    case RFB_SESSION_CONNECTED_PASSIVE:
        receivedBytes = receiveData(1, (char *)cmdBuffer);
        if(receivedBytes < 1) {
            qDebug()<< "Error occured while protocolFSM";
            handleSocketError();
            break;
        }
/*        if (m_socket->read((char *)cmdBuffer, 1) != 1) {
            m_rfbSessionStatus = ERROR_RFB_SESSION_DISCONNECTED;
            emit sessionDisconnected();
            break;
        }*/
//        qDebug()<<"Message type:" << cmdBuffer[0] <<
//                (cmdBuffer[0] == MSG_FRAMEBUFFER_UPDATE);
        switch (cmdBuffer[0]) {
        case MSG_FRAMEBUFFER_UPDATE:
#ifdef QT_DEBUG_TM_RFBCLIENT
            qDebug() << "QTmRfbClient::protocolFsm:" << "MSG_FRAMEBUFFER_UPDATE";
#endif
            framebufferUpdate();
            break;
        case MSG_SET_COLOUR_MAP_ENTRIES:
#ifdef QT_DEBUG_TM_RFBCLIENT
            qDebug() << "QTmRfbClient::protocolFsm:" << "MSG_SET_COLOUR_MAP_ENTRIES";
#endif
            setColourMapEntries();
            break;
        case MSG_BELL:
#ifdef QT_DEBUG_TM_RFBCLIENT
            qDebug() << "QTmRfbClient::protocolFsm:" << "MSG_BELL";
#endif
            bell();
            break;
        case MSG_SERVER_CUT_TEXT:
#ifdef QT_DEBUG_TM_RFBCLIENT
            qDebug() << "QTmRfbClient::protocolFsm:" << "MSG_SERVER_CUT_TEXT";
#endif
            serverCutText();
            break;
        case MSG_TERMINAL_MODE:
            receivedBytes = receiveData(3, (char*)cmdBuffer+1);
            if(receivedBytes < 3) {
                qDebug() << "Error occured while protocol fsm";
                handleSocketError();
                break;
            }
            length = (cmdBuffer[2]<<8) | cmdBuffer[3];
#ifdef QT_DEBUG_TM_RFBCLIENT
            qDebug() << "QTmRfbClient::protocolFsm: TM msg payload len: " << length;
#endif
            read   = 0;
            switch (cmdBuffer[1]) {
            case TM_MSG_SERVER_DISPLAY_CONFIG:
#ifdef QT_DEBUG_TM_RFBCLIENT
                qDebug() << "QTmRfbClient::protocolFsm:" << "TM_MSG_SERVER_DISPLAY_CONFIG";
#endif
                read = serverDisplayConfig();
                break;
            case TM_MSG_SERVER_EVENT_CONFIG:
#ifdef QT_DEBUG_TM_RFBCLIENT
                qDebug() << "QTmRfbClient::protocolFsm:" << "TM_MSG_SERVER_EVENT_CONFIG";
#endif
                read = serverEventConfig();
                break;
            case TM_MSG_EVENT_MAPPING:
#ifdef QT_DEBUG_TM_RFBCLIENT
                qDebug() << "QTmRfbClient::protocolFsm:" << "TM_MSG_EVENT_MAPPING";
#endif
                read = eventMapping();
                break;
            case TM_MSG_KEY_EVENT_LISTING:
#ifdef QT_DEBUG_TM_RFBCLIENT
                qDebug() << "QTmRfbClient::protocolFsm:" << "TM_MSG_KEY_EVENT_LISTING";
#endif
                read = keyEventListing();
                break;
            case TM_MSG_KEYBOARD_TRIGGER:
#ifdef QT_DEBUG_TM_RFBCLIENT
                qDebug() << "QTmRfbClient::protocolFsm:" << "TM_MSG_KEYBOARD_TRIGGER";
#endif
                read = virtualKeyboardTrigger();
                break;
            case TM_MSG_DEVICE_STATUS:
#ifdef QT_DEBUG_TM_RFBCLIENT
                qDebug() << "QTmRfbClient::protocolFsm:" << "TM_MSG_DEVICE_STATUS";
#endif
                read = deviceStatus();
                break;
            case TM_MSG_CONTENT_ATTESTATION_RESP:
#ifdef QT_DEBUG_TM_RFBCLIENT
                qDebug() << "QTmRfbClient::protocolFsm:" << "TM_MSG_CONTENT_ATTESTATION_RESP";
#endif
                read = contentAttestation();
                break;
            case TM_MSG_FRAMEBUFFER_ALT_TEXT:
#ifdef QT_DEBUG_TM_RFBCLIENT
                qDebug() << "QTmRfbClient::protocolFsm:" << "TM_MSG_FRAMEBUFFER_ALT_TEXT";
#endif
                read = framebufferAlternativeText();
                break;
            default:
                read = 0;
                qDebug() << "QRfbClientTm::rfb_protocol_ext_handle_msg: Unknown Type:" << cmdBuffer[1];
                break;
            }
            //Payload remaining from the message will be read and ignored.
            if (read < length) {
                qDebug() << "Reading remaining payload !!";
                char *dummyBuffer = (char *) malloc(length-read);
                receivedBytes = receiveData(length-read, (char*)dummyBuffer);
                qDebug() << "receivedBytes : " << receivedBytes;
                if(receivedBytes < length-read) {
                    qDebug() << "Error occured while protocol FSM";
                    handleSocketError();
                    break;
                }
                delete dummyBuffer;
            }
            break;
        default:
            qDebug() << "QTmRfbClient:" << "Error - Wrong VNC message " << cmdBuffer[0];
            //m_rfbSessionStatus = QTmRfbClient::ERROR_WRONG_MSG_TYPE;
            //emit sessionDisconnected();
            m_socket->readAll();
            break;
        }
        break;
    case RFB_SESSION_SERVER_VERSION:
#ifdef QT_DEBUG_TM_RFBCLIENT
        qDebug() << "QTmRfbClient::protocolFsm:" << "RFB_SESSION_SERVER_VERSION";
#endif
        protocolHandshake();
        break;
    case RFB_SESSION_SECURITY_TYPES:
#ifdef QT_DEBUG_TM_RFBCLIENT
        qDebug() << "QTmRfbClient::protocolFsm:" << "RFB_SESSION_SECURITY_TYPES";
#endif
        securityHandshake();
        break;
    case RFB_SESSION_SECURITY_RESPONSE:
#ifdef QT_DEBUG_TM_RFBCLIENT
        qDebug() << "QTmRfbClient::protocolFsm:" << "RFB_SESSION_SECURITY_RESPONSE";
#endif
        securityResponse();
        break;
    case RFB_SESSION_SERVER_INIT:
#ifdef QT_DEBUG_TM_RFBCLIENT
        qDebug() << "QTmRfbClient::protocolFsm:" << "RFB_SESSION_SERVER_INIT";
#endif
        serverInit();
        break;
    case RFB_SESSION_NONE:
        break;
    default:
        qDebug() << "QTmRfbClient:" << "Error - Wrong FSM State 2 " << cmdBuffer[0];
        //m_rfbSessionStatus = QTmRfbClient::ERROR_WRONG_FSM_STATE;
        //emit sessionDisconnected();
        m_socket->readAll();
        break;
    }
}


/**
 * <b>Implementation Details:</b>
 *
 * Go through the framebuffer update area, line-by-line.
 * Read each line individually and copy the received bytes into the client framebuffer.
 */
int
QTmRfbClient::rawEncoding(int x, int y, int width, int height) {
    int receivedBytes =0 ;
    int totalBytes = 0;
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::rawEncoding:" << "Receive RAW Encoding" << QRect(x, y, width, height);
#endif
    QTmGlobal::TmColorFormat colorFormat;
    colorFormat   = (QTmGlobal::TmColorFormat) m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::RGBNone);
    //int fb_width  = m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbWidth,  0);
    //int fb_width  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelWidth,  0);
    int fb_width  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigRequestedFbWidth,  0);
    qDebug()<< "Color format :" << colorFormat << ", fb_width :"<< fb_width;
    switch (colorFormat) {
      case QTmGlobal::RGB888: {
        unsigned int *fb_down = (unsigned int *) m_framebuffer + y * fb_width + x;
        for (;height>0; height--) {
            receivedBytes = receiveData(width<<2, (char *) fb_down);
            if(receivedBytes < (width<<2))
            {
                qDebug() << "Error occured while reading raw data";
                handleSocketError();
                return receivedBytes;
            }
            fb_down += fb_width;
        }
      } break;
      case QTmGlobal::RGB565:
      case QTmGlobal::RGB555:
      case QTmGlobal::RGB444: {
        unsigned short int *fb_down = (unsigned short int *) m_framebuffer + (y * fb_width) + x;
        int width_byte = width<<1;
        int bytes_read_row;
        for (;height>0; height--) {
            bytes_read_row = 0;
            while (bytes_read_row < width_byte) {
                receivedBytes = receiveData(width_byte - bytes_read_row, ((char *)fb_down)+bytes_read_row);
                if(receivedBytes < 0)
                {
                    qDebug()<< "Error occured while reading raw";
                    //handleSocketError();
                    //break;
                    return 0;
                }
                bytes_read_row += receivedBytes;
                totalBytes += receivedBytes;
                if (bytes_read_row < width_byte) {
                    qDebug() << "insufficient read. wait.";
                    //sleep(1);
                }
            }
            fb_down += fb_width;
        }
        //qDebug() << "QTmRfbClient::rawEncoding:" << "totalBytes read: "  << totalBytes;
      } break;
      case QTmGlobal::RGB343: {
        unsigned short int *fb_down = (unsigned short int *) m_framebuffer + y * fb_width + x;
        for (;height>0; height--) {
            receivedBytes = receiveData(width<<1, (char *) fb_down);
            if(receivedBytes < (width<<1))
            {
                qDebug()<< "Error occured while reading raw";
                handleSocketError();
                return receivedBytes;
            }
            for (int i=width; i>0; i--)
            {
                /**
                 * For RGB343 do local color transformation to RGB444, as QImage does not support RGB343
                 * Add 0x0101 to the color value to increase luminance.
                 */
                unsigned short int color = *fb_down;
                *fb_down++ = ((color & 0x0380) << 2) | ((color & 0x007F) << 1) | 0x0101;
            }
            fb_down += (fb_width - width);
        }
      } break;
      default:
        qDebug() << "QRfbClientFbProcessing:" << "Unknown Color Format";
        return receivedBytes;
    }
    return totalBytes;
}

#ifndef QT_UI_CLEANUP_ENABLE_N97
/**
 * <b>Implementation Details:</b>
 *
 * Go through the framebuffer update area, line-by-line.
 * Read each line individually. The first two bytes of each line, give the number of run-length encodings for this line.
 * The number of bytes read for each run-length, depend on the color format.
 * For every run-length encoding, do the following steps:
 * - decode the run-length @c number
 * - decode the color value @c color
 * - copy the color value @c color, @c number times into the client framebuffer.
 */
#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
int QTmRfbClient::rleEncoding(int x, int y, int width, int height) {
#else
int QTmRfbClient::rleEncoding(int x, int y, int width, int height) {
#endif
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::rleEncoding:" << "Receive RLE Encoding" << QRect(x, y, width, height);
#endif

#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
    int number_of_runs = 0;
#endif
    unsigned int number;
    unsigned int no_runs;
    unsigned int position;
    unsigned int totalBytes = 0;
    unsigned int totalDecodedBytes = 0;
    QTmGlobal::TmColorFormat colorFormat;
    colorFormat   = (QTmGlobal::TmColorFormat) m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::RGBNone);
    //int fb_width  = m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbWidth,  0);
    //int fb_height = m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbHeight, 0);
    //int fb_width  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelWidth,  0);
    //int fb_height = m_rfbConfigClient->configuration(QTmConfiguration::ConfigDisplayPixelHeight, 0);
    int fb_width  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigRequestedFbWidth,  0);
    int fb_height = m_rfbConfigClient->configuration(QTmConfiguration::ConfigRequestedFbHeight, 0);

    qDebug() << "QTmRfbClient::rleEncoding(), fb_width=" << fb_width << ", fb_height=" << fb_height << "colorFormat=" << colorFormat << "m_framebuffer=" << m_framebuffer;

    switch(colorFormat) {
      case QTmGlobal::RGB888: {
        unsigned int  *fb_down = (unsigned int *) m_framebuffer + y * fb_width + x;
        unsigned int   color;
#ifdef VNC_OPTIMIZATION
        unsigned char *cmdBuffer = m_encodingCmdBuffer;
#else
        unsigned char *cmdBuffer = (unsigned char *) malloc(width << 2);
#endif
        for (;height>0; height--) {
            receiveData(2, (char *) cmdBuffer);
            position = 0;
            no_runs  = (cmdBuffer[0] << 8) | cmdBuffer[1];
            receiveData(no_runs << 2, (char *) cmdBuffer);
#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
            number_of_runs += no_runs;
#endif
            for (; no_runs>0; no_runs--) {
                color  = cmdBuffer[position++];
                color |= cmdBuffer[position++] <<  8;
                color |= cmdBuffer[position++] << 16;
                number = cmdBuffer[position++] + 1;
                for (;number>0; number--)
                    *fb_down++ = color;
            }
            fb_down += (fb_width - width);
        }
#ifdef VNC_OPTIMIZATION
        // nothing here
#else
        free(cmdBuffer);
#endif
      } break;
      case QTmGlobal::RGB565:
      case QTmGlobal::RGB555: {
        unsigned short int *fb_down = (unsigned short int *) m_framebuffer + (y * fb_width) + x;
        unsigned short int *fb_end = (unsigned short int *) m_framebuffer + fb_width * fb_height;
        //unsigned short int *fb_end = (unsigned short int *) m_framebuffer + ((y + height) * fb_width) + x;
        unsigned short int  color;
#ifdef VNC_OPTIMIZATION
        unsigned char *cmdBuffer = m_encodingCmdBuffer;
#else
        unsigned char      *cmdBuffer = (unsigned char *) malloc(width*3);
#endif
        for (;height>0; height--) {
            int bytesRead = receiveData(2, (char *) cmdBuffer);
            totalBytes += bytesRead;
            position = 0;
            no_runs  = (cmdBuffer[0] << 8) | cmdBuffer[1];
            bytesRead = receiveData(no_runs*3, (char *) cmdBuffer);
            totalBytes += bytesRead;
            if(bytesRead < 0)
            {
                qDebug() << "Error occured while reading rle data";
                qDebug() << "QTmRfbClient::rleEncoding(), no_runs=" << no_runs << "at height=" << height;
                break;
            }

#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
            number_of_runs += no_runs;
#endif

            for (; no_runs>0; no_runs--) {
                switch (colorFormat) {
                  case QTmGlobal::RGB565:
                    color  = cmdBuffer[position++];
                    color |= cmdBuffer[position++] << 8;
                    number = cmdBuffer[position++] + 1;
                    break;
                  case QTmGlobal::RGB555:
                    color   =  cmdBuffer[position++];
                    color  |= (cmdBuffer[position]   & 0x07) <<  8;
                    number  = (cmdBuffer[position++] >> 7);
                    number |= (cmdBuffer[position++] << 1);
                    break;
                  default:
                    number = 0;
                    color  = 0;
                    break;
                }
                for (;number>0; number--) {
                    if (fb_down > fb_end) {
                        break;
                    }
                    *fb_down = color;
                    totalDecodedBytes++;
                    fb_down++;
                }
            }
            //qDebug() << "QTmRfbClient::rleEncoding: fb_down = " << fb_down << "totalDecodedBytes = " << totalDecodedBytes;
            fb_down += (fb_width - width);
        }
#ifdef VNC_OPTIMIZATION
        // nothing here
#else
        free(cmdBuffer);
#endif
        //qDebug() << "QTmRfbClient::rleEncoding:" << "totalBytes read: "  << totalBytes;
        //qDebug() << "QTmRfbClient::rleEncoding:" << "totalDecodedBytes read: "  << totalDecodedBytes;
      } break;
      case QTmGlobal::RGB444:
      case QTmGlobal::RGB343: {
        unsigned short int *fb_down = (unsigned short int *) m_framebuffer + y * fb_width + x;
        unsigned short int  color;
#ifdef VNC_OPTIMIZATION
        unsigned char *cmdBuffer = m_encodingCmdBuffer;
#else
        unsigned char      *cmdBuffer = (unsigned char *) malloc(width*2);
#endif
        for (;height>0; height--) {
            receiveData(2, (char *) cmdBuffer);
            position = 0;
            no_runs  = (cmdBuffer[0] << 8) | cmdBuffer[1];
            receiveData(no_runs*2, (char *) cmdBuffer);
#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
            number_of_runs += no_runs;
#endif
            for (; no_runs>0; no_runs--) {
                switch (colorFormat) {
                  case QTmGlobal::RGB444:
                    color  =   cmdBuffer[position++];
                    color |=  (cmdBuffer[position]   & 0x0F) << 8;
                    number = ((cmdBuffer[position++] & 0xF0) >> 4) + 1;
                    break;
                  case QTmGlobal::RGB343:
                    color  =   cmdBuffer[position++];
                    color |=  (cmdBuffer[position]   & 0x03) << 8;
                    number = ((cmdBuffer[position++] & 0xFC) >> 2) + 1;
                    /**
                     * For RGB343 do local color transformation to RGB444, as QImage does not support RGB343
                     * Add 0x0101 to the color value to increase luminance.
                     */
                    color  = ((color & 0x0380) << 2) | ((color & 0x007F) << 1) | 0x0101;
                    break;
                  default:
                    number = 0;
                    color  = 0;
                    break;
                }
                for (;number>0; number--)
                    *fb_down++ = color;
            }
            fb_down += (fb_width - width);
        }
#ifdef VNC_OPTIMIZATION
        // nothing here
#else
        free(cmdBuffer);
#endif
      } break;
      default:
        break;
    }
    //qDebug() << "QTmRfbClient::rleEncoding: end of function.";
#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
    return number_of_runs;
#else
    return totalBytes;
#endif
}

#else   //QT_UI_CLEANUP_ENABLE_N97

#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
int QTmRfbClient::rleEncoding(int x, int y, int width, int height) {
#else
void QTmRfbClient::rleEncoding(int x, int y, int width, int height) {
#endif
    int receivedBytes = 0;
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::rleEncoding:" << "Receive RLE Encoding" << QRect(x, y, width, height);
#endif

#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
    int number_of_runs = 0;
#endif
    unsigned int number;
    unsigned int no_runs;
    unsigned int position;
    QTmGlobal::TmColorFormat colorFormat;
    colorFormat   = (QTmGlobal::TmColorFormat) m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbColorFormat, QTmGlobal::RGBNone);
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::rleEncoding:" << "color format :" << colorFormat;
#endif
    int fb_width  = m_rfbConfigServer->configuration(QTmConfiguration::ConfigFbWidth,  0);
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::rleEncoding:" << " fb_width :" << fb_width;
#endif
    switch(colorFormat) {
      case QTmGlobal::RGB888: {
        qDebug() << "RGB 888 ***********";
        unsigned int  *fb_down = (unsigned int *) m_framebuffer + y * fb_width + x;
        unsigned int   color;
        unsigned char *cmdBuffer = (unsigned char *) malloc(width << 2);


        for (;height>0; height--) {
            receivedBytes = receiveData(2, (char *) cmdBuffer);
            if(receivedBytes < 2)
            {
                qDebug("Error occured while rle encoding");
                handleSocketError();
                return;
            }
            position = 0;
            no_runs  = (cmdBuffer[1] << 8) | cmdBuffer[0];      //<<< modified for N97
            receivedBytes = receiveData(no_runs << 2, (char *) cmdBuffer);
            if(receivedBytes < (no_runs << 2))
            {
                qDebug("Error occured while rle encoding");
                handleSocketError();
                return;
            }
#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
            number_of_runs += no_runs;
#endif
            for (; no_runs>0; no_runs--) {
                color  = cmdBuffer[position++];
                color |= cmdBuffer[position++] <<  8;
                color |= cmdBuffer[position++] << 16;
                number = cmdBuffer[position++] + 1;
                for (;number>0; number--)
                    *fb_down++ = color;
            }
            fb_down += (fb_width - width);
        }
        free(cmdBuffer);
      } break;
      case QTmGlobal::RGB565:
      case QTmGlobal::RGB555: {
        unsigned short int *fb_down = (unsigned short int *) m_framebuffer + y * fb_width + x;
        unsigned short int  color;
        unsigned char      *cmdBuffer = (unsigned char *) malloc(width*3);

        for (;height>0; height--) {
            receivedBytes = receiveData(2, (char *) cmdBuffer);
            if(receivedBytes < 2)
            {
                qDebug("Error occured while rle encoding");
                handleSocketError();
                return;
            }
            position = 0;
            no_runs  = (cmdBuffer[1] << 8) | cmdBuffer[0];      //<<< modified for N97
            receivedBytes = receiveData(no_runs*3, (char *) cmdBuffer);
            if(receivedBytes < no_runs*3)
            {
                qDebug("Error occured while rle encoding");
                handleSocketError();
                return;
            }
#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
            number_of_runs += no_runs;
#endif
            for (; no_runs>0; no_runs--) {
                switch (colorFormat) {
                  case QTmGlobal::RGB565:
                    color  = cmdBuffer[position++];
                    color |= cmdBuffer[position++] << 8;
                    number = cmdBuffer[position++] + 1;
                    break;
                  case QTmGlobal::RGB555:
                    color   =  cmdBuffer[position++];
                    color  |= (cmdBuffer[position]   & 0x07) <<  8;
                    number  = (cmdBuffer[position++] >> 7);
                    number |= (cmdBuffer[position++] << 1);
                    break;
                  default:
                    number = 0;
                    color  = 0;
                    break;
                }
                for (;number>0; number--)
                    *fb_down++ = color;
            }
            fb_down += (fb_width - width);
        }
        free(cmdBuffer);
      } break;
      case QTmGlobal::RGB444:
      case QTmGlobal::RGB343: {
        unsigned short int *fb_down = (unsigned short int *) m_framebuffer + y * fb_width + x;
        unsigned short int  color;
        unsigned char      *cmdBuffer = (unsigned char *) malloc(width*2);
        for (;height>0; height--) {
            receivedBytes = receiveData(2, (char *) cmdBuffer);
            if(receivedBytes < 2)
            {
                qDebug("Error occured while rle encoding");
                handleSocketError();
                return;
            }
            position = 0;
            no_runs  = (cmdBuffer[1] << 8) | cmdBuffer[0];      //<<< modified for N97
            receivedBytes = receiveData(no_runs*2, (char *) cmdBuffer);
            if(receivedBytes < no_runs*2)
            {
                qDebug("Error occured while rle encoding");
                handleSocketError();
                return;
            }
#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
            number_of_runs += no_runs;
#endif
            for (; no_runs>0; no_runs--) {
                switch (colorFormat) {
                  case QTmGlobal::RGB444:
                    color  =   cmdBuffer[position++];
                    color |=  (cmdBuffer[position]   & 0x0F) << 8;
                    number = ((cmdBuffer[position++] & 0xF0) >> 4) + 1;
                    break;
                  case QTmGlobal::RGB343:
                    color  =   cmdBuffer[position++];
                    color |=  (cmdBuffer[position]   & 0x03) << 8;
                    number = ((cmdBuffer[position++] & 0xFC) >> 2) + 1;
                    /**
                     * For RGB343 do local color transformation to RGB444, as QImage does not support RGB343
                     * Add 0x0101 to the color value to increase luminance.
                     */
                    color  = ((color & 0x0380) << 2) | ((color & 0x007F) << 1) | 0x0101;
                    break;
                  default:
                    number = 0;
                    color  = 0;
                    break;
                }
                for (;number>0; number--)
                    *fb_down++ = color;
            }
            fb_down += (fb_width - width);
        }
        free(cmdBuffer);
      } break;
      default:
        break;
    }
#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
    return number_of_runs;
#endif
}

#endif  //QT_UI_CLEANUP_ENABLE_N97

// ********************************************************************
// *****           Desktop Size Pseudo Encoding message           *****
//*********************************************************************
/*!
  \fn QTmRfbClient::framebufferResized(QSize size)
  Signal emitted from a RFB client to inform about a Desktop Size Pseudo Encoding message received.
  The new framebuffer size is given in \a size.
 */
/*!
  Receive a RFB Desktop Size Pseudo Encoding message.
  A new framebuffer surface must be allocated, with the new framebuffer resolution.
  In case the new allocation fails, a sessionDisconnected() signal is emitted.
  Otherwise a framebufferResized() message is emitted.
 */
void QTmRfbClient::desktopSizePseudoEncoding(QSize framebufferSize) {
    QSize expectedClientSize;
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::desktopSizePseudoEncoding:" << "Desktop Size signal received" << framebufferSize;
#endif
    if (!m_surface)
    {
        qDebug() << "QTmRfbClient::desktopSizePseudoEncoding:" << "Surface not set before!";
        return;
    }
    framebufferSize = setServerFbSize(framebufferSize);
    getExpectedClientSize(framebufferSize, &expectedClientSize, true);

    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbWidth, expectedClientSize.width());
    m_rfbConfigClient->setConfiguration(QTmConfiguration::ConfigRequestedFbHeight, expectedClientSize.height());

    QTmGlobal::TmColorFormat colorFormat = (QTmGlobal::TmColorFormat) m_rfbConfigClient->configuration(QTmConfiguration::ConfigFbColorFormat);

    m_framebuffer = m_surface->allocateSurface(expectedClientSize, colorFormat, m_calculatedXScaleFactor, m_calculatedYScaleFactor);
    if(!m_framebuffer) {
        qDebug() << "QTmRfbClient::desktopSizePseudoEncoding:" << "Allocate surface failed!";
        //m_rfbSessionStatus = QTmRfbClient::ERROR_CANNOT_ALLOCATE_SURFACE;
        //emit sessionDisconnected();
        return;
    }

    qDebug() << "QTmRfbClient::desktopSizePseudoEncoding: About to emit framebufferResized";
    emit framebufferResized(framebufferSize);
}


// ********************************************************************
// *****       Context Information Pseudo Encoding message        *****
//*********************************************************************
/*!
  \fn QTmRfbClient::contextInformationReceived(QRect area, int appId, QTmAppInfo *appCategory, QTmDisplayInfo *conCategory)
  Signal emitted from a RFB client to inform about a Context Information Pseudo Encoding message received.
  \a area gives the rectangular area, for which the context information is given.
  \a appId gives the application id, of the application, providing the content in question.
  \a appCategory gives the application information.
  \a conCategory gives the content information.
 */
/*!
  Receive a RFB Context Information Pseudo Encoding message.
  After reception, the signal contextInformationReceived() is emitted.
 */
int QTmRfbClient::contextInformationPseudoEncoding(unsigned char *buffer) {
    QRect area   = QRect((buffer[0]  << 8) | buffer[1], (buffer[2]  << 8) | buffer[3],
                         (buffer[4]  << 8) | buffer[5], (buffer[6]  << 8) | buffer[7]);
    unsigned char cmdBuffer[20];
    int receivedBytes = receiveData(20, (char*)cmdBuffer);
    if(receivedBytes < 20)
    {
        qDebug()<< "Error occured while contextInformationPseudoEncoding()";
        handleSocketError();
        return 0;
    }

    int appId    = (cmdBuffer[0] << 24) | (cmdBuffer[1] << 16) | (cmdBuffer[2] << 8) | cmdBuffer[3];

    QTmAppInfo     *appCategory = new QTmAppInfo();
    QTmDisplayInfo *conCategory = new QTmDisplayInfo();
    appCategory->setCategory ((unsigned int) (cmdBuffer[8] << 24) | (cmdBuffer[9] << 16) | (cmdBuffer[10] << 8) | cmdBuffer[11]);

    switch ((cmdBuffer[12] << 24) | (cmdBuffer[13] << 16) | (cmdBuffer[14] << 8) | cmdBuffer[15]) {
    case (int) QTmGlobal::VisualText:           conCategory->setCategory(QTmGlobal::VisualText); break;
    case (int) QTmGlobal::VisualVideo:          conCategory->setCategory(QTmGlobal::VisualVideo); break;
    case (int) QTmGlobal::VisualImage:          conCategory->setCategory(QTmGlobal::VisualImage); break;
    case (int) QTmGlobal::VisualVectorGraphics: conCategory->setCategory(QTmGlobal::VisualVectorGraphics); break;
    case (int) QTmGlobal::Visual3DGraphics:     conCategory->setCategory(QTmGlobal::Visual3DGraphics); break;
    case (int) QTmGlobal::VisualUserInterface:  conCategory->setCategory(QTmGlobal::VisualUserInterface); break;
    case (int) QTmGlobal::VisualCarMode:        conCategory->setCategory(QTmGlobal::VisualCarMode); break;
    case (int) QTmGlobal::VisualMisc:           conCategory->setCategory(QTmGlobal::VisualMisc); break;
    default: conCategory->setCategory(QTmGlobal::VisualUnknown);
    }

    switch ((cmdBuffer[16] << 24) | (cmdBuffer[17] << 16) | (cmdBuffer[18] << 8) | cmdBuffer[19]) {
    case (int) QTmGlobal::VisualRuleMaxFeedbackTime:    conCategory->setRules(QTmGlobal::VisualRuleMaxFeedbackTime); break;
    case (int) QTmGlobal::VisualRuleMinFontSize:        conCategory->setRules(QTmGlobal::VisualRuleMinFontSize); break;
    case (int) QTmGlobal::VisualRuleNoAutoScrolling:    conCategory->setRules(QTmGlobal::VisualRuleNoAutoScrolling); break;
    case (int) QTmGlobal::VisualRuleNoVideo:            conCategory->setRules(QTmGlobal::VisualRuleNoVideo); break;
    default: conCategory->setRules(QTmGlobal::VisualRuleUnknown);
    }

    switch ((cmdBuffer[4] <<  8) |  cmdBuffer[5]) {
    case (int) QTmGlobal::TrustUser:           appCategory->setTrust(QTmGlobal::TrustUser);           break;
    case (int) QTmGlobal::TrustSelfRegistered: appCategory->setTrust(QTmGlobal::TrustSelfRegistered); break;
    case (int) QTmGlobal::TrustRegistered:     appCategory->setTrust(QTmGlobal::TrustRegistered);     break;
    case (int) QTmGlobal::TrustCertified:      appCategory->setTrust(QTmGlobal::TrustCertified);      break;
    default: appCategory->setTrust(QTmGlobal::TrustUnknown);
    }
    switch ((cmdBuffer[6] <<  8) |  cmdBuffer[7]) {
    case (int) QTmGlobal::TrustUser:           conCategory->setTrust(QTmGlobal::TrustUser);           break;
    case (int) QTmGlobal::TrustSelfRegistered: conCategory->setTrust(QTmGlobal::TrustSelfRegistered); break;
    case (int) QTmGlobal::TrustRegistered:     conCategory->setTrust(QTmGlobal::TrustRegistered);     break;
    case (int) QTmGlobal::TrustCertified:      conCategory->setTrust(QTmGlobal::TrustCertified);      break;
    default: conCategory->setTrust(QTmGlobal::TrustUnknown);
    }

    emit contextInformationReceived(area, appId, appCategory, conCategory);

    return receivedBytes;
}


/**
    Calculate the X and Y scaling factors considering the single pixel width and height of client display
  **/

void QTmRfbClient::calculateScalingFactors(int clientDisplayXPixels, int clientDisplayYPixels,int serverXPixels, int serverYPixels)
{
    double displayRatio = (double)((double)clientDisplayXPixels / (double) clientDisplayYPixels);

    double pixelRatio = (double)((double)m_clientSinglePixelWidth / (double)m_clientSinglePixelHeight);

    double phoneAspect = (double)((double) serverXPixels / (double) serverYPixels);

    qDebug()<< "QTmRfbClient::calculateScalingFactors()  clientSinglePixelWidth : " << m_clientSinglePixelWidth;
    qDebug()<< "QTmRfbClient::calculateScalingFactors()  clientSinglePixelHeight : " << m_clientSinglePixelHeight;
    qDebug()<< "QTmRfbClient::calculateScalingFactors()  displayRatio : " << displayRatio;
    qDebug()<< "QTmRfbClient::calculateScalingFactors()  pixelRatio : " << pixelRatio;
    qDebug()<< "QTmRfbClient::calculateScalingFactors()  phoneAspect : " << phoneAspect;

    bool matchVertically;
    if(this->m_considerPixelComp){
        matchVertically = (phoneAspect < (displayRatio * pixelRatio));
    }
    else{
        matchVertically = (phoneAspect < displayRatio);
    }

    if (matchVertically)
    {
       double f = clientDisplayYPixels / (double) serverYPixels;
       if(this->m_considerPixelComp){
            m_calculatedXScaleFactor = f / pixelRatio;
       }
       else
       {
           m_calculatedXScaleFactor = f;
       }
       m_calculatedYScaleFactor = f;
    }
    else
    {
       double f = clientDisplayXPixels / (double) serverXPixels;
       m_calculatedXScaleFactor = f;
       if(this->m_considerPixelComp){
            m_calculatedYScaleFactor = f * pixelRatio;
       }
       else{
           m_calculatedYScaleFactor = f;
       }
    }

    qDebug()<< "QTmRfbClient::calculateScalingFactors()  calculatedXScaleFactor : " << m_calculatedXScaleFactor;
    qDebug()<< "QTmRfbClient::calculateScalingFactors()  calculatedYScaleFactor : " << m_calculatedYScaleFactor;

}

/** END ***********/



// ********************************************************************
// *****             Send and Receive Data from Socket            *****
// ********************************************************************
/*!
  Wait until the number of bytes to be read bytes are available from the TCP socket.
  If bytes are not available (either completely or in part) return 0.
  Timeout counter is set to 5 seconds.
  If bytes are available read \a length bytes and store them in \a buffer
  Returns the number of read bytes.
 */








#ifdef QT_UI_CLEANUP

void QTmRfbClient::handleSocketError()
{
    QProcess pr;
    qDebug()<< "SOCKET ERROR : connection state :" << m_connected;
    if(m_connected)
    {
        m_socketError = true;
        qDebug() << pr.error();
        qDebug() << pr.errorString();
        sessionStop();
    }
}

void QTmRfbClient::setClientId(int clientId)
{
    this->m_clientId = clientId;
}

void QTmRfbClient::setRfbThreadListener(RfbThreadListener* listener)
{
    this->m_threadListenr = listener;
}

void QTmRfbClient::run()
{
#ifdef QT_CODE_DEBUG
        qDebug()<< "RFB client thread started ...";
#endif
        while(m_connected)
        {
#ifdef QT_CODE_DEBUG
      //  qDebug()<< "QTmRfbClient::run() " << m_connected;
#endif
           if(m_socket)
                protocolFsm();
           else
               break;
        }

#ifdef QT_CODE_DEBUG
        qDebug()<< "Exiting RFB client thread ..."<< m_threadListenr << ", error state :" << m_socketError;
#endif
        if(m_threadListenr && m_socketError)
        {
            m_threadListenr->rfbThreadExited(m_clientId);
        }


        // send a callback to app saying coud not start rfb session // preferabel onstatus
        // NOTE : As some of the RFB threads may not start it is advised that vncDisconnectClient
        // do a check on isRunning before quiting it
}

int QTmRfbClient::forceStopRfbRead()
{
    if(m_socket){
        m_socket->forceStopSocketRead();
    }
}

int QTmRfbClient::receiveData(int length, char *buffer) {
//#ifdef QT_CODE_DEBUG
//    qDebug()<<"QTmRfbClient::receiveData(" << length <<")";
//#endif
    if(!m_socket)
        return 0;
    int read = m_socket->socketRead((void *)buffer, length);
    if(read == AppInterface::FAILURE){
        delete m_socket;
        m_socket = 0;
    }
//#ifdef QT_CODE_DEBUG
//    qDebug()<<"returning (" << read <<")";
//#endif
    return read;
}
/*!
  Write \a length number of bytes from the \a buffer to the TCP socket.
  Return the number of successful written bytes.
 */
int QTmRfbClient::transmitData(int length, char *buffer) {
    if (!m_socket)
        return 0;
    int written = m_socket->socketWrite((void*)buffer, length);
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::transmitData written :" <<written;
#endif
    //if (written != length)
      //  emit sessionDisconnected();
    return written;
}

#else
int QTmRfbClient::receiveData(int length, char *buffer) {
    while (m_socket->bytesAvailable() < length) {
        if (m_socket && !m_socket->waitForReadyRead(5 * 1000)) {
            qDebug() << "error" << m_socket->error() << m_socket->errorString();
            emit sessionDisconnected();
            return 0;
        }
    }
    int read = m_socket->read((char *)buffer, length);
    return read;
}
/*!
  Write \a length number of bytes from the \a buffer to the TCP socket.
  Return the number of successful written bytes.
 */
int QTmRfbClient::transmitData(int length, char *buffer) {
    if (!m_socket)
        return 0;
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::transmitData m_mutex.lock();" ;
#endif
    m_mutex.lock();
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::transmitData m_mutex.lock acquired " ;
#endif
    int written = m_socket->write(buffer, length);
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "written :" <<written;
#endif
    m_mutex.unlock();
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::released mutex" ;
#endif
    if (written != length)
        emit sessionDisconnected();
    return written;
}
#endif

void QTmRfbClient::setClientSinglePixelWidth(double singlePixelWidth)
{
    m_clientSinglePixelWidth = singlePixelWidth;
}

void QTmRfbClient::setClientSinglePixelHeight(double singlePixelHeight)
{
    m_clientSinglePixelHeight = singlePixelHeight;
}

void QTmRfbClient::sendFbUpdateRequest(bool incremental)
{
    unsigned char cmdBuffer[10];
    cmdBuffer[0]  = MSG_FRAMEBUFFER_UPDATE_REQUEST;
    cmdBuffer[1]  = (incremental) ? 1 : 0;
    cmdBuffer[2]  = 0;
    cmdBuffer[3]  = 0;
    cmdBuffer[4]  = 0;
    cmdBuffer[5]  = 0;
    cmdBuffer[6]  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigRequestedFbWidth, 0) >> 8;
    cmdBuffer[7]  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigRequestedFbWidth, 0);
    cmdBuffer[8]  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigRequestedFbHeight, 0) >> 8;
    cmdBuffer[9]  = m_rfbConfigClient->configuration(QTmConfiguration::ConfigRequestedFbHeight, 0);


    if(m_fbUpdateRequestFlag == false)
    {
        transmitData(10, (char*)cmdBuffer);
        m_fbUpdateRequestFlag = true;

//#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::sendFbUpdateRequest:" << "... done";
    qDebug() << "QTmRfbClient::sendFbUpdateRequest() : " << "time msec :" << QString::number ( getCurrMillisec(), 'g', 15);
//#endif
    }
    else
    {
#ifdef QT_DEBUG_TM_RFBCLIENT
    qDebug() << "QTmRfbClient::sendFbUpdateRequest: Already sent, ignoring !!";
#endif
    }

}

double QTmRfbClient::getCurrMillisec()
{
    struct timeb timeVal;
    double timeMsec;

    ftime(&timeVal);

    timeMsec = ((1000.0*timeVal.time) + timeVal.millitm);

    //qDebug() << "QTmRfbClient::getCurrMillisec: curr msec" << QString::number(timeMsec, 'f', 0);
    return timeMsec;
}


void  QTmRfbClient::setConsiderPixelComp(bool considerPixelComp)
{
    this->m_considerPixelComp = considerPixelComp;
}

bool  QTmRfbClient::getConsiderPixelComp()
{
    return this->m_considerPixelComp;
}

/*!@{*/ // observer method
void QTmRfbClient::updateIdleBufferAvailability()
{

    unsigned char * buffer = m_surface->getSurface();
    if(buffer)
    {
        m_framebuffer = buffer;
    }
    else
    {
        return;
    }    

    if(m_isSendingFbUpdateRequestFirstTime)
    {
        usleep(1000*500);
        framebufferUpdateRequest(false);
        m_isSendingFbUpdateRequestFirstTime = false;
    }
    else
    {
         framebufferUpdateRequest(clientConfig(QTmConfiguration::ConfigIncrementalUpdate, false));
    }

}
/*!@}*/
