/****************************************************************************
**
** Copyright (C) 2012 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Lesser General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public
** License along with this library; if not, write to the
** Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
**
****************************************************************************/
/*
 * ChangeLog -  
 *				July 05, 2012
 * 				1. Added member function performHTTPAction() that reads based on HTTP header length.
 *              2. Removed the dependency of QnetworkAccessManager() due to its inconsistent behavior.
 *              3. Uses performHTTPAction() explicitly and then calls readDeviceDescription() for reading
 *                 the root device description.
 *              4. As application icons are not needed, the application icon fetching sequence is disabled.
 *                 This improved the upnp negotiation time significantly.
 *
 *              August 02, 2013
 *              1. Added logic to wait 500ms before deleting the TmRemoteServer object; if other event processing of qt is in progress.
 *              2. Added retryWait() function to retry in case of server that dont connect or send data 1st time.
 *              3. Always emit 'remoteServerDeviceRemoved' to ensure the cleanup at upper layers/managers on every ML phone disconnect.
 *              4. Closed the QUpnpHttpServer socket in case of UPnP control point socket error.
 *              5. Changed M_SEARCH_QUERY string to 'ssdp:all' instead of 'TmServerDevice'.
 */

#include <QtCore>
#include <QtCore/QDir>
#include <QtCore/QDebug>
#include <QtCore/QStringList>
#include <QtCore/QUrl>
#include <QtNetwork/QTcpSocket>
#include "TmHeaders/TCPSocket.h"
#include "TmHeaders/QTmGlobal.h"
#include "QTmUpnpControlPoint.h"
#include <QNetworkInterface>
#include <QTime>

#ifdef Q_OS_WIN
#include <winsock2.h>
#include <ws2tcpip.h>
#undef interface
#endif

#ifdef QT_UI_CLEANUP
#define Q_OS_LINUX
#endif
#ifdef Q_OS_LINUX
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <errno.h>
#include <net/if.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <netdb.h>
#ifndef QT_UI_CLEANUP
#include <ifaddrs.h>
#endif
#include <ctype.h>
#include <unistd.h>
#endif

//Types of UPnP Device Discovery Messages
#define SSDP_ALIVE	1
#define SSDP_BYEBYE	2
#define SSDP_UPDATE	3

//UPnP Terminal Mode
#define UPNP_TM_SERVER_DEVICE_TYPE "urn:schemas-upnp-org:device:TmServerDevice:1"

// Class constructor
#ifdef QT_UI_CLEANUP
QTmUpnpControlPoint::QTmUpnpControlPoint(QObject *parent) :
        QObject(parent),
        m_socket(0)
        ,m_url(0)
{}
#else
QTmUpnpControlPoint::QTmUpnpControlPoint(QObject *parent) :
        QObject(parent),
        m_socket(0)
{
    //Init network access manager

    connect(&m_netManager, SIGNAL(finished(QNetworkReply*)),
            this,          SLOT(readDeviceDescription(QNetworkReply *)),
            Qt::AutoConnection);
}
#endif

// Class Destructor
QTmUpnpControlPoint::~QTmUpnpControlPoint()
{
    //Close & Delete all sockets
    m_socketLock.lockForWrite();
    if (m_socket) {
#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "Stop listening for UPnP service discovery advertisements";
#endif
        m_socket->close();
        disconnect(m_socket, SIGNAL(readyRead()), this, SLOT(readDiscoveryMessage()));
        delete m_socket;
        m_socket = 0;
    }

    QHash<QHostAddress, QTmUpnpControlPointTx*>::iterator i = m_socketMap.begin();
    while (i != m_socketMap.end()) {
        delete i.value();
        i = m_socketMap.erase(i);
    }

#ifdef QT_DEBUG_TM_UPNP
    qDebug() << "QTmUpnpControlPoint:" << "Stop UPnP HTTP Server from receiving event notifications";
#endif
    if (m_eventingHttpServer.isListening())
        m_eventingHttpServer.close();
    m_socketLock.unlock();

    //Clear remote server devices
    m_serverLock.lockForWrite();
    QList<QTmRemoteServer*>::iterator j = m_serverList.begin();
    while (j != m_serverList.end()) {
        QTmRemoteServer *server = *j;
        j = m_serverList.erase(j);
        emit remoteServerDeviceRemoved(server);
        server->deleteLater();
    }
    m_serverLock.unlock();
}

bool QTmUpnpControlPoint::connectUpnp(QString interfaceName, QHostAddress interfaceAddress)
{
    //Create Multi-cast socket to listen for UPnP service discovery advertisements
    qDebug()<< "QTmUpnpControlPoint::connectUpnp(QString interfaceName, QHostAddress interfaceAddress)  :" << interfaceName << "," << interfaceAddress;
    if (!m_socket) {
#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "Start listen for UPnP service discovery advertisements";
#endif
        m_socket = new QUdpSocket();
        if (!m_socket->bind(1900, QUdpSocket::ReuseAddressHint)) {            
            qDebug() << "QTmUpnpControlPoint:" << "cannot bind socket to port 1900";
            delete m_socket;
            m_socket = 0;
            return false;
        }
       m_localAddress = interfaceAddress;
       connect(m_socket, SIGNAL(readyRead()), this,     SLOT (readDiscoveryMessage()), Qt::AutoConnection);

    }

    //Start HTTP Server of UPnP Control Point
    if (!m_eventingHttpServer.isListening()) {
#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "Starting UPnP HTTP Server (for receiving event notifications)";
#endif
        if (!m_eventingHttpServer.listen(QHostAddress::Any)) {
            qDebug() << "QTmUpnpControlPoint: cannot start HTTP Server";
            return false;
        }
    }

    //Check if already listening
    if (m_socketMap.contains(interfaceAddress)) {
        qDebug() << "QTmUpnpControlPoint: already listening to" << interfaceName << "(" << interfaceAddress.toString() << ")";
        sendMSearchQuery(interfaceAddress);
        //return false;
        return true;
    }

    //Setup UPnP multicasting
#ifdef QT_DEBUG_TM_UPNP
    qDebug() << "QTmUpnpControlPoint:" << "Starting UPnP Control Point" << "for" << interfaceName << "(" << interfaceAddress.toString() << ")";
#endif
    if (interfaceAddress.protocol() == QAbstractSocket::IPv4Protocol) {
        if(!setupUpnpIPv4(interfaceName, interfaceAddress)) {
            qDebug() << "QTmUpnpControlPoint: cannot start UPnP control point on IPv4 for" << interfaceName << "(" << interfaceAddress.toString() << ")";
            return false;
        }
    } else if (interfaceAddress.protocol() == QAbstractSocket::IPv6Protocol) {
        if(!setupUpnpIPv6(interfaceName, interfaceAddress)) {
            qDebug() << "QTmUpnpControlPoint: cannot start UPnP control point on IPv6 for" << interfaceName << "(" << interfaceAddress.toString() << ")";
            return false;
        }
    } else {
        qDebug() << "QTmUpnpControlPoint: protocol for" << interfaceName << "(" << interfaceAddress.toString() << "not supported";
        return false;
    }

#ifdef QT_DEBUG_TM_UPNP
    qDebug() << "QTmUpnpControlPoint:" << "Started UPnP Control Point for" << interfaceName << "(" << interfaceAddress.toString() << ")";
#endif
    return true;
}

void QTmUpnpControlPoint::disconnectUpnp(QHostAddress interfaceAddress)
{
    int emitFlag = false;
    //TODO leave multicast group

    //Close socket
    QTmUpnpControlPointTx *tx = m_socketMap.value(interfaceAddress);
    if (tx) {
        m_socketMap.remove(interfaceAddress);
        delete tx;
    }

    //Remove all devices within same network
    m_serverLock.lockForWrite();
    QList<QTmRemoteServer*>::iterator i = m_serverList.begin();
    while (i != m_serverList.end()) {
        QTmRemoteServer *server = *i;
        if (server->localAddress() == interfaceAddress) {

            if(server->isReadyReadPending())
            {
                server->forceStopReadyRead();
                qDebug() << "QTmUpnpControlPoint:" << "lets wait for remaining event processing !!";
                usleep(500*1000);
            }


#ifdef QT_DEBUG_TM_UPNP
            qDebug() << "QTmUpnpControlPoint:" << "Removed server device:" << server->USN();
#endif
            //Inform others
            emit remoteServerDeviceRemoved(server);

            emitFlag = true;

            //Delete device
            qDebug() << "QTmUpnpControlPoint:" << "deleting active QTmRemoteServer!!";
            server->deleteLater();
            i = m_serverList.erase(i);
        } else
            i++;
    }
    m_serverLock.unlock();

    if(emitFlag == false)
    {
        //perform dummy emit to ensure the cleanup at upper layers/managers on every ML phone disconnect.
        emit remoteServerDeviceRemoved(NULL);
    }
}

void QTmUpnpControlPoint::readDiscoveryMessage()
{
#ifdef QT_DEBUG_TM_UPNP
    qDebug() << "QTmUpnpControlPoint:" << "read discovery message";
#endif
    readMessage(m_socket);
}

void QTmUpnpControlPoint::readMessage(QUdpSocket *socket)
{
#ifdef QT_DEBUG_TM_UPNP
   qDebug() << "QTmUpnpControlPoint:" << "read message1" << socket->hasPendingDatagrams();
#endif

   while(socket->hasPendingDatagrams()) {
        //Read datagram
        QByteArray datagram;
        datagram.resize(socket->pendingDatagramSize());
        QHostAddress sender;
        quint16 senderPort;
        socket->readDatagram(datagram.data(), datagram.size(), &sender, &senderPort);

        //Parse datagram
        parseMessage(QString(datagram));
    }
}

void QTmUpnpControlPoint::parseMessage(const QString &buffer)
{
#ifdef QT_DEBUG_TM_UPNP
      qDebug() << "QTmUpnpControlPoint: Parse discovery message/response:\n" << buffer;
#endif

    //Parse header
    int msgType = 0;
    QString location, USN, devType;
    QStringList lines = buffer.split('\n');
    foreach( QString line, lines) {
        line = line.trimmed();
        if (line.startsWith("NTS:", Qt::CaseInsensitive)) {
            // Check content of NTS header
            if (line.contains("ssdp:alive", Qt::CaseInsensitive)) {
                msgType = SSDP_ALIVE;
            } else if (line.contains("ssdp:byebye", Qt::CaseInsensitive)) {
                msgType = SSDP_BYEBYE;
            } else if (line.contains("ssdp:update", Qt::CaseInsensitive)) {
                msgType = SSDP_UPDATE;
            } else {
                msgType = 0;
                return;
            }
        } else if (line.startsWith("NT:", Qt::CaseInsensitive)) {
            // Check content of NT header
            devType = line.remove("NT:", Qt::CaseInsensitive).trimmed();
        }
        else if (line.startsWith("ST:", Qt::CaseInsensitive)) {
            // Check content of ST header
            devType = line.remove("ST:", Qt::CaseInsensitive).trimmed();
            // If we find ST header then it means it is a response to a M-Search request
            // hence, set msgType equivalent to an SSDP service advertisement
            msgType = SSDP_ALIVE;
        }
        else if (line.startsWith("USN:", Qt::CaseInsensitive)) {
            // Check content of USN header
            USN = line.remove("USN:", Qt::CaseInsensitive).trimmed();
        } else if (line.startsWith("LOCATION:", Qt::CaseInsensitive)) {
            // Get content of Location header
            location = line.remove("LOCATION:", Qt::CaseInsensitive).trimmed();
        }
    }

    //Sanity check
    if (!msgType || USN.isEmpty() || devType.isEmpty())
    {
#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "Parse message/response failed!";
#endif
        return; // do not proceed any further
    }

    if ((msgType == SSDP_ALIVE || msgType == SSDP_UPDATE) && location.isEmpty())
    {
#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "Did not found the location entry!";
#endif
        return; // do not proceed any further
    }

    //Check whether the device is supported
    if (devType != UPNP_TM_SERVER_DEVICE_TYPE)
    {
#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "Unknown device type:" << devType;
#endif
        return; // do nothing
    }

    //Get device if already known
    m_serverLock.lockForRead();
    QTmRemoteServer *dev = remoteServerByUsn(USN);
    m_serverLock.unlock();

    //If discovery message is of right kind, then proceed
    switch(msgType) {
    case SSDP_ALIVE:
        {
            //Check whether the device is already known
            if (dev)
                return; // do nothing
#ifdef QT_DEBUG_TM_UPNP
            qDebug() << "QTmUpnpControlPoint:" << "SSDP advertisement/M-Search response received, from" << USN << "type:" << devType << "location:" << location;
#endif
            //Retrieve device description
#ifdef QT_UI_CLEANUP
            if(m_url)
            {
                delete m_url;
                m_url = 0;
            }
            m_url = new QUrl(location);
            #ifdef QT_DEBUG_TM_UPNP
            qDebug()<< " ************************   path :" +  m_url->path();
            #endif


            QString request = QString("GET %1 HTTP/1.1\r\nHOST: %2:%3\r\n\r\n").arg(m_url->path()).arg(m_url->host()).arg(m_url->port());
            QByteArray response;
            if(!performHTTPAction(m_url,request.toAscii(), response, TEXT_CONTENT))
                return;
            readDeviceDescription(response,m_url, m_socket->localAddress()) ;//QHostAddress("192.168.77.2"));
            //QString responseString = QString::fromAscii(response.constData());

            //qDebug()<<"Http response btained :\n"<<responseString<<"]";
#else
            qDebug()<< "#######    QNetworkRequest(QUrl(location):" + location;

            m_netManager.get(QNetworkRequest(QUrl(location)));
#endif

            break;
        }

    case SSDP_BYEBYE:
        {
#if 0 //this code is disabled to avoid unexpected 'DeviceRemoved' callback. This occures only in 20111024 img
            if (!dev)
                return; // do nothing
#ifdef QT_DEBUG_TM_UPNP
            qDebug() << "QTmUpnpControlPoint:" << "SSDP byebye received, from" << USN;
#endif
            //Remove from device list
            m_serverLock.lockForWrite();
            m_serverList.removeAll(dev);
            m_serverLock.unlock();
#ifdef QT_DEBUG_TM_UPNP
            qDebug() << "QTmUpnpControlPoint:" << "Removed server device:" << dev->USN();
#endif

            //Inform others
            emit remoteServerDeviceRemoved(dev);

            //Delete device
            dev->deleteLater();
#endif
            break;
        }

    case SSDP_UPDATE:
        {
#ifdef QT_DEBUG_TM_UPNP
            qDebug() << "QTmUpnpControlPoint:" << "SSDP update received, from" << USN << "type:" << devType;
#endif
            //TODO SSDP_UPDATE

            break;
        }
    }
}

bool QTmUpnpControlPoint::retryWait(QTcpSocket &socket, const QUrl &url, RetryWaitReason reason)
{
    bool connState = false;
    int errorId = -1;
    int retryCount = 0;
    while(1)
    {
        if(reason == RETRY_REASON_SOCKET_CONNECT)
            connState = socket.waitForConnected(SOCKET_READ_TIMEOUT);
        else
            connState = socket.waitForReadyRead(SOCKET_READ_TIMEOUT);

        if(connState == false)
        {
            errorId = socket.error();
            qDebug() << "QTerminalModeUPnPClient:" << "error" << socket.errorString();
            if(errorId != QAbstractSocket::SocketTimeoutError)
            {
                qDebug() << "QTerminalModeUPnPClient:" << "Failed to connect to " << (url) << "error" << socket.errorString();
                qDebug() << "QTerminalModeUPnPClient:" << "closing HTTP session";
                socket.close();
                qDebug() << "QTerminalModeUPnPClient:" << "closing eventing Http Server socket";
                m_eventingHttpServer.close();
                return false;
            }
            else
            {
                qDebug() << "QTerminalModeUPnPClient:" << "Failed to connect to "  << (url) << "Timed Out ..Retrying !!";
                retryCount++;
                if(retryCount == MAX_SOCKET_RETRY_CNT)
                {
                    qDebug() << "QTerminalModeUPnPClient:" << "stopping retry !!";
                    qDebug() << "QTerminalModeUPnPClient:" << "Failed to connect to " << (url) << "error" << socket.errorString();
                    qDebug() << "QTerminalModeUPnPClient:" << "closing HTTP session";
                    m_eventingHttpServer.close();
                    socket.close();
                    return false;
                }
            }
        }
        else
        {
            qDebug() << "QTerminalModeUPnPClient:" << "Connected to " << (url);
            break;
        }
    }//while(1)
}

#define QTCP_SOCKET_IMPL
#ifdef QT_UI_CLEANUP
int QTmUpnpControlPoint::performHTTPAction(QUrl* url,const QByteArray& sendBuffer, QByteArray& receiveBuffer, int contentType)
{

#ifdef QTCP_SOCKET_IMPL


    QTcpSocket socket;

    QTime timer;

    timer.start();
    qDebug() << "QTmUpnpControlPoint:  performHTTPAction() start" << "timer start !!";

    socket.connectToHost(url->host(), url->port());

    qDebug() << "QTmUpnpControlPoint:  performHTTPAction()" << "connectToHost: done" << timer.elapsed() << "msec";

    socket.setSocketOption(QAbstractSocket::LowDelayOption, 1);

    if(retryWait(socket, (*url), RETRY_REASON_SOCKET_CONNECT) == false)
    {
        qDebug() << "QTerminalModeUPnPClient:" << "Failed to connect to Remote UPnP HTTP Server" << url->host() << ":" << url->port();
        return false;
    }

    qDebug() << "QTmUpnpControlPoint: performHTTPAction()" << "waitForConnected: done" << timer.elapsed() << "msec";

    qDebug() << "QTmUpnpControlPoint: writing the data on socket ..";
    if(!socket.write(sendBuffer))
    {
        qDebug() << "QTmUpnpControlPoint:" << "Failed to connect write data $";
        //Close HTTP session
        qDebug() << "QTerminalModeUPnPClient:" << "closing HTTP session";
        socket.close();
        qDebug() << "QTerminalModeUPnPClient:" << "closing eventing Http Server socket";
        m_eventingHttpServer.close();
        return false;
    }
    qDebug() << "QTmUpnpControlPoint: wait for write complete ..";
    if(!socket.waitForBytesWritten(SOCKET_READ_TIMEOUT))
    {
        qDebug() << "QTmUpnpControlPoint:" << "Failed to connect write data #";
        //Close HTTP session
        qDebug() << "QTerminalModeUPnPClient:" << "closing HTTP session";
        socket.close();
        qDebug() << "QTerminalModeUPnPClient:" << "closing eventing Http Server socket";
        m_eventingHttpServer.close();
        return false;
    }

    //While there's data, read and store it in receivedBuffer
    int headerLength = -1;
    long totalLength = -1;
    bool valid = false;
    //while (socket.waitForReadyRead(SOCKET_READ_TIMEOUT))
    while(1)
    {
        if(retryWait(socket, (*url), RETRY_REASON_SOCKET_READYREAD) == false)
        {
            qDebug() << "QTerminalModeUPnPClient:" << "Failed to connect to Remote UPnP HTTP Server" << url->host() << ":" << url->port();
            return false;
        }

        while (socket.bytesAvailable())
        {
            if (totalLength == -1)
                receiveBuffer += socket.readAll();
            else
                receiveBuffer += socket.read(totalLength - receiveBuffer.length());
        }

        // Parse header
        if (headerLength == -1) {
            //Find header
            int index = receiveBuffer.indexOf("\r\n\r\n");
            if (index != -1) {
                headerLength = index + 4; // skip "\r\n\r\n"
                //Parse header lines
                QString header = receiveBuffer.mid(0, index);
                QStringList lines = header.split('\n');
                foreach( QString line, lines) {
                    line = line.trimmed();
                    if (line.startsWith("HTTP/", Qt::CaseInsensitive)) {
                        if (!line.contains("200")) {
                            qDebug() << "QTmUpnpControlPoint:" << "HTTP request failed!\"" << line << "\"";
                            qDebug() << "QTerminalModeUPnPClient:" << "closing HTTP session";
                            socket.close();
                            qDebug() << "QTerminalModeUPnPClient:" << "closing eventing Http Server socket";
                            m_eventingHttpServer.close();
                            return false;
                        } else {
                            valid = true;
                        }
                    } else if (line.startsWith("CONTENT-LENGTH:", Qt::CaseInsensitive)) {
                        long contentLength = line.remove("CONTENT-LENGTH:", Qt::CaseInsensitive).trimmed().toLong();
                        totalLength = contentLength + headerLength;
                        qDebug() << "QTmUpnpControlPoint:" << "content-length:" << contentLength << "total-length:" << totalLength;
                    }
                }
            }
        }

        // Check conetent length
        if (totalLength != -1 && receiveBuffer.length() >= totalLength) {
            break;
        }


        qDebug() << "QTmUpnpControlPoint, mod, sleep 200";
        usleep(200);
    }

    //Close HTTP session
    socket.close();
    return valid;

#else //QTCP_SOCKET_IMPL

    TCPSocket  socket;

    QTime timer;

    timer.start();
    qDebug() << "QTmUpnpControlPoint:  performHTTPAction() start" << "timer start !!";

    qDebug() << "QTmUpnpControlPoint:  performHTTPAction() ip" << url->host().toLocal8Bit().data() << "port:" << url->port();
    socket.connectToIp(url->host().toLocal8Bit().data(), url->port());

    qDebug() << "QTmUpnpControlPoint:  performHTTPAction()" << "connectToHost: done" << timer.elapsed() << "msec";

    //socket.setSocketOption(QAbstractSocket::LowDelayOption, 1);
    qDebug() << "QTmUpnpControlPoint: performHTTPAction()" << "Bytes to write:" << sendBuffer.length();
    socket.socketWrite((void*)sendBuffer.data(), sendBuffer.length());
    qDebug() << "QTmUpnpControlPoint: performHTTPAction()" << "waitForBytesWritten: done" << timer.elapsed() << "msec";

    //While there's data, read and store it in receivedBuffer
    int ret = 0;
    int  bytes_received;
    //int total_bytes_received = 0;

    usleep(500);

    ret = socket.socketReadOnce(receiveBuffer, &bytes_received);
    if(ret != 0)
    {
        qDebug() << "Read error !!";
    }
    /*
    int no_data_cnt = 0;
    while(timer.elapsed() < 500) {
        bytes_received = 0;
        ret = socket.socketReadOnce(receiveBuffer, &bytes_received);
        if (ret != 0) {
            qDebug() << "QTmUpnpControlPoint::performHTTPAction(), socketRead() reurned non-zero value";
        }
        if (bytes_received == 0) {
            // increment no data counter
            no_data_cnt++;
        }
        total_bytes_received += bytes_received;
        if (no_data_cnt > 4) break;
        usleep(25);
    }
    */
    qDebug() << "QTmUpnpControlPoint: performHTTPAction()" << "Bytes read1:" << bytes_received;
    qDebug() << "QTmUpnpControlPoint: performHTTPAction()" << "socket.readAll(): done" << timer.elapsed() << "msec";

    qDebug() << "QTmUpnpControlPoint:  performHTTPAction() stop" << "timer stop !!";

    //Close HTTP session
    socket.closeSocket();

    return 1;

#endif //QTCP_SOCKET_IMPL
}


void QTmUpnpControlPoint::readDeviceDescription(QByteArray &response, QUrl* url,QHostAddress lAddress)
{
    QDomDocument doc;
    QString errorString;
    int errorLine;

    //Parse XML into struct for processing
    QString responseString;
    responseString.reserve(3096);
    responseString = QString::fromAscii(response.constData());

    qDebug()<< "readDeviceDescription : " ;
    qDebug()<< "About to parse responseString:[" + responseString <<"]" ;

    QString strRoot(ROOT_NODE);
    int rootIndex = responseString.indexOf( strRoot);
    if(rootIndex < 0)
        return;

    responseString = responseString.mid(rootIndex - 1, responseString.length() - strRoot.length());
    //qDebug()<< "About to parse :[" + responseString <<"]" ;
    errorString.reserve(3096);
    if (!doc.setContent(responseString, false, &errorString, &errorLine)) {
      qDebug() << "QTmUpnpControlPoint:" << "XML format error";
      qDebug() << "Error :" << errorString << "line - " << errorLine ;
      return;
    }

    QDomElement specVersionElement = doc.documentElement().firstChildElement("specVersion");
    QDomElement minorVersionElement;
    QDomElement majorVersionElement;
    if(!specVersionElement.isNull())
    {
        minorVersionElement = specVersionElement.firstChildElement("minor");
        majorVersionElement = specVersionElement.firstChildElement("major");
    }


    QDomNodeList deviceNodes = doc.elementsByTagName("device");
    if (deviceNodes.isEmpty()) {
#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "Error: No device found within root device description.";
#endif
        return;
    }

    for(int i = 0; ; i++) {
        QDomNode deviceNode = deviceNodes.at(i);
        if(deviceNode.isNull())
            break;

        //Get device type & id
        QDomElement devTypeElement = deviceNode.firstChildElement("deviceType");
        QDomElement udnElement = deviceNode.firstChildElement("UDN");
        if (udnElement.isNull() || devTypeElement.isNull())
            continue;
        QString USN = udnElement.text() + "::" + devTypeElement.text();

        //Check whether the device is supported
        if (devTypeElement.text() != UPNP_TM_SERVER_DEVICE_TYPE)
            continue;

        //Check whether server device
        m_serverLock.lockForWrite();
        QTmRemoteServer *tmServerDevice = remoteServerByUsn(USN);
        if (tmServerDevice) {
            m_serverLock.unlock();
            continue; // server device already known
        }

#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "New server device:" << USN;
#endif
        //Init new server device
        tmServerDevice = new QTmRemoteServer(USN, m_localAddress, (*url), &m_eventingHttpServer);
        tmServerDevice->setUDN(udnElement.text());

        QDomElement friendlyNameElement = deviceNode.firstChildElement("friendlyName");
        QDomElement manufacturerNameElement = deviceNode.firstChildElement("manufacturer");
        QDomElement manufacturerUrlElement = deviceNode.firstChildElement("manufacturerURL");
        QDomElement modelNameElement = deviceNode.firstChildElement("modelName");
        QDomElement modelDescriptionElement = deviceNode.firstChildElement("modelDescription");
        QDomElement modelNumberElement = deviceNode.firstChildElement("modelNumber");
        QDomElement modelUrlElement = deviceNode.firstChildElement("modelURL");
        QDomElement serialNumberElement = deviceNode.firstChildElement("serialNumber");
        QDomElement upcElement = deviceNode.firstChildElement("UPC");

        if(!minorVersionElement.isNull())
            tmServerDevice->setTmSpecVersionMinor(minorVersionElement.text().toInt());

        if(!majorVersionElement.isNull())
            tmServerDevice->setTmSpecVersionMinor(majorVersionElement.text().toInt());

        if(!friendlyNameElement.isNull())
            tmServerDevice->setFriendlyName(friendlyNameElement.text());

        if(!manufacturerNameElement.isNull())
            tmServerDevice->setManufacturerName(manufacturerNameElement.text());

        if(!manufacturerUrlElement.isNull())
            tmServerDevice->setManufacturerUrl(QUrl(manufacturerUrlElement.text()));

        if(!modelNameElement.isNull())
            tmServerDevice->setModelName(modelNameElement.text());

        if(!modelDescriptionElement.isNull())
            tmServerDevice->setModelDescription(modelDescriptionElement.text());

        if(!modelNumberElement.isNull())
            tmServerDevice->setModelNumber(modelNumberElement.text());

        if(!modelUrlElement.isNull())
            tmServerDevice->setModelUrl(QUrl(modelUrlElement.text()));

        if(!serialNumberElement.isNull())
            tmServerDevice->setSerialNumber(serialNumberElement.text());

        if(!upcElement.isNull())
            tmServerDevice->setUPC(upcElement.text());
#ifdef FETCH_APP_ICON_LIST
        QList<QTmIcon *> iconList;
        QDomElement iconListElement = deviceNode.firstChildElement("iconList");
        if(!iconListElement.isNull())
        {
            QDomElement iconElement = iconListElement.firstChildElement("icon");
            if(!iconElement.isNull())
            {
                QDomElement iconUrlElement = iconElement.firstChildElement("url");
                QDomElement iconMimeTypeElement = iconElement.firstChildElement("mimetype");
                QDomElement iconHeightElement = iconElement.firstChildElement("height");
                QDomElement iconWidthElement = iconElement.firstChildElement("width");
                QDomElement iconDepthElement = iconElement.firstChildElement("depth");
                QTmIcon *icon = new QTmIcon();
                if(!iconWidthElement.isNull())
                    icon->setWidth(iconWidthElement.text().toInt());
                if(!iconHeightElement.isNull())
                    icon->setHeight(iconHeightElement.text().toInt());
                if(!iconDepthElement.isNull())
                    icon->setDepth(iconDepthElement.text().toInt());
                if(!iconMimeTypeElement.isNull())
                    icon->setMimeType(iconMimeTypeElement.text());
                if(!iconUrlElement.isNull())
                {
                    //Retrieve and cache the application icon
                    QUrl cacheUrl;
                    QUrl iconUrl = url->resolved(QUrl(iconUrlElement.text()));
                    if (tmServerDevice->retrieveIcon(iconUrl, cacheUrl))
                        icon->setUrl(cacheUrl);
                    else
                        icon->setUrl(iconUrl);
                }
                tmServerDevice->iconList().append(icon);
            }
        }
        tmServerDevice->setIconList(iconList);
#endif  //FETCH_APP_ICON_LIST
        QDomElement connElement = deviceNode.firstChildElement("X_connectivity");
        QDomElement btElement, bdAddrElement, startConnElement;
        if(!connElement.isNull())
            btElement = connElement.firstChildElement("bluetooth");
        if(!btElement.isNull())
        {
            bdAddrElement = btElement.firstChildElement("bdAddr");
            startConnElement = btElement.firstChildElement("startConnection");
        }

        if(!bdAddrElement.isNull())
            tmServerDevice->setServerBTAddress(bdAddrElement.text());
        if(!startConnElement.isNull())
        {
            if(!startConnElement.text().compare("true", Qt::CaseInsensitive))
                tmServerDevice->setCanServerStartBT(true);
            else
                tmServerDevice->setCanServerStartBT(false);
        }

        //Parse services
        QDomElement serviceListElement = deviceNode.firstChildElement("serviceList");
        if (serviceListElement.isNull() || serviceListElement.parentNode() != deviceNode) {
            qDebug() << "QTmUpnpControlPoint:" << "Error: No service list found within root device description.";
            delete tmServerDevice;
            m_serverLock.unlock();
            continue;
        }
        QDomNodeList serviceNodes = serviceListElement.elementsByTagName("service");
        if (serviceNodes.isEmpty()) {
            qDebug() << "QTmUpnpControlPoint:" << "Error: No services found within service list.";
            delete tmServerDevice;
            m_serverLock.unlock();
            continue;
        }

        for (int j = 0; ; j++) {
            //Parse service
            QDomNode serviceNode = serviceNodes.at(j);
            if (serviceNode.isNull())
                break;

            QDomElement serviceTypeElement = serviceNode.firstChildElement("serviceType");
            QDomElement serviceIdElement   = serviceNode.firstChildElement("serviceId");
            QDomElement scpdUrlElement     = serviceNode.firstChildElement("SCPDURL");
            QDomElement controlUrlElement  = serviceNode.firstChildElement("controlURL");
            QDomElement eventSubUrlElement = serviceNode.firstChildElement("eventSubURL");
            //Check required elements
            if (serviceTypeElement.isNull() || serviceIdElement.isNull() || scpdUrlElement.isNull() || controlUrlElement.isNull())
                continue;

            //TODO QUrl::resolved ( const QUrl & relative ) const
            QTmRemoteServer::QUpnpService service;
            service.serviceType     = serviceTypeElement.text();
            service.serviceId       = serviceIdElement.text();
            service.scdpURL         = url->resolved(QUrl(scpdUrlElement.text()));
            service.controlUrl      = url->resolved(QUrl(controlUrlElement.text()));
            if (!eventSubUrlElement.isNull())
                service.eventSubUrl = url->resolved(QUrl(eventSubUrlElement.text()));
#ifdef QT_DEBUG_TM_UPNP
            qDebug() << "QTmUpnpControlPoint:" << "Retrieving Service details...";
            qDebug() << "QTmUpnpControlPoint:" << "serviceType:" << service.serviceType;
            qDebug() << "QTmUpnpControlPoint:" << "serviceId:" << service.serviceId;
            qDebug() << "QTmUpnpControlPoint:" << "SCPDURL:" << service.scdpURL.toString();
            qDebug() << "QTmUpnpControlPoint:" << "controlURL:" << service.controlUrl.toString();
            qDebug() << "QTmUpnpControlPoint:" << "eventSubURL:" << service.eventSubUrl.toString();
#endif
            //Store service
            tmServerDevice->setService(service);
        }

        //TODO Verifying UPnP Service is Terminal Mode compliant

        //Add to server device list
        m_serverList.append(tmServerDevice);
        m_serverLock.unlock();
#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "Added server device:" << tmServerDevice->USN();
#endif

        //Inform others
        emit remoteServerDeviceAdded(tmServerDevice);

#ifndef QT_UI_CLEANUP_ENABLE_N97
        //TODO to be moved to TmClient
        //Subscribe to events from TmApplicationServer:1 service
#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "Subscribing to Events for TmApplicationServer:1 service";
#endif
        if(!tmServerDevice->subscribeToTmApplicationServerEvents())
            qDebug() << "QTmUpnpControlPoint:" << "Error subscribing to events for TmApplicationServer:1 service";


        //Subscribe to events from TmClientProfile:1 service
#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "Subscribing to Events for TmClientProfile:1 service";
#endif
        if(!tmServerDevice->subscribeToTmClientProfileEvents())
            qDebug() << "QTmUpnpControlPoint:" << "Error subscribing to events for TmClientProfile:1 service";
#endif //QT_UI_CLEANUP_ENABLE_N97

    }
}
#else
void QTmUpnpControlPoint::readDeviceDescription(QNetworkReply *reply)
{
    if (reply->error() != QNetworkReply::NoError) {
        qDebug() << "QTmUpnpControlPoint:" << "Error retrieving device description";
        return;
    }

    QUrl url = reply->url();
    qDebug()<<"******   *****   *****  QTmUpnpControlPoint::readDeviceDescription() : url:" << url;
#ifdef QT_DEBUG_TM_UPNP
    qDebug() << "QTmUpnpControlPoint:" << "Parsing UPnP device description from location:" << url.path();
#endif

    QDomDocument doc;
    //Parse XML into struct for processing
    if (!doc.setContent(reply)) {
        qDebug() << "QTmUpnpControlPoint:" << "XML format error";
        return;
    }

    QDomElement specVersionElement = doc.documentElement().firstChildElement("specVersion");
    QDomElement minorVersionElement;
    QDomElement majorVersionElement;
    if(!specVersionElement.isNull())
    {
        minorVersionElement = specVersionElement.firstChildElement("minor");
        majorVersionElement = specVersionElement.firstChildElement("major");
    }

    QDomElement urlBaseElement = doc.documentElement().firstChildElement("URLBase");
    if(!urlBaseElement.isNull())
        url = QUrl(urlBaseElement.text());
    qDebug()<<"******   *****   *****  QTmUpnpControlPoint::readDeviceDescription() :url2:" << url;

    QDomNodeList deviceNodes = doc.elementsByTagName("device");
    if (deviceNodes.isEmpty()) {
#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "Error: No device found within root device description.";
#endif
        return;
    }

    QHostAddress localAddress = getLocalAddress(QHostAddress(url.host()));
    qDebug()<<"******   *****   *****  QTmUpnpControlPoint::readDeviceDescription() : localAddress:" << localAddress;
    for(int i = 0; ; i++) {
        QDomNode deviceNode = deviceNodes.at(i);
        if(deviceNode.isNull())
            break;

        //Get device type & id
        QDomElement devTypeElement = deviceNode.firstChildElement("deviceType");
        QDomElement udnElement = deviceNode.firstChildElement("UDN");
        if (udnElement.isNull() || devTypeElement.isNull())
            continue;
        QString USN = udnElement.text() + "::" + devTypeElement.text();

        //Check whether the device is supported
        if (devTypeElement.text() != UPNP_TM_SERVER_DEVICE_TYPE)
            continue;

        //Check whether server device
        m_serverLock.lockForWrite();
        QTmRemoteServer *tmServerDevice = remoteServerByUsn(USN);
        if (tmServerDevice) {
            m_serverLock.unlock();
            continue; // server device already known
        }

#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "New server device:" << USN;
#endif
        //Init new server device
        tmServerDevice = new QTmRemoteServer(USN, localAddress, url, &m_eventingHttpServer);
        tmServerDevice->setUDN(udnElement.text());

        QDomElement friendlyNameElement = deviceNode.firstChildElement("friendlyName");
        QDomElement manufacturerNameElement = deviceNode.firstChildElement("manufacturer");
        QDomElement manufacturerUrlElement = deviceNode.firstChildElement("manufacturerURL");
        QDomElement modelNameElement = deviceNode.firstChildElement("modelName");
        QDomElement modelDescriptionElement = deviceNode.firstChildElement("modelDescription");
        QDomElement modelNumberElement = deviceNode.firstChildElement("modelNumber");
        QDomElement modelUrlElement = deviceNode.firstChildElement("modelURL");
        QDomElement serialNumberElement = deviceNode.firstChildElement("serialNumber");
        QDomElement upcElement = deviceNode.firstChildElement("UPC");

        if(!minorVersionElement.isNull())
            tmServerDevice->setTmSpecVersionMinor(minorVersionElement.text().toInt());

        if(!majorVersionElement.isNull())
            tmServerDevice->setTmSpecVersionMinor(majorVersionElement.text().toInt());

        if(!friendlyNameElement.isNull())
            tmServerDevice->setFriendlyName(friendlyNameElement.text());

        if(!manufacturerNameElement.isNull())
            tmServerDevice->setManufacturerName(manufacturerNameElement.text());

        if(!manufacturerUrlElement.isNull())
            tmServerDevice->setManufacturerUrl(QUrl(manufacturerUrlElement.text()));

        if(!modelNameElement.isNull())
            tmServerDevice->setModelName(modelNameElement.text());

        if(!modelDescriptionElement.isNull())
            tmServerDevice->setModelDescription(modelDescriptionElement.text());

        if(!modelNumberElement.isNull())
            tmServerDevice->setModelNumber(modelNumberElement.text());

        if(!modelUrlElement.isNull())
            tmServerDevice->setModelUrl(QUrl(modelUrlElement.text()));

        if(!serialNumberElement.isNull())
            tmServerDevice->setSerialNumber(serialNumberElement.text());

        if(!upcElement.isNull())
            tmServerDevice->setUPC(upcElement.text());

        QList<QTmIcon *> iconList;
        QDomElement iconListElement = deviceNode.firstChildElement("iconList");
        if(!iconListElement.isNull())
        {
            QDomElement iconElement = iconListElement.firstChildElement("icon");
            if(!iconElement.isNull())
            {
                QDomElement iconUrlElement = iconElement.firstChildElement("url");
                QDomElement iconMimeTypeElement = iconElement.firstChildElement("mimetype");
                QDomElement iconHeightElement = iconElement.firstChildElement("height");
                QDomElement iconWidthElement = iconElement.firstChildElement("width");
                QDomElement iconDepthElement = iconElement.firstChildElement("depth");
                QTmIcon *icon = new QTmIcon();
                if(!iconWidthElement.isNull())
                    icon->setWidth(iconWidthElement.text().toInt());
                if(!iconHeightElement.isNull())
                    icon->setHeight(iconHeightElement.text().toInt());
                if(!iconDepthElement.isNull())
                    icon->setDepth(iconDepthElement.text().toInt());
                if(!iconMimeTypeElement.isNull())
                    icon->setMimeType(iconMimeTypeElement.text());
                if(!iconUrlElement.isNull())
                {
                    //Retrieve and cache the application icon
                    QUrl cacheUrl;
                    QUrl iconUrl = url.resolved(QUrl(iconUrlElement.text()));
                    if (tmServerDevice->retrieveIcon(iconUrl, cacheUrl))
                        icon->setUrl(cacheUrl);
                    else
                        icon->setUrl(iconUrl);
                }
                tmServerDevice->iconList().append(icon);
            }
        }
        tmServerDevice->setIconList(iconList);

        QDomElement connElement = deviceNode.firstChildElement("X_connectivity");
        QDomElement btElement, bdAddrElement, startConnElement;
        if(!connElement.isNull())
            btElement = connElement.firstChildElement("bluetooth");
        if(!btElement.isNull())
        {
            bdAddrElement = btElement.firstChildElement("bdAddr");
            startConnElement = btElement.firstChildElement("startConnection");
        }

        if(!bdAddrElement.isNull())
            tmServerDevice->setServerBTAddress(bdAddrElement.text());
        if(!startConnElement.isNull())
        {
            if(!startConnElement.text().compare("true", Qt::CaseInsensitive))
                tmServerDevice->setCanServerStartBT(true);
            else
                tmServerDevice->setCanServerStartBT(false);
        }

        //Parse services
        QDomElement serviceListElement = deviceNode.firstChildElement("serviceList");
        if (serviceListElement.isNull() || serviceListElement.parentNode() != deviceNode) {
            qDebug() << "QTmUpnpControlPoint:" << "Error: No service list found within root device description.";
            delete tmServerDevice;
            m_serverLock.unlock();
            continue;
        }
        QDomNodeList serviceNodes = serviceListElement.elementsByTagName("service");
        if (serviceNodes.isEmpty()) {
            qDebug() << "QTmUpnpControlPoint:" << "Error: No services found within service list.";
            delete tmServerDevice;
            m_serverLock.unlock();
            continue;
        }

        for (int j = 0; ; j++) {
            //Parse service
            QDomNode serviceNode = serviceNodes.at(j);
            if (serviceNode.isNull())
                break;

            QDomElement serviceTypeElement = serviceNode.firstChildElement("serviceType");
            QDomElement serviceIdElement   = serviceNode.firstChildElement("serviceId");
            QDomElement scpdUrlElement     = serviceNode.firstChildElement("SCPDURL");
            QDomElement controlUrlElement  = serviceNode.firstChildElement("controlURL");
            QDomElement eventSubUrlElement = serviceNode.firstChildElement("eventSubURL");
            //Check required elements
            if (serviceTypeElement.isNull() || serviceIdElement.isNull() || scpdUrlElement.isNull() || controlUrlElement.isNull())
                continue;

            //TODO QUrl::resolved ( const QUrl & relative ) const
            QTmRemoteServer::QUpnpService service;
            service.serviceType     = serviceTypeElement.text();
            service.serviceId       = serviceIdElement.text();
            service.scdpURL         = url.resolved(QUrl(scpdUrlElement.text()));
            service.controlUrl      = url.resolved(QUrl(controlUrlElement.text()));
            if (!eventSubUrlElement.isNull())
                service.eventSubUrl = url.resolved(QUrl(eventSubUrlElement.text()));
#ifdef QT_DEBUG_TM_UPNP
            qDebug() << "QTmUpnpControlPoint:" << "Retrieving Service details...";
            qDebug() << "QTmUpnpControlPoint:" << "serviceType:" << service.serviceType;
            qDebug() << "QTmUpnpControlPoint:" << "serviceId:" << service.serviceId;
            qDebug() << "QTmUpnpControlPoint:" << "SCPDURL:" << service.scdpURL.toString();
            qDebug() << "QTmUpnpControlPoint:" << "controlURL:" << service.controlUrl.toString();
            qDebug() << "QTmUpnpControlPoint:" << "eventSubURL:" << service.eventSubUrl.toString();
#endif
            //Store service
            tmServerDevice->setService(service);
        }

        //TODO Verifying UPnP Service is Terminal Mode compliant

        //Add to server device list
        m_serverList.append(tmServerDevice);
        m_serverLock.unlock();
#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "Added server device:" << tmServerDevice->USN();
#endif

        //Inform others
        emit remoteServerDeviceAdded(tmServerDevice);

#ifndef QT_UI_CLEANUP_ENABLE_N97
        //TODO to be moved to TmClient
        //Subscribe to events from TmApplicationServer:1 service
#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "Subscribing to Events for TmApplicationServer:1 service";
#endif
        if(!tmServerDevice->subscribeToTmApplicationServerEvents())
            qDebug() << "QTmUpnpControlPoint:" << "Error subscribing to events for TmApplicationServer:1 service";


        //Subscribe to events from TmClientProfile:1 service
#ifdef QT_DEBUG_TM_UPNP
        qDebug() << "QTmUpnpControlPoint:" << "Subscribing to Events for TmClientProfile:1 service";
#endif
        if(!tmServerDevice->subscribeToTmClientProfileEvents())
            qDebug() << "QTmUpnpControlPoint:" << "Error subscribing to events for TmClientProfile:1 service";
#endif //QT_UI_CLEANUP_ENABLE_N97

    }
}
#endif



// Set up multi-cast listener to listen for UPnP SSDP advertisements
// Returns 1 for success, 0 for failure
int QTmUpnpControlPoint::setupUpnpIPv4(QString interfaceName, QHostAddress interfaceAddress)
{
    qDebug() << "QTmUpnpControlPoint:" << "Setup Upnp IPv4 on" << interfaceName << "(" << interfaceAddress.toString() << ")";

    //Join UPnP Device Discovery multicast group
    struct ip_mreq group;
    memset(&group, 0, sizeof(group));
    group.imr_multiaddr.s_addr = inet_addr("239.255.255.250");
    group.imr_interface.s_addr = inet_addr(interfaceAddress.toString().toAscii().constData());

    qDebug() << "QTmUpnpControlPoint:" << "Interface Address is:" << inet_ntoa(group.imr_interface);

    int err = setsockopt(m_socket->socketDescriptor(), IPPROTO_IP, IP_ADD_MEMBERSHIP, (const char *) &group, sizeof(group));
    if (err < 0) {
        qDebug() << "QTmUpnpControlPoint:" << "Adding multicast group error (" << err << ")";
        perror("QTmUpnpControlPoint: IP_ADD_MEMBERSHIP");
        return 0;
    }

    //Create Multi-cast socket to send UPnP M-SEARCH multicast queries
    QUdpSocket *socketTx = new QUdpSocket();
    if (!socketTx->bind()) {
            qDebug() << "QTmUpnpControlPoint:" << "Cannot bind socket to send UPnP M-SEARCH multicast queries";
            delete socketTx;
            return 0;
    }

    //Disable MC loopback
    char loopch = 0;
    err = setsockopt(socketTx->socketDescriptor(), IPPROTO_IP, IP_MULTICAST_LOOP, (char *)&loopch, sizeof(loopch));
    if (err < 0) {
         qDebug() << "QTmUpnpControlPoint:" << "Disabling multicast loopback error (" << err << ")";
         perror("QTmUpnpControlPoint: IP_MULTICAST_LOOP");
         socketTx->close();
         delete socketTx;
         return 0;
    }

    //Set local interface for outbound multicast datagrams
    //The IP address specified must be associated with the preferred interface
    struct in_addr localInterface;
    memset(&localInterface, 0, sizeof(localInterface));
    localInterface.s_addr = inet_addr(interfaceAddress.toString().toAscii().constData());
    err = setsockopt(socketTx->socketDescriptor(), IPPROTO_IP, IP_MULTICAST_IF, (char *)&localInterface, sizeof(localInterface));
    if (err < 0) {
         qDebug() << "QTmUpnpControlPoint:" << "Error while setting local network interface (" << err << ")";
         perror("QTmUpnpControlPoint: IP_MULTICAST_IF");
         socketTx->close();
         delete socketTx;
         return 0;
    }
#ifdef QT_DEBUG_TM_UPNP
      else
         qDebug() << "QTmUpnpControlPoint:" << "Setting the local interface for outbound datagrams on: " << interfaceAddress.toString() << "...OK.\n";
#endif

    //Add to transceiver list
    m_socketMap.insert(interfaceAddress, new QTmUpnpControlPointTx(this, socketTx));

    return 1;
}

// Set up multi-cast listener to listen for UPnP SSDP advertisements
// Returns 1 for success, 0 for failure
int QTmUpnpControlPoint::setupUpnpIPv6(QString interfaceName, QHostAddress interfaceAddress)
{
    qDebug() << "QTmUpnpControlPoint:" << "Setup Upnp IPv6 on" << interfaceName << "(" << interfaceAddress.toString() << ")";

    //Join UPnP Device Discovery multicast group
    struct ipv6_mreq group;
    memset(&group, 0, sizeof(group));
    inet_pton(AF_INET6, (const char*)"FF02::C", group.ipv6mr_multiaddr.s6_addr);

    //Obtain index associated with network interface
    group.ipv6mr_interface = if_nametoindex(interfaceName.toAscii().constData());
    //If index cannot be found, exit
    if(group.ipv6mr_interface == 0)
        return 0;

    qDebug() << "QTmUpnpControlPoint:" << "Interface Address is:" << interfaceAddress.toString();

    int err = setsockopt(m_socket->socketDescriptor(), IPPROTO_IPV6, IPV6_JOIN_GROUP, (const char *) &group, sizeof(group));
    if (err < 0) {
        qDebug() << "QTmUpnpControlPoint:" << "Adding IPv6 multicast group error (" << err << ")";
        perror("QTmUpnpControlPoint: IPV6_JOIN_GROUP");
        return 0;
    }

    //Create Multi-cast socket to send UPnP M-SEARCH multicast queries
    QUdpSocket *socketTx = new QUdpSocket();
    if (!socketTx->bind()) {
            qDebug() << "QTmUpnpControlPoint:" << "Cannot bind socket to send UPnP M-SEARCH multicast queries";
            delete socketTx;
            return 0;
    }

    //Disable MC loopback
    char loopch = 0;
    err = setsockopt(socketTx->socketDescriptor(), IPPROTO_IPV6, IPV6_MULTICAST_LOOP, (char *)&loopch, sizeof(loopch));
    if (err < 0) {
         qDebug() << "QTmUpnpControlPoint:" << "Disabling IPv6 multicast loopback error (" << err << ")";
         perror("QTmUpnpControlPoint: IPV6_MULTICAST_LOOP");
         socketTx->close();
         delete socketTx;
         return 0;
    }

    //Set local interface for outbound multicast datagrams
    //The IP address specified must be associated with the preferred interface
    struct in6_addr localInterface;
    memset(&localInterface, 0, sizeof(localInterface));
    inet_pton(AF_INET6, interfaceAddress.toString().toAscii().constData(), localInterface.s6_addr);
    err = setsockopt(socketTx->socketDescriptor(), IPPROTO_IPV6, IPV6_MULTICAST_IF, (char *)&localInterface, sizeof(localInterface));
    if (err < 0) {
         qDebug() << "QTmUpnpControlPoint:" << "Error while setting IPv6 local network interface (" << err << ")";
         perror("QTmUpnpControlPoint: IPV6_MULTICAST_IF");
         socketTx->close();
         delete socketTx;
         return 0;
    }
#ifdef QT_DEBUG_TM_UPNP
      else
         qDebug() << "QTmUpnpControlPoint:" << "Setting the local interface for outbound IPv6 datagrams on: " << interfaceAddress.toString() << "...OK.\n";
#endif

    //Add to transceiver list
    m_socketMap.insert(interfaceAddress, new QTmUpnpControlPointTx(this, socketTx));

    return 1;
}

QTmRemoteServer* QTmUpnpControlPoint::remoteServerById(int remoteServerId)
{
    QWriteLocker wLocker(&m_serverLock);
    foreach (QTmRemoteServer *server, m_serverList) {
        if (remoteServerId == server->remoteServerId())
            return server;
    }
    return 0;
}

QTmRemoteServer* QTmUpnpControlPoint::remoteServerByUsn(const QString &USN)
{
    foreach (QTmRemoteServer *server, m_serverList) {
        if (USN == server->USN()) {
            return server;
        }
    }
    return 0;
}

// Send M-Search Query to discover Terminal Mode UPnP Services
bool QTmUpnpControlPoint::sendMSearchQuery(QHostAddress interfaceAddress)
{
    qDebug() << "QTmUpnpControlPoint: Sending M-Search query for Terminal Mode UPnP services.";
    QWriteLocker wLocker(&m_socketLock);
    QTmUpnpControlPointTx *tx = m_socketMap.value(interfaceAddress);
    return tx ? tx->sendMSearchQuery() : 0;
}

QHostAddress QTmUpnpControlPoint::getLocalAddress(QHostAddress address)
{
    QList<QNetworkInterface> interfaces = QNetworkInterface::allInterfaces();
    foreach (QNetworkInterface interface, interfaces) {
        QList<QNetworkAddressEntry> entries = interface.addressEntries();
        foreach (QNetworkAddressEntry entry, entries) {
            if (address.isInSubnet(entry.ip(), entry.prefixLength()))
                return entry.ip();
        }
    }
    return QHostAddress();
}

/******************************************************************************/

QTmUpnpControlPointTx::QTmUpnpControlPointTx(QTmUpnpControlPoint *controlPoint, QUdpSocket *socket)
    : QObject(controlPoint)
    , m_controlPoint(controlPoint)
    , m_socketS(socket)
{
#ifdef QT_DEBUG_TM_UPNP
    qDebug() << "QTmUpnpControlPointTx constructor";
#endif
    if(m_socketS->localAddress().protocol() == QAbstractSocket::IPv4Protocol)
    {
        //Initialize the group sockaddr structure with
        //SSDP group IPv4 address of 239.255.255.250 and port 1900
        memset((char *) &groupSock, 0, sizeof(groupSock));
        groupSock.sin_family = AF_INET;
        groupSock.sin_addr.s_addr = inet_addr("239.255.255.250");
        //inet_aton((const char*)"239.255.255.250", &groupSock.sin_addr);
        groupSock.sin_port = htons(1900);
    }
    else if(m_socketS->localAddress().protocol() == QAbstractSocket::IPv6Protocol)
    {
        //Initialize the group sockaddr structure with
        //SSDP group IPv6 address of FF02::C and port 1900
        memset((char *) &groupSock6, 0, sizeof(groupSock6));
        groupSock6.sin6_family = AF_INET6;
        inet_pton(AF_INET6, (const char*)"FF02::C", groupSock6.sin6_addr.s6_addr);
        groupSock6.sin6_flowinfo = 0;
        groupSock6.sin6_port = htons(1900);
    }
    connect(m_socketS, SIGNAL(readyRead()),
            this,      SLOT  (readUnicastResponse()),
            Qt::AutoConnection);
}

QTmUpnpControlPointTx::~QTmUpnpControlPointTx()
{
#ifdef QT_DEBUG_TM_UPNP
    qDebug() << "~QTmUpnpControlPointTx destructor";
#endif
    m_socketS->close();
    disconnect(m_socketS, SIGNAL(readyRead()), this, SLOT(readUnicastResponse()));
    delete m_socketS;
}


void QTmUpnpControlPointTx::readUnicastResponse()
{
#ifdef QT_DEBUG_TM_UPNP
    qDebug() << "QTmUpnpControlPointTx:" << "read unicast message";
#endif
    m_controlPoint->readMessage(m_socketS);
}

bool QTmUpnpControlPointTx::sendMSearchQuery()
{
    qDebug() << "QTmUpnpControlPointTx:" << "sendMSearchQuery";
    //Send multicast M-SEARCH query
    if(m_socketS->localAddress().protocol() == QAbstractSocket::IPv4Protocol)
    {
        //char *M_SEARCH_QUERY = (char*)"M-SEARCH * HTTP/1.1\r\nHOST: 239.255.255.250:1900\r\nMAN: \"ssdp:discover\"\r\nMX: 1\r\nST: urn:schemas-upnp-org:device:TmServerDevice:1\r\n\r\n";
        //char *M_SEARCH_QUERY = (char*)"M-SEARCH * HTTP/1.1\r\nHOST: 239.255.255.250:1900\r\nMAN: \"ssdp:discover\"\r\nMX: 1\r\nST: urn:schemas-upnp-org:service:TmApplicationServer:1\r\n\r\n";
        //char *M_SEARCH_QUERY = (char*)"M-SEARCH * HTTP/1.1\r\nHOST: 239.255.255.250:1900\r\nMAN: \"ssdp:discover\"\r\nMX: 1\r\nST: urn:schemas-upnp-org:service:TmClientProfile:1\r\n\r\n";
        char *M_SEARCH_QUERY = (char*)"M-SEARCH * HTTP/1.1\r\nHOST: 239.255.255.250:1900\r\nMAN: \"ssdp:discover\"\r\nMX: 1\r\nST: ssdp:all\r\n\r\n";
        sendto(m_socketS->socketDescriptor(), M_SEARCH_QUERY, strlen(M_SEARCH_QUERY), 0, (struct sockaddr*)&groupSock, sizeof(groupSock));
    }
    else if(m_socketS->localAddress().protocol() == QAbstractSocket::IPv6Protocol)
    {
        char *M_SEARCH_QUERY = (char*)"M-SEARCH * HTTP/1.1\r\nHOST: FF02::C:1900\r\nMAN: \"ssdp:discover\"\r\nMX: 1\r\nST: urn:schemas-upnp-org:device:TmServerDevice:1\r\n\r\n";
        //char *M_SEARCH_QUERY = (char*)"M-SEARCH * HTTP/1.1\r\nHOST: 239.255.255.250:1900\r\nMAN: \"ssdp:discover\"\r\nMX: 1\r\nST: urn:schemas-upnp-org:service:TmApplicationServer:1\r\n\r\n";
        //char *M_SEARCH_QUERY = (char*)"M-SEARCH * HTTP/1.1\r\nHOST: 239.255.255.250:1900\r\nMAN: \"ssdp:discover\"\r\nMX: 1\r\nST: urn:schemas-upnp-org:service:TmClientProfile:1\r\n\r\n";
        //char *M_SEARCH_QUERY = (char*)"M-SEARCH * HTTP/1.1\r\nHOST: 239.255.255.250:1900\r\nMAN: \"ssdp:discover\"\r\nMX: 1\r\nST: ssdp:all\r\n\r\n";
        sendto(m_socketS->socketDescriptor(), M_SEARCH_QUERY, strlen(M_SEARCH_QUERY), 0, (struct sockaddr*)&groupSock6, sizeof(groupSock6));
    }
    else
    {
        return false;
    }
    return true;
}

