/****************************************************************************
**
** Copyright (C) 2012 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Lesser General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public
** License along with this library; if not, write to the
** Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
**
****************************************************************************/
/*
 * ChangeLog -  
 *				July 05, 2012
 *				1. Support for multiple UpnP control points and VNC connections is removed.
 *              So to avoid unecessary memory usage and processing, use of 'm_rfbClientMap' and 'm_threadMap' is disabled. Instead 'm_activeRfbClient' is added.
 *              2. Call to rfbClient->sessionStop() is moved from vncDisconnectClient() to vncStopClient(). The function is called after the RFB thread is exited.
 *              This makes sure that no network I/O happenning while socket close (i.e. rfbClient->sessionStop()).
 *              3. Use of 'm_widgetMap'is disabled to remove the dependency on Qt GUI library.
 *              4. Added keyEvent(), pointerEvent(), member functions. These gets triggered from TM application. 
 *				These functions internally call QtmRfbClient APIs for sending command to TM server
 *				March 19, 2013
 *				1. Added connectSurfaceToRfbClient() api to store the active surface instance in rfbclient.
 *				2. Added vncSupportedKeyEvents() api to register the application supported key events.
 *				3. Added vncPauseFbUpdateRequest() and vncResumeFbUpdateRequest() api for pausing/resuming the fb updates from the ML server.
 */

#ifndef QTMCLIENT_H
#define QTMCLIENT_H

#include <QtCore/QHash>
#include <QtCore/QSignalMapper>

#ifndef QT_UI_CLEANUP
#include "QTmWidget.h"
#endif
#include "QTmRfbClient.h"
#include "QTmUpnpControlPoint.h"
#include "QTmNetworkObserver.h"
#include "QTmClientProfile.h"
#include "QTmClientRules.h"
#include <qrect.h>
#include <qsize.h>
#include <qthread.h>
#include <TmHeaders/RfbThreadListener.h>
#ifndef QT_UI_CLEANUP
#include "QTmDeclarativeItem.h"
#endif

class QTmClient : public QObject, public RfbThreadListener {
    Q_OBJECT

public:
                            QTmClient(QObject *parent = 0);
                            ~QTmClient();

    void rfbThreadExited(int clientId);
    void                    networkStartObserver(bool monitorIPv4 = true, bool monitorIPv6 = false);
    void                    networkStopObserver();
    QNetworkConfiguration::BearerType networkBearerType(QString interfaceName);

    int                     vncStartClient();
#ifdef QT_UI_CLEANUP
    bool                    vncConnectClient(QTmApplication *application, int id);

#else
    void                    vncConnectClient(QTmApplication *application, int id);
#endif

    void                    vncDisconnectClient(int id);
    void                    vncStopClient(int id);

#ifndef QT_UI_CLEANUP
    QWidget                *vncClientWidget(int id);
    void                    setClientDeclarativeItem(QTmDeclarativeItem* item, int id);
#else    //QT_UI_CLEANUP
    bool                    connectRfbClientToSurface(int id, QTmSurface* tmSurface);
    /*!@{*/
    bool                    connectSurfaceToRfbClient(int id, QTmSurface* tmSurface);
    /*!@}*/
    int                     vncSupportedKeyEvents(int clientId, int knobKey, int deviceKey, int mediaKey);
    void                    keyEvent(int vncId, int key_id, bool key_pressed);
    //void                    pointerEvent(int vncId, QPoint pointer_event, unsigned int pointer_pressed);
    void                    pointerEvent(int vncId, int x, int y, unsigned int pointer_pressed);
    int                     translateKey(int key, Qt::KeyboardModifiers modifier);
#endif

    void                    vncSetColorFormat      (QTmGlobal::TmColorFormat colorFormat, int id);
    void                    vncSetRunLengthEncoding(bool enable, int id);
    void                    vncSetIncrementalUpdate(bool enable, int id);
    void                    vncSetPreferredSize(QSize size, bool scaling, int id, double clientSinglePixelWidth, double clientSinglePixelHeight, bool considerPixelComp);
    void                    vncClickVirtualKey     (int keySym,  int id);
    void                    vncBlockFramebuffer(QRect area, unsigned int appId, unsigned int reason, int id);
    void                    vncResetFbUpdateRequestTimer();
    void                    vncPauseFbUpdateRequest();
    void                    vncResumeFbUpdateRequest();


    bool                    upnpStartControlPoint();
    bool                    upnpConnectControlPoint(QString interfaceName, QHostAddress interface);
    void                    upnpDisconnectControlPoint(QHostAddress hostAddress);
    void                    upnpStopControlPoint();
    bool                    upnpLaunchApplication(QTmApplication *application);
    bool                    upnpTerminateApplication(QTmApplication *application);
    QTmGlobal::TmStatus     upnpGetApplicationStatus(QTmApplication *application);

signals:
    void     networkDeviceDetected(QString interfaceName, QHostAddress ipAddress);
    void     networkDeviceLost(QHostAddress ipAddress);

    void     vncClientResized(QSize size, int id);
    void     vncClientConnected(int id);
    void     vncClientDisconnected(int id);
    void     vncClientDisplayConfiguration(int id);
    void     vncClientEventConfiguration(int id, int knobKey, int deviceKey, int mediaKey);
    void     vncClientKeyboardTrigger(QPoint point, bool remove, int id);
    void     vncClientKeyEventList(QList<unsigned int> *list, unsigned int counter, int id);
    void     vncClientFbAlternativeText(unsigned int appId, QString text, int id);
    void     vncClientContextInformation(QRect area, int appId, QTmAppInfo *appCategory, QTmDisplayInfo *conCategory, int id);

    void     remoteServerDeviceAdded(QTmRemoteServer *remoteServerDevice);
    void     remoteServerDeviceRemoved(QTmRemoteServer *remoteServerDevice);

private slots:
    void     deviceDetected(QString networkInterfaceName, QNetworkAddressEntry networkEntry);
    void     deviceLost(QNetworkAddressEntry networkEntry);

    // RfbClient signal-wrappers:
    void rfbDesktopSize(QSize framebufferSize);
    void rfbSessionConnected(QString serverName);
    void rfbSessionDisconnected();
    void rfbEventConfiguration(int knobKey, int deviceKey, int mediaKey);
    void rfbDisplayConfiguration();
    void rfbVirtualKeyboardTrigger(QPoint point, bool remove);
    void rfbKeyEventList(QList<unsigned int> *list, unsigned int counter);
    void rfbFramebufferAlternativeText(unsigned int appId, QString text);
    void rfbContextInformation(QRect area, int appId, QTmAppInfo *appCategory, QTmDisplayInfo *conCategory);
    void rfbDeviceStatus(unsigned int);

private:
    QTmRemoteServer* remoteServerByApplication(QTmApplication *application);

#ifndef QT_UI_CLEANUP
    bool connectRfbClientToSurface(int id);
#endif
    void disconnectRfbClientFromSurface(int id);

    QTmUpnpControlPoint            *m_upnpControlPoint;

    //QHash<int, QTmRfbClient*>       m_rfbClientMap;   //no need of map,since only 1 rfb session is reqd

    QTmRfbClient                    *m_activeRfbClient;

#ifndef QT_UI_CLEANUP
    QHash<int, QTmWidget* >         m_widgetMap;
#endif

    //QHash<int, QThread* >           m_threadMap;      //no need of map,since only 1 rfb session is reqd

#ifndef QT_UI_CLEANUP
    QHash<int, QTmDeclarativeItem*> m_declarativeMap;
#endif

    int                             m_maxVncId;

    QTmNetworkObserver             *m_networkObserver;
};

#endif // QTMCLIENT_H
