/****************************************************************************
**
** Copyright (C) 2012 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Lesser General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public
** License along with this library; if not, write to the
** Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
**
****************************************************************************/
/*
 * ChangeLog -  
 *				July 05, 2012
 * 				1. Added logic to limit the max frame update rate to 10.
 *              2. Added checks for detecting and handling socket errors.
 *              3. Appended missing '\n' in RFB protocol version string
 *              4. Added readSecurityFailReason() function to read the fail reason string and
 *                 also handles socket reconnection in this case.
 *              5. Changed the sequence of sending the framebufferUpdateRequest().
 *                 Its now sent only after complete VNC frame is received.
 *                 Previously it was getting sent before reading last frame rectangle.
 *              6. Added logic to determine the up/down scaling support based on the server side scaling capability.
 *              7. Added logic to request the best fitresolution values to  TM server that support scaling.
 *              8. Added ogic to calculate the scalling factors depending on the negotiated resolution values
 *                 and then notifying it to application.
 *				March 19, 2013
 *				1. Added getter/setter methods for enabling/disabling consideration of pixel compensation.
 *			    2. Added pauseFbUpdateRequest()/resumeFbUpdateRequest() api to support pause and resume of VNC fb update.
 *			    3. Added updateIdleBufferAvailability() observer method to get the notification from the double buffer manager.
 */


#ifndef QTMRFBCLIENT_H
#define QTMRFBCLIENT_H

#include <QTcpSocket>
#include <QMutex>
#include <QObject>
#include <QHash>
#include <QUrl>
#include <qsize.h>
#include <qrect.h>
#include <QtCore/qthread.h>

#include "QTmConfiguration.h"
#include "QTmAppInfo.h"
#include "QTmDisplayInfo.h"
#include "QTmSurface.h"
#include "QTmTouchEvent.h"
#include "QTmKeyEvent.h"
#include "QTmPointerEvent.h"
#include "TCPSocket.h"
#include <TmHeaders/RfbThreadListener.h>
#include <QDateTime>
#include <SurfaceObserver.h>


#ifdef QT_UI_CLEANUP
class QTmApplication;
//class TcpSocket;
class QTmRfbClient : public QThread, public SurfaceObserver{
#else
class QTmRfbClient : public QObject {
#endif

    Q_OBJECT
    Q_PROPERTY(bool sessionActive READ sessionActive)
    Q_PROPERTY(QUrl sessionUrl    READ sessionUrl)

public:

    enum QTmSessionStatus {
        RFB_SESSION_NONE,
        RFB_SESSION_CONNECTED_ACTIVE,
        RFB_SESSION_CONNECTED_PASSIVE,
        RFB_SESSION_SERVER_VERSION,
        RFB_SESSION_SECURITY_TYPES,
        RFB_SESSION_SECURITY_RESPONSE,
        RFB_SESSION_SERVER_INIT,
        ERROR_RFB_SESSION_DISCONNECTED,
        ERROR_WRONG_SERVER_VERSION,
        ERROR_WRONG_SECURITY_TYPES,
        ERROR_WRONG_FRAMEBUFFER_ENCODING,
        ERROR_WRONG_MSG_TYPE,
        ERROR_WRONG_FSM_STATE,
        ERROR_CANNOT_ALLOCATE_SURFACE,
    };

    QTmRfbClient(QObject *parent = 0);
    ~QTmRfbClient();


#ifdef QT_UI_CLEANUP
    void run();
    void setClientId(int clientId);
    void setRfbThreadListener(RfbThreadListener* listener);
    void handleSocketError();
    int forceStopRfbRead();
#endif

    void      setClientConfig(QTmConfiguration::QTmConfig key, unsigned int value = 0);
    unsigned int clientConfig(QTmConfiguration::QTmConfig key, unsigned int defaultValue = 0);
    unsigned int serverConfig(QTmConfiguration::QTmConfig key, unsigned int defaultValue = 0);

    void clientCutText(QString text);
    void triggerFramebufferUpdates();

    void clientDisplayConfig();
    void clientEventConfig();
    void keyEventListingRequest(bool enable);
    void deviceStatusRequest();
    void eventMappingRequest(unsigned int clientKey, unsigned int serverKey);
    void virtualKeyboardTriggerRequest(bool enable);
    void contentAttestationRequest(char *nounce, char attestation, char signature, char key, char *session_key);
    void framebufferAlternativeTextRequest(unsigned short int max_length);
    void framebufferBlockingNotification(QRect fbArea, unsigned int appId, unsigned int reason);
    void audioBlockingNotification(unsigned int appId, unsigned int reason);
    void setClientSinglePixelWidth(double singlePixelWidth);
    void setClientSinglePixelHeight(double singlePixelHeight);
    void setSurface(QTmSurface *surface);
    QSize setServerFbSize(QSize size);
    QSize setClientFbSize(QSize size, bool scaling);
    void  setConsiderPixelComp(bool considerPixelComp);
    bool  getConsiderPixelComp();
    void getExpectedClientSize(QSize serverSize, QSize * clientFrameBufferSize, bool afterFbResize);
    void resetFbUpdateRequestTimer();
    void resumeFbUpdateRequest();
    void pauseFbUpdateRequest();

    bool sessionStart(QUrl url);
    void sessionStop();
    bool sessionActive();
    QUrl sessionUrl();
    QTmSessionStatus sessionStatus();
    QHash<unsigned int, unsigned int> *eventMap();

    void readSecurityFailReason();

    /*!@{*/ // observer method
    void updateIdleBufferAvailability();
    /*!@}*/

public slots:
    void keyEvent(QTmKeyEvent *event);
#ifdef QT_UI_CLEANUP
    void pointerEvent(int x, int y, unsigned int pointer_pressed);
#else
    void pointerEvent(QTmPointerEvent *pointerEvent);
#endif

    void touchEvent(QList<QTmTouchEvent *> *touchEvent);

signals:
    void sessionConnected(QString serverName);
    void sessionDisconnected();
    void framebufferResized(QSize size);
    void displayConfigurationReceived();
    void eventConfigurationReceived(int knobKey, int deviceKey, int mediaKey);
    void eventMappingReceived(unsigned int clientKey, unsigned int serverKey);
    void deviceStatusReceived(unsigned int deviceStatus);
    void keyEventListReceived(QList<unsigned int> *list, unsigned int counter);
    void virtualKeyboardTriggerReceived(QPoint point, bool remove);
    void framebufferAlternativeTextReceived(unsigned int appId, QString text);
    void contextInformationReceived(QRect area, int appId, QTmAppInfo *appCategory, QTmDisplayInfo *conCategory);

private:
    void protocolHandshake();
    void securityHandshake();
    void securityResponse();
    void clientInit();
    void serverInit();
    //void setEncodings(int encoding);
    void setEncodings();
    void setPixelFormat();
    void serverCutText();
    void setColourMapEntries();
    void bell();
    void framebufferUpdateRequest(bool incremental);
    void framebufferUpdate();    

    int  serverDisplayConfig();
    int  serverEventConfig();
    int  keyEventListing();
    int  deviceStatus();
    int  eventMapping();
    int  virtualKeyboardTrigger();
    int  contentAttestation();
    int  framebufferAlternativeText();
    void calculateScalingFactors(int clientDisplayXPixels, int clientDisplayYPixels,int serverXPixels, int serverYPixels);
    void desktopSizePseudoEncoding(QSize framebufferSize);
    int contextInformationPseudoEncoding(unsigned char *cmdBuffer);
    int rawEncoding(int x, int y, int width, int height);
#ifdef QT_DEBUG_TM_PERFORMANCE_COUNTER
    int  rleEncoding(int x, int y, int width, int height);
#else
    int rleEncoding(int x, int y, int width, int height);
#endif

    int  receiveData(int length, char *buffer);
    int  transmitData(int length, char *buffer);

    void sendFbUpdateRequest(bool incremental);
    double getCurrMillisec();

private slots:
    void protocolFsm();

private:
    volatile bool                      m_connected;
    QHash<unsigned int, unsigned int> *m_eventMap;
    QTmSurface                        *m_surface;
    QUrl                               m_url;
#ifdef QT_UI_CLEANUP
    TCPSocket                        *m_socket;
    RfbThreadListener                *m_threadListenr;
    int m_clientId;
    bool m_socketError;
#else
    QTcpSocket                        *m_socket;
#endif

    int                                m_supportedEncodingCount;

    QTmConfiguration                  *m_rfbConfigClient;
    QTmConfiguration                  *m_rfbConfigServer;
    QMutex                             m_mutex;
    QTmSessionStatus                   m_rfbSessionStatus;
    unsigned int                       m_keyEventCounter;
    unsigned char                     *m_framebuffer;
    bool                               m_displayStatusFlag;
    bool                               m_fbUpdateRequestFlag;
    int                                m_frameCount;    
    double                             m_timeLastFBRequest;
    double                             m_timeRoundTrip;         //time between framebuffer update request and
                                                                //complete receival of framebuffer update.
                                                                //this doesnt include rendering time

    /** Support Added to calculte the FB X and Y scale factors in case of Model Variation **/
    double                             m_calculatedXScaleFactor;
    double                             m_calculatedYScaleFactor;
    double                             m_clientSinglePixelWidth;
    double                             m_clientSinglePixelHeight;
    bool                               m_considerPixelComp;

    /*!@{*/
    bool                               m_isSendingFbUpdateRequestFirstTime;
    unsigned char                      *m_encodingCmdBuffer;
    /*!@}*/
};

#endif // QTMRFBCLIENT_H
