/*******************************************************************************
**
** Copyright 2013 Sony Corporation
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Lesser General Public
** License, version 2.1, as published by the Free Software Foundation.
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** Lesser General Public License for more details.
** You should have received a copy of the GNU Lesser General Public
** License along with this library; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
**
*******************************************************************************/
/*
 * ChangeLog -  
 *				March 19, 2013
 *				1. Added vncPauseFbUpdateRequest() and vncResumeFbUpdateRequest() api for pausing/resuming the fb updates from the ML server.
 *              2. Added support to enable/disable the consideration of pixel compensation of client display panel.
 *              3. Added support to handle knob key and media key events.
 *              4. Added suooprt for integrating the double buffer implementation.
 *
 *              August 02, 2013
 *              1. Changed the prototype of upnpLaunchApplication() for avoid data corruption while passing data using pointer.
 *              2. Changed the prototype of getApplicationById() for avoid data corruption while passing data using pointer.
 *              3. Always emit 'remoteServerDeviceRemoved' to ensure the cleanup at upper layers/managers on every ML phone disconnect.
 */

#include <libinterfaceimpl.h>
#include <appinterface.h>
#include <QTmAppInfo.h>
#include <QTmDisplayInfo.h>
#include <QTmClient.h>
#include <QTmRemoteServer.h>
#include <QTmApplication.h>
#include <qtmsurfaceimpl.h>
#include <qdebug.h>
#include <libinterfaceemitter.h>

/** -------------------------------------------------------------------------
    construction & destruction
---------------------------------------------------------------------------**/

LibInterfaceImpl::LibInterfaceImpl(LibInterfaceEmitter* emitter)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::LibInterfaceImpl() ---  created";
#endif
    m_remoteServerList = new QList<QTmRemoteServer *>;
    m_remoteServerIdToVncIdMap = new QMultiHash<int, int>();
    m_appIdToApplicationMap = new QMultiHash<QString, QTmApplication*>();
    m_TerminalMode = new QTmClient(this);
    m_surfaceMap = new QMultiHash<int,QTmSurface*>();

    connect(emitter, SIGNAL(sig_upnpStartControlPoint()), this, SLOT(upnpStartControlPoint()));
    connect(emitter, SIGNAL(sig_upnpConnectControlPoint(char* , char* )), this, SLOT(upnpConnectControlPoint(char* , char* )));
    connect(emitter, SIGNAL(sig_verifyTmClientProfile(QTmClientProfile*, int, int)),
            this,SLOT(verifyTmClientProfile(QTmClientProfile*,int,int)));

    connect(emitter, SIGNAL(sig_upnpStopControlPoint()), this, SLOT(upnpStopControlPoint()));
    connect(emitter, SIGNAL(sig_upnpDisconnectControlPoint(char*)), this, SLOT(upnpDisconnectControlPoint(char*)));
    connect(emitter, SIGNAL(sig_upnpGetApplicationStatus(char*)), this, SLOT(upnpGetApplicationStatus(char*)));
    //connect(emitter, SIGNAL(sig_upnpLaunchApplication(char*)), this, SLOT(upnpLaunchApplication(char*)));
    connect(emitter, SIGNAL(sig_upnpLaunchApplication(const QString)), this, SLOT(upnpLaunchApplication(const QString)));
    connect(emitter, SIGNAL(sig_upnpTerminateApplication(char* )), this, SLOT(upnpTerminateApplication(char* )));
    connect(emitter, SIGNAL(sig_getTmClientProfile(int,int)), this, SLOT(getTmClientProfile(int,int)));
    connect(emitter, SIGNAL(sig_setTmClientProfile(QTmClientProfile *,int,int)), this, SLOT(setTmClientProfile(QTmClientProfile *,int,int)));
    connect(emitter, SIGNAL(sig_vncStartClient()), this, SLOT(vncStartClient()));
    connect(emitter, SIGNAL(sig_vncConnectClient(char*,int)), this, SLOT(vncConnectClient(char*,int)));
    connect(emitter, SIGNAL(sig_vncDisconnectClient(int)), this, SLOT(vncDisconnectClient(int)));
    connect(emitter, SIGNAL(sig_vncStopClient(int)), this, SLOT(vncStopClient(int)));
    connect(emitter, SIGNAL(sig_vncSetColorFormat(int,int)), this, SLOT(vncSetColorFormat(int,int)));
    connect(emitter, SIGNAL(sig_vncSetRunLengthEncoding(bool,int)), this, SLOT(vncSetRunLengthEncoding(bool,int)));
    connect(emitter, SIGNAL(sig_vncSetIncrementalUpdate(bool,int)), this, SLOT(vncSetIncrementalUpdate(bool,int)));
    connect(emitter, SIGNAL(sig_vncSetPreferredSize(int,int,bool,int, double, double, bool)), this, SLOT(vncSetPreferredSize(int,int,bool,int, double, double, bool)));
    connect(emitter, SIGNAL(sig_vncOnFBReadComplete(int)), this, SLOT(vncOnFBReadComplete(int)));
    connect(emitter, SIGNAL(sig_vncClickVirtualKey(int,int)), this, SLOT(vncClickVirtualKey(int,int)));
    connect(emitter, SIGNAL(sig_vncPointerEvent(int,int,int,unsigned int)), this, SLOT(vncPointerEvent(int,int,int,unsigned int)));
    connect(emitter, SIGNAL(sig_vncSupportedKeyEvents(int,int,int,int)), this, SLOT(vncSupportedKeyEvents(int,int,int,int)));
    connect(emitter, SIGNAL(sig_vncKeyEvent(int,int,bool)), this, SLOT(vncKeyEvent(int,int,bool)));
    connect(emitter, SIGNAL(sig_vncBlockFramebuffer(int,int,int,int,int,int,int)), this, SLOT(vncBlockFramebuffer(int,int,int,int,int,int,int)));
    connect(emitter, SIGNAL(sig_vncResetFbUpdateRequestTimer()), this, SLOT(vncResetFbUpdateRequestTimer()));
    connect(emitter, SIGNAL(sig_vncPauseFbUpdateRequest()), this, SLOT(vncPauseFbUpdateRequest()));
    connect(emitter, SIGNAL(sig_vncResumeFbUpdateRequest()), this, SLOT(vncResumeFbUpdateRequest()));


    connect(m_TerminalMode, SIGNAL(vncClientConnected(int)),      this, SLOT(vncConnected(int)));
    connect(m_TerminalMode, SIGNAL(vncClientDisconnected(int)),   this, SLOT(vncDisconnected(int)));
    connect(m_TerminalMode, SIGNAL(vncClientResized(QSize, int)), this, SLOT(vncResized(QSize, int)));
    connect(m_TerminalMode, SIGNAL(vncClientContextInformation(QRect,int,QTmAppInfo*,QTmDisplayInfo*,int)),
            this,           SLOT  (vncClientContextInformation(QRect,int,QTmAppInfo*,QTmDisplayInfo*,int)));
    connect(m_TerminalMode, SIGNAL(remoteServerDeviceAdded(QTmRemoteServer*)),
            this,           SLOT  (remoteServerDeviceAdded(QTmRemoteServer*)));
    connect(m_TerminalMode, SIGNAL(remoteServerDeviceRemoved(QTmRemoteServer*)),
            this,           SLOT  (remoteServerDeviceRemoved(QTmRemoteServer*)));
    connect(m_TerminalMode, SIGNAL(vncClientEventConfiguration(int, int, int, int)),      this, SLOT(vncClientEventConfiguration(int, int, int, int)));

}

LibInterfaceImpl::~LibInterfaceImpl()
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::~LibInterfaceImpl()"<< m_TerminalMode;
    qDebug()<<"About to disconnect signals";
#endif
    if(m_TerminalMode)
    {
        disconnect(m_TerminalMode, SIGNAL(vncClientConnected(int)),      this, SLOT(vncConnected(int)));
        disconnect(m_TerminalMode, SIGNAL(vncClientDisconnected(int)),   this, SLOT(vncDisconnected(int)));
        disconnect(m_TerminalMode, SIGNAL(vncClientResized(QSize, int)), this, SLOT(vncResized(QSize, int)));
        disconnect(m_TerminalMode, SIGNAL(vncClientContextInformation(QRect,int,QTmAppInfo*,QTmDisplayInfo*,int)),
                this,           SLOT  (vncClientContextInformation(QRect,int,QTmAppInfo*,QTmDisplayInfo*,int)));
        disconnect(m_TerminalMode, SIGNAL(remoteServerDeviceAdded(QTmRemoteServer*)),
                this,           SLOT  (remoteServerDeviceAdded(QTmRemoteServer*)));
        disconnect(m_TerminalMode, SIGNAL(remoteServerDeviceRemoved(QTmRemoteServer*)),
                this,           SLOT  (remoteServerDeviceRemoved(QTmRemoteServer*)));
        disconnect(m_TerminalMode, SIGNAL(vncClientEventConfiguration(int, int, int, int)), this, SLOT(vncClientEventConfiguration(int, int, int, int)));
#ifdef QT_CODE_DEBUG
        qDebug()<<"All library signals disconnected";
#endif
    }
#ifdef QT_CODE_DEBUG
        qDebug()<<"Deleting m_remoteServerIdToVncIdMap";
#endif
    delete m_remoteServerIdToVncIdMap;
#ifdef QT_CODE_DEBUG
        qDebug()<<"Deleting  m_appIdToApplicationMap";
#endif
    delete m_appIdToApplicationMap;
#ifdef QT_CODE_DEBUG
        qDebug()<<"Deleting m_remoteServerList";
#endif
    delete m_remoteServerList;
#ifdef QT_CODE_DEBUG
        qDebug()<<"Deleting m_surfaceMap";
#endif
    delete m_surfaceMap;
#ifdef QT_CODE_DEBUG
        qDebug()<<"Deleting m_TerminalMode";
#endif
    delete m_TerminalMode;

}

/** -------------------------------------------------------------------------
    private methods:
---------------------------------------------------------------------------**/

//QTmApplication* LibInterfaceImpl::getApplicationById(char* appId)
QTmApplication* LibInterfaceImpl::getApplicationById(const QString appId)
{
    //QString key = QString(appId);
    QString key = appId;
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::getApplicationById() appId :"<< key;
#endif

    if(m_appIdToApplicationMap->contains(key))
    {
        return m_appIdToApplicationMap->value(key);
    }
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::getApplicationById() returning appId :"<< 0;
#endif
    return 0;
}

QTmRemoteServer* LibInterfaceImpl::getRemoteServer(int serverId)
{
    QTmRemoteServer *existing_remoteServer = 0;
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::getRemoteServer() serverId :"<< serverId;
#endif
    foreach (QTmRemoteServer *remoteServer , *m_remoteServerList) {
        if(remoteServer->remoteServerId() == serverId)
        {
            existing_remoteServer = remoteServer;
            break;
        }
    }
    return existing_remoteServer;
}


/** -------------------------------------------------------------------------
    public methods
---------------------------------------------------------------------------**/

void LibInterfaceImpl::setAppInterface(AppInterface* appInterface)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::setAppInterface() "<< appInterface;
#endif
    this->m_appInterface = appInterface;
}

/** methods inherited from LibInterface*/

int LibInterfaceImpl::upnpStartControlPoint()
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpStartControlPoint() By SIGNAL EMMISION";
#endif
    bool controlPointStarted = m_TerminalMode->upnpStartControlPoint();
    if(controlPointStarted)
    {
        m_appInterface->onStatus(LibInterface::UPNP_START_CONTROL_POINT, AppInterface::SUCCESS);
    }
    else
    {
        m_appInterface->onStatus(LibInterface::UPNP_START_CONTROL_POINT, AppInterface::FAILURE);
    }
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpStartControlPoint() controlPointStarted :" << controlPointStarted;
#endif
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::upnpStopControlPoint()
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpStopControlPoint() By SIGNAL EMMISION";
#endif
    m_TerminalMode->upnpStopControlPoint();
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::upnpConnectControlPoint(char* interfaceName, char* address)
{
    QString qinterfaceName = QString(QLatin1String(interfaceName));
    QString qaddress = QString(QLatin1String(address));
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpConnectControlPoint() - interface :" << qinterfaceName << ", qaddress :" << qaddress;
#endif
    QHostAddress hostAddress = QHostAddress(qaddress);

    bool connected = m_TerminalMode->upnpConnectControlPoint(qinterfaceName,hostAddress);
    if(connected)
    {
        m_appInterface->onStatus(LibInterface::UPNP_CONNECT_CONTROL_POINT,AppInterface::SUCCESS);
    }
    else
    {
        m_appInterface->onStatus(LibInterface::UPNP_CONNECT_CONTROL_POINT,AppInterface::FAILURE);
    }
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::verifyTmClientProfile(QTmClientProfile *clientProfile, int profileID, int serverId)
{

#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::verifyTmClientProfile() - profileID :" << profileID << ", server id:" << serverId << "client profile :" << clientProfile;
    qDebug()<<"-----------------------------------------";
    qDebug()<<"1. client name :" << clientProfile->clientName() ;
    qDebug()<<"2. client id :" << clientProfile->clientId() ;
    qDebug()<<"3. manufacturer  :" << clientProfile->manufacturerName() ;
    qDebug()<<"4. model no :" << clientProfile->modelNumber();

    qDebug()<<"rtp ipl :" << clientProfile->rtpipl();
    qDebug()<<"rtp mpl :" << clientProfile->rtpmpl();
    qDebug()<<"rtp payload :" << clientProfile->rtppayload();

#endif

    bool testOK = false;
    QTmRemoteServer* existing_remoteServer = getRemoteServer(serverId);
    #ifdef QT_CODE_DEBUG
    qDebug()<<"Remote server :" << existing_remoteServer;
    #endif

    if(existing_remoteServer)
    {
        #ifdef QT_CODE_DEBUG
        qDebug()<<"About to call setClientProfile ..." <<clientProfile;
        #endif
        QTmClientProfile clientProfile1 = existing_remoteServer->setClientProfile(clientProfile, profileID);
        #ifdef QT_CODE_DEBUG
        qDebug()<<"clientProfile1" <<  clientProfile1.isNull();
        qDebug()<<"clientProfile1.clientName:" <<  clientProfile1.clientName();
         qDebug()<<"libinterfaceImpl - 1 ::: Client Id ######## "<<clientProfile1.clientId();
        #endif
        if(!clientProfile1.isNull())
        {
            testOK = clientProfile->equals(&clientProfile1);
            #ifdef QT_CODE_DEBUG
            qDebug()<<"setClient profile response testOK :" << testOK;
            #endif
            //Test GetClientProfile
            QTmClientProfile clientProfile2 = existing_remoteServer->getClientProfile(profileID);
            #ifdef QT_CODE_DEBUG
            qDebug()<<"libinterfaceImpl - 2 ::: Client Id ######## "<<clientProfile2.clientId();
            qDebug()<<"clientProfile1 :" << clientProfile2.isNull();
            #endif
            if(!clientProfile2.isNull())
            {
                testOK = clientProfile->equals(&clientProfile2);
                #ifdef QT_CODE_DEBUG
                qDebug()<<"get client profile response testOK :" << testOK;
             #endif
            }
            else
            {
                qDebug() << "LibInterfaceImpl: verifyTmClientProfile(): GetClientProfile has returned a null profile";
            }
        }
        else
        {
            qDebug() << "LibInterfaceImpl::verifyTmClientProfile(): SetClientProfile has returned a null profile";
        }
    }
    else
    {
        qDebug() << "LibInterfaceImpl::verifyTmClientProfile(): invalid server id provided";
    }
    qDebug()<<"Test OK : "<< testOK;
    //testOK = true;
    if(testOK)
    {
        m_appInterface->onStatus(LibInterface::TM_VERIFY_CLIENT_PROFILE,AppInterface::SUCCESS);
        QString appListingFilter = QString("*");
        qDebug()<< "About to call getApplicationList ...";
        existing_remoteServer->getApplicationList(appListingFilter,profileID);
    }
    else
    {
        m_appInterface->onStatus(LibInterface::TM_VERIFY_CLIENT_PROFILE,AppInterface::FAILURE);
    }
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::verifyTmClientProfile() - testOK :" << testOK;
#endif
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::getTmClientProfile(int profileID, int serverId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::getTmClientProfile() - profileID :" << profileID << ", server id:" << serverId;
#endif
    QTmRemoteServer* existing_remoteServer = getRemoteServer(serverId);
    if(existing_remoteServer)
    {
        QTmClientProfile clientProfile = existing_remoteServer->getClientProfile(profileID);
        m_appInterface->onProfileUpdate(LibInterface::TM_GET_CLIENT_PROFILE,&clientProfile,profileID);
    }
    else
    {
        qDebug() << "LibInterfaceImpl::verifyTmClientProfile(): invalid server id provided";
        m_appInterface->onStatus(LibInterface::TM_GET_CLIENT_PROFILE,AppInterface::ERR_INVALID_SERVER_ID);
    }
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::setTmClientProfile(QTmClientProfile *clientProfile, int profileID, int serverId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::setTmClientProfile() - profileID :" << profileID << ", server id:" << serverId;
#endif
    QTmRemoteServer* existing_remoteServer = getRemoteServer(serverId);
    if(existing_remoteServer)
    {
        QTmClientProfile clientProfile2 =  existing_remoteServer->setClientProfile(clientProfile, profileID);
        m_appInterface->onProfileUpdate(LibInterface::TM_SET_CLIENT_PROFILE,&clientProfile2,profileID);
    }
    else
    {
        qDebug() << "LibInterfaceImpl::verifyTmClientProfile(): invalid server id provided";
        m_appInterface->onStatus(LibInterface::TM_SET_CLIENT_PROFILE,AppInterface::ERR_INVALID_SERVER_ID);
    }
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::upnpDisconnectControlPoint(char* address)
{
    QString qaddress = QString(QLatin1String(address));
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpDisconnectControlPoint() - qaddress :" << qaddress;
#endif
    QHostAddress hostAddress = QHostAddress(qaddress);
    m_TerminalMode->upnpDisconnectControlPoint(hostAddress);
    m_appInterface->onStatus(LibInterface::UPNP_DISCONNECT_CONTROL_POINT,AppInterface::SUCCESS);
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::upnpGetApplicationStatus(char*  appId)
{
    QTmApplication* application = getApplicationById(appId);
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpGetApplicationStatus - application :" << application;
#endif
    int ret_status = AppInterface::FAILURE;
    if(application)
    {
        ret_status = m_TerminalMode->upnpGetApplicationStatus(application);
    }
    else
    {
        m_appInterface->onStatus(LibInterface::UPNP_GET_APPLICATION_STATUS,AppInterface::ERR_INVALID_APP_ID);
    }
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpGetApplicationStatus - ret_status :" << ret_status;
#endif    
    return AppInterface::SUCCESS;
}

//int LibInterfaceImpl::upnpLaunchApplication(char* appId)
int LibInterfaceImpl::upnpLaunchApplication(const QString appId)
{
    qDebug()<<"LibInterfaceImpl::upnpLaunchApplication - received application id :" << appId;
    QTmApplication* application = getApplicationById(appId);
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpLaunchApplication - application :" << application;
#endif
    bool launched = false;
    if(application)
    {
        launched = m_TerminalMode->upnpLaunchApplication(application);
        if(launched)
        {
            m_appInterface->onStatus(LibInterface::UPNP_LAUNCH_APPLICATION,AppInterface::SUCCESS);
        }
        else
        {
             m_appInterface->onStatus(LibInterface::UPNP_LAUNCH_APPLICATION,AppInterface::FAILURE);
        }
    }
    else
    {
        //m_appInterface->onStatus(LibInterface::UPNP_LAUNCH_APPLICATION,AppInterface::ERR_INVALID_APP_ID);
    }
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpLaunchApplication - launched :" << launched;
#endif    
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::upnpTerminateApplication(char* appId)
{
    QTmApplication* application = getApplicationById(appId);
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpTerminateApplication - application :" << application;
#endif
    bool terminated = false;
    if(application)
    {
        qDebug()<< "About to terminate "<< application->name();
        terminated = m_TerminalMode->upnpTerminateApplication(application);
        if(terminated)
        {
            m_appInterface->onStatus(LibInterface::UPNP_TERMINATE_APPLICATION,AppInterface::SUCCESS);
        }
        else
        {
            m_appInterface->onStatus(LibInterface::UPNP_TERMINATE_APPLICATION,AppInterface::FAILURE);
        }
    }
    else
    {
        m_appInterface->onStatus(LibInterface::UPNP_TERMINATE_APPLICATION,AppInterface::ERR_INVALID_APP_ID);
    }
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpTerminateApplication - terminated :" << terminated;
#endif    
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncStartClient()
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncStartClient()";
#endif
    int clientId = m_TerminalMode->vncStartClient();
    QTmSurfaceImpl* libSurface =  QTmSurfaceImpl::getInstance();//new QTmSurfaceImpl();

    libSurface->setClientId(clientId);

    m_TerminalMode->connectRfbClientToSurface(clientId,(QTmSurface*)libSurface);
    libSurface->setAppInterface(m_appInterface);
    m_appInterface->vncClientStarted(clientId);

#ifdef BUFFER_MANAGER_INTEGRATED 
    /*!@{*/
    m_TerminalMode->connectSurfaceToRfbClient(clientId,(QTmSurface*)libSurface);
#endif

#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncStartClient() client Id:" <<clientId;
#endif    
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncConnectClient(char* appId, int clientId)
{
    QTmApplication* application = getApplicationById(appId);
    bool connected = false;
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncConnectClient() application:" << application << ", clientId :" << clientId;
#endif

    if(application)
    {
        connected = m_TerminalMode->vncConnectClient(application,clientId);
        if(connected)
        {
            m_appInterface->onStatus(LibInterface::VNC_CONNECT_CLIENT,AppInterface::SUCCESS);
        }
        else
        {
            m_appInterface->onStatus(LibInterface::VNC_CONNECT_CLIENT,AppInterface::FAILURE);
        }
    }
    else
    {
        m_appInterface->onStatus(LibInterface::VNC_CONNECT_CLIENT,AppInterface::ERR_INVALID_APP_ID);
    }
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncDisconnectClient(int clientId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncDisconnectClient() clientId :" << clientId;
#endif
    m_TerminalMode->vncDisconnectClient(clientId);
    m_appInterface->onStatus(LibInterface::VNC_DISCONNECT_CLIENT,AppInterface::SUCCESS);
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncStopClient(int clientId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncStopClient() clientId :" << clientId;
#endif
    m_TerminalMode->vncStopClient(clientId);
    m_appInterface->onStatus(LibInterface::VNC_STOP_CLIENT,AppInterface::SUCCESS);
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncSetColorFormat(int colorFormat, int clientId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncSetColorFormat() colorFormat:" << colorFormat << ", clientId :" << clientId;
#endif
    m_TerminalMode->vncSetColorFormat((QTmGlobal::TmColorFormat)colorFormat,clientId);
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncSetRunLengthEncoding(bool enable, int clientId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncSetRunLengthEncoding() enable:" << enable << ", clientId :" << clientId;
#endif
    m_TerminalMode->vncSetRunLengthEncoding(enable,clientId);
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncSetIncrementalUpdate(bool enable, int clientId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncSetIncrementalUpdate() enable:" << enable << ", clientId :" << clientId;
#endif
    m_TerminalMode->vncSetIncrementalUpdate(enable, clientId);
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncSetPreferredSize(int width, int height, bool scaling, int clientId, double clientSinglePixelWidth, double clientSinglePixelHeight, bool considerPixelComp)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncSetPreferredSize() :" << width <<","<<height <<", scaling :"<< scaling << ", clientId :" << clientId;
#endif
    m_TerminalMode->vncSetPreferredSize(QSize(width,height),scaling,clientId, clientSinglePixelWidth, clientSinglePixelHeight, considerPixelComp);
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncOnFBReadComplete(int bufferId){
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncOnFBReadComplete() :";
#endif
    QTmSurfaceImpl* libSurface =  QTmSurfaceImpl::getInstance();
    libSurface->onBufferReadComplete(bufferId);
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncClickVirtualKey(int clientId, int keySym)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncClickVirtualKey() keySym:" << keySym << ", clientId :" << clientId;
#endif
    m_TerminalMode->vncClickVirtualKey(keySym, clientId);
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncPointerEvent(int clientId, int x, int y, unsigned int pressed)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncPointerEvent() pressed:"<<pressed<< ","<<  x <<","<< y << ", clientId :" << clientId ;
#endif
    m_TerminalMode->pointerEvent(clientId, x, y, pressed);
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncSupportedKeyEvents(int clientId, int knobKey, int deviceKey, int mediaKey)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncSupportedKeyEvents() : "<< "knobKey - "<< knobKey<< ", deviceKey - "<<  deviceKey <<",mediaKey - "<< mediaKey;
#endif
    m_TerminalMode->vncSupportedKeyEvents(clientId, knobKey, deviceKey, mediaKey);
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncKeyEvent(int clientId, int keyId, bool pressed)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncKeyEvent() pressed:" << pressed << " keyId :"<< keyId <<  ", clientId :" << clientId;
#endif
    m_TerminalMode->keyEvent(clientId,keyId,pressed);
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncBlockFramebuffer(int x, int y, int width, int height, int appId, int reason, int id)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncBlockFramebuffer() :" << x <<","<<y<<","<< width <<","<< height <<",appid:"<<appId <<"reason:"<<reason;
#endif
    QRect rect(x, y, width, height);
    m_TerminalMode->vncBlockFramebuffer(rect, appId, reason, id);
    return AppInterface::SUCCESS;
}


int LibInterfaceImpl::vncResetFbUpdateRequestTimer()
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncResetFbUpdateRequestTimer()";
#endif
    m_TerminalMode->vncResetFbUpdateRequestTimer();
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncPauseFbUpdateRequest()
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncPauseFbUpdateRequest()";
#endif
    m_TerminalMode->vncPauseFbUpdateRequest();
    return AppInterface::SUCCESS;
}

int LibInterfaceImpl::vncResumeFbUpdateRequest()
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncResumeFbUpdateRequest()";
#endif
    m_TerminalMode->vncResumeFbUpdateRequest();
    return AppInterface::SUCCESS;
}

/** -------------------------------------------------------------------------
    public slots:
---------------------------------------------------------------------------**/


void LibInterfaceImpl::vncResized(QSize size, int clientId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncResized() size:" << size <<", clientId:" << clientId;
#endif
    m_appInterface->vncClientResized(size.width(),size.height(),clientId);
}

void LibInterfaceImpl::vncClientContextInformation(QRect area, int appId, QTmAppInfo *appCategory, QTmDisplayInfo *displayCategory, int clientId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncClientContextInformation() area:" << area <<","<< appId<<","<< appCategory->category() << "," <<appCategory->trust() << "," <<displayCategory->category()<<","<<displayCategory->rules()<<","<<displayCategory->trust()<<","<<clientId;
#endif

    m_appInterface->vncClientContextInformation(area.x(),area.y(),area.width(),area.height(),appId,appCategory->category(),appCategory->trust(),displayCategory->category(),displayCategory->rules(),displayCategory->trust(),clientId);
}

void LibInterfaceImpl::vncDisconnected(int clientId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncDisconnected() clientId:" << clientId;
#endif
    m_appInterface->vncClientDisconnected(clientId);
}

void LibInterfaceImpl::vncConnected(int clientId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncConnected() clientId:" << clientId;
#endif
    m_appInterface->vncClientConnected(clientId);
}

void LibInterfaceImpl::vncClientDisplayConfiguration(int clientId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncClientDisplayConfiguration() clientId:" << clientId;
#endif
    m_appInterface->vncClientDisplayConfiguration(clientId);
}

void LibInterfaceImpl::vncClientEventConfiguration(int clientId, int knobKey, int deviceKey, int mediaKey)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncClientEventConfiguration() clientId:" << clientId;
#endif
    qDebug()<<"LibInterfaceImpl::vncClientEventConfiguration() knobKey:" << knobKey;
    qDebug()<<"LibInterfaceImpl::vncClientEventConfiguration() deviceKey:" << deviceKey;
    qDebug()<<"LibInterfaceImpl::vncClientEventConfiguration() mediaKey:" << mediaKey;
    m_appInterface->vncClientEventConfiguration(clientId, knobKey, deviceKey, mediaKey);
}

void LibInterfaceImpl::vncClientKeyboardTrigger(QPoint point, bool remove, int clientId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncClientKeyboardTrigger() :" << point <<","<<remove<< ","<< clientId;
#endif
    m_appInterface->vncClientKeyboardTrigger(point.x(),point.y(),remove,clientId);
}

void LibInterfaceImpl::vncClientKeyEventList(QList<unsigned int> *list, unsigned int counter, int clientId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncClientKeyEventList() size:" << list->size() <<", clientId:" << clientId;
#endif
    m_appInterface->vncClientKeyEventList(list,counter,clientId);
}

void LibInterfaceImpl::vncClientFbAlternativeText(unsigned int appId, QString text, int clientId)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::vncClientFbAlternativeText() appId:" << appId <<",text:"<<text<<", clientId:" << clientId;
#endif
    QByteArray ba = text.toLatin1();
    m_appInterface->vncClientFbAlternativeText(appId,ba.data(),clientId);
}

void LibInterfaceImpl::upnpApplicationListingChanged(QTmRemoteServer *remoteServer)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpApplicationListingChanged()";
#endif
    QList<QTmApplication *> appList = remoteServer->applicationList();

    foreach (QTmApplication *application , appList) {
        if(!m_appIdToApplicationMap->contains(application->appId()))
        {
            m_appIdToApplicationMap->insert(application->appId(),application);
#ifdef QT_CODE_DEBUG
    qDebug()<<"application::" << application->name() << ", id:" <<application->appId();
#endif
        }
    }
    m_appInterface->applicationListingChanged(&appList);
}

void LibInterfaceImpl::upnpApplicationAdded(QTmApplication *application)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpApplicationAdded() application:" << application->name() << ", id:" <<application->appId();
#endif
    if(!m_appIdToApplicationMap->contains(application->appId()))
    {
        m_appIdToApplicationMap->insert(application->appId(), application);
        m_appInterface->applicationAdded(application);
    }
}

void LibInterfaceImpl::upnpApplicationStarted(QTmApplication *application)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpApplicationStarted() application:" << application->name() << ", id:" <<application->appId();
#endif
    QByteArray ba = application->appId().toLatin1();
    QByteArray ba1 = application->url().toString().toLatin1();
    m_appInterface->applicationLaunched(ba.data(), ba1.data());
}

void LibInterfaceImpl::upnpApplicationTerminated(QTmApplication *application)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpApplicationTerminated() application:" << application->name() << ", id:" <<application->appId();
#endif
    QByteArray ba = application->appId().toLatin1();
    m_appInterface->applicationTerminated(ba.data());
}

void LibInterfaceImpl::upnpApplicationStatusChanged(QTmApplication *application)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::upnpApplicationStatusChanged() application:" << application->name() << ", id:" <<application->appId();
#endif
    QByteArray ba = application->appId().toLatin1();
    m_appInterface->applicationStatusChanged(ba.data(),application->status());
}

void LibInterfaceImpl::remoteServerDeviceAdded(QTmRemoteServer *remoteServer)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::remoteServerDeviceAdded()";
#endif
    if(!m_remoteServerList->contains(remoteServer))
    {
        connect(remoteServer, SIGNAL(applicationLaunched(QTmApplication*)),
                this,         SLOT(upnpApplicationStarted(QTmApplication*)));
        connect(remoteServer, SIGNAL(applicationTerminated(QTmApplication*)),
                this,         SLOT(upnpApplicationTerminated(QTmApplication*)));
        connect(remoteServer, SIGNAL(applicationStatusChanged(QTmApplication*)),
                this,         SLOT(upnpApplicationStatusChanged(QTmApplication*)));
        connect(remoteServer, SIGNAL(applicationAdded(QTmApplication*)),
                this,         SLOT(upnpApplicationAdded(QTmApplication*)));
        connect(remoteServer, SIGNAL(applicationListingChanged(QTmRemoteServer*)),
                this,  SLOT(upnpApplicationListingChanged(QTmRemoteServer*)));

        //Add remote server from list

        m_remoteServerList->append(remoteServer);

#ifdef QT_CODE_DEBUG
        qDebug() << "MainApplication: New Terminal Mode UPnP Server detected.";
        qDebug() << "MainApplication: - UDN:"  << remoteServer->UDN();
        qDebug() << "MainApplication: - id:"   << remoteServer->remoteServerId();
        qDebug() << "MainApplication: - Host:" << remoteServer->remoteServerUrl().host();
#endif
    }
    else
    {
#ifdef QT_CODE_DEBUG
        qDebug() << "MainApplication: Terminal Mode UPnP Server with UDN:" << remoteServer->UDN() << " id:" << remoteServer->remoteServerId() << "already exists.";
#endif
    }
    m_appInterface->remoteServerDeviceAdded(remoteServer);
}

void LibInterfaceImpl::remoteServerDeviceRemoved(QTmRemoteServer *remoteServer)
{
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::remoteServerDeviceRemoved()" << remoteServer;
#endif

    if(remoteServer == NULL)
    {
        //perform dummy emit to ensure the cleanup at upper layers/managers on every ML phone disconnect.
        m_appInterface->remoteServerDeviceRemoved(-1);
        return;
    }

    //Clear applications list
    int remoteServerId = remoteServer->remoteServerId();
#ifdef QT_CODE_DEBUG
    qDebug()<<"LibInterfaceImpl::remoteServerDeviceRemoved() server id:" << remoteServerId;
#endif
    //Close all VNC clients associated with this remote server device
    int vncId;
    QHash<int, int>::iterator i = m_remoteServerIdToVncIdMap->find(remoteServerId);

#ifdef QT_CODE_DEBUG
    qDebug()<<"About to call disconnect on all vnc clients of " << remoteServerId ;
    qDebug()<<"Remote Server ID previously Stored " << (i != m_remoteServerIdToVncIdMap->end() && i.key() == remoteServerId);
#endif

    while (i != m_remoteServerIdToVncIdMap->end() && i.key() == remoteServerId) {
        vncId = i.value();
#ifdef QT_CODE_DEBUG
        qDebug()<<"Disconnecting VNC client id:" << vncId;
#endif
        m_TerminalMode->vncDisconnectClient(vncId);
        m_TerminalMode->vncStopClient(vncId);
        i = m_remoteServerIdToVncIdMap->find(remoteServerId);
    }
    //Remove all mappings for this remote server device
    m_remoteServerIdToVncIdMap->remove(remoteServerId);

    //Deinit signal/slot mechanism
    disconnect(remoteServer, SIGNAL(applicationLaunched(QTmApplication*)),
               this,         SLOT(upnpApplicationStarted(QTmApplication*)));
    disconnect(remoteServer, SIGNAL(applicationTerminated(QTmApplication*)),
               this,         SLOT(upnpApplicationTerminated(QTmApplication*)));
    disconnect(remoteServer, SIGNAL(applicationStatusChanged(QTmApplication*)),
               this,         SLOT(upnpApplicationStatusChanged(QTmApplication*)));
    disconnect(remoteServer, SIGNAL(applicationAdded(QTmApplication*)),
               this,         SLOT(upnpApplicationAdded(QTmApplication*)));
    disconnect(remoteServer, SIGNAL(applicationListingChanged(QTmRemoteServer*)),
               this,  SLOT(upnpApplicationListingChanged(QTmRemoteServer*)));

    m_remoteServerList->removeAll(remoteServer);

#ifdef QT_CODE_DEBUG
    qDebug() << "Removing application ids ...";
#endif
    QList<QString> appIds = m_appIdToApplicationMap->keys();
    foreach (QString appId , appIds)
    {
        m_appIdToApplicationMap->remove(appId);
#ifdef QT_CODE_DEBUG
        qDebug() << "Removed :" << appId;
#endif
    }
    m_appInterface->remoteServerDeviceRemoved(remoteServer->remoteServerId());
}
